package fi.helsinki.dacopan.model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Default {@link fi.helsinki.dacopan.model.DataView} implementation.
 * 
 * <p>Note: this class is not thread-safe, it should be populated once
 * and thereafter it can be accessed by multiple concurrent threads.</p>
 * 
 * @version $Id: DefaultDataView.java,v 1.21 2005/04/06 16:54:25 pamietti Exp $
 */
public class DefaultDataView implements DataView {

    /*
     * Model objects (hosts, flows etc) are stored in maps because we need
     * to have lookups by id. The maps are LinkedHashMaps, since instances
     * of that class retain the same iteration order (most maps don't).
     */
    
    /** hosts by id */
    protected Map hosts = new LinkedHashMap();

    /** flows by id */
    protected Map flows = new LinkedHashMap();
    
    /** links by id */
    protected Map links = new LinkedHashMap();
    
    /** layers by id */
    protected Map layers = new HashMap();
    
    /** protocols by id */
    protected Map protocols = new LinkedHashMap();
    
    /** all the transfer units by id */
    protected Map units = new HashMap();
    
    /** maps layers to lists of units on that level */
    protected Map layersToUnits = new HashMap();
    
    /** cached values for end times of each layer */
    protected Map endTimes = new HashMap();
    
    /** end time of the whole scenario */
    protected float scenarioEndTime = 0.0f;
    
    /** maps variable definitions to the maximum lengths of their values */
    protected Map variableLengths = new HashMap();
    
    /**
     * Gets the unit list for the layer of the specified unit, creates
     * a new list if none exists.
     */
    protected List getOrCreateUnitList(TransferUnit unit) {
        List units = (List)layersToUnits.get(unit.getProtocol().getLayer());
        if(units == null) {
            units = new ArrayList();
            layersToUnits.put(unit.getProtocol().getLayer(), units);
        }
        return units;
    }
    
    // DataView implementation follows
    
    public void addHost(Host host) {
        hosts.put(new Long(host.getId()), host);
    }
    
    public void addFlow(Flow flow) {
        flows.put(new Long(flow.getId()), flow);
    }
    
    public void addLink(Link link) {
        links.put(new Long(link.getId()), link);
    }
    
    public void addLayer(Layer layer) {
        layers.put(new Long(layer.getId()), layer);
    }
    
    public void addProtocol(Protocol protocol) {
        protocols.put(new Long(protocol.getId()), protocol);
    }

    public void addTransferUnit(TransferUnit unit, boolean recurse) {
        List unitsForLayer = getOrCreateUnitList(unit);
        unitsForLayer.add(unit);
        Collections.sort(unitsForLayer);
        units.put(new Long(unit.getId()), unit);
        updateEndTime(unit);
        updateMaxLengths(unit);
        if(recurse) {
            for(Iterator it = unit.getChildren().iterator(); it.hasNext(); ) {
                addTransferUnit((TransferUnit)it.next(), true);
            }
        }
    }
    
    public List getUnitsForLayer(Layer layer) {
        List units = (List)layersToUnits.get(layer);
        if(units == null) {
            return Collections.EMPTY_LIST;
        }
        return units;
    }

    public List getUnitsForLayer(Layer layer, float start, float end) {
        if(start > end) {
            throw new IllegalArgumentException("start > end");
        }
        List units = getUnitsForLayer(layer);
        List result = new ArrayList();
        for(Iterator i = units.iterator(); i.hasNext(); ) {
            TransferUnit unit = (TransferUnit)i.next();
            if(unit.getSendStart() <= end && unit.getReceiveEnd() >= start) {
                result.add(unit);
            }
        }
        return result;
    }
    
    public TransferUnit getUnitById(long id) {
        return (TransferUnit)units.get(new Long(id));
    }
    
    public List getHosts() {
        return new ArrayList(hosts.values());
    }
    
    public Host getHostById(long id) {
        return (Host)hosts.get(new Long(id));
    }
    
    public List getFlows() {
        return new ArrayList(flows.values());
    }
    
    public Flow getFlowById(long id) {
        return (Flow)flows.get(new Long(id));
    }
    
    public List getLayers() {
        List result = new ArrayList(layers.values());
        Collections.sort(result);
        return result;
    }
    
    public Layer getLayerById(long id) {
        return (Layer)layers.get(new Long(id));
    }
    
    public List getProtocols() {
        return new ArrayList(protocols.values());
    }
    
    public Protocol getProtocolById(long id) {
        return (Protocol)protocols.get(new Long(id));
    }
    
    public List getLinks() {
        return new ArrayList(links.values());
    }
    
    public Link getLinkById(long id) {
        return (Link)links.get(new Long(id));
    }
    
    public int getValueMaxLength(VariableDefinition def) {
        if(def == null) {
            throw new IllegalArgumentException("Null variable definition");
        }
        Integer maxLength = (Integer)this.variableLengths.get(def);
        if(maxLength == null) {
            return 0;
        }
        return maxLength.intValue();
    }
    
    public List getAvailableVariables(Layer layer) {
        List result = new LinkedList();
        for (Iterator it = getProtocols().iterator(); it.hasNext();) {
            Protocol proto = (Protocol) it.next();
            if(proto.getLayer().equals(layer)) {
                for (Iterator vars = proto.getVariables().iterator(); vars.hasNext();) {
                    VariableDefinition def = (VariableDefinition) vars.next();
                    if(getValueMaxLength(def) > 0) {
                        result.add(def);
                    }
                }
            }
        }
        return result;
    }
    
    public StepIterator getStepIterator(Layer layer) {
        List units = getUnitsForLayer(layer);
        Set timestamps = new HashSet();
        timestamps.add(new Float(0.0f));
        timestamps.add(new Float(scenarioEndTime));
        for(Iterator i = units.iterator(); i.hasNext(); ) {
            TransferUnit unit = (TransferUnit)i.next();
            timestamps.add(new Float(unit.getSendStart()));
            if(!unit.isDropped()) {
            	timestamps.add(new Float(unit.getReceiveEnd()));
            }
            if(!unit.isOneUnit()) {
                timestamps.add(new Float(unit.getSendEnd()));
                timestamps.add(new Float(unit.getReceiveStart()));
            }
        }
        Float[] sortedTimes = (Float[])timestamps.toArray(new Float[timestamps.size()]);
        Arrays.sort(sortedTimes);
        return new DefaultStepIterator(sortedTimes);
    }
    
    public float getEndTime() {
        return scenarioEndTime;
    }
    
    public float getEndTimeForLayer(Layer layer) {
        Float endTime = (Float)endTimes.get(layer);
        if(endTime == null) {
            return 0.0f;
        }
        return endTime.floatValue();
    }
    
    /**
     * Updates the end times of the unit's layer and the whole scenario, if necessary.
     */
    private void updateEndTime(TransferUnit unit) {
        float end = getEndTimeForLayer(unit.getProtocol().getLayer());
        if(unit.getReceiveEnd() > end) {
            endTimes.put(unit.getProtocol().getLayer(), new Float(unit.getReceiveEnd()));
        }
        if(unit.getReceiveEnd() > scenarioEndTime) {
            scenarioEndTime = unit.getReceiveEnd();
        }
    }

    /**
     * Updates the max lengths of the variables of the unit's protocol.
     */
    private void updateMaxLengths(TransferUnit unit) {
        for (Iterator it = unit.getProtocol().getVariables().iterator(); it.hasNext();) {
            VariableDefinition def = (VariableDefinition)it.next();
            String value = unit.getValue(def);
            int length = value == null ? 0 : value.length();
            Integer max = (Integer)variableLengths.get(def);
            if(max == null || length > max.intValue()) {
                variableLengths.put(def, new Integer(length));
            }
        }
    }
    
    private static class DefaultStepIterator implements StepIterator {

        private int index;
        private Float[] steps;

        private DefaultStepIterator(Float[] steps) {
            this.steps = steps;
            this.index = 0;
        }
        
        public boolean hasNext() {
            return index < steps.length - 1;
        }
        
        public boolean hasPrevious() {
            return index > 0;
        }
        
        public float next() {
            if (!hasNext()) {
                throw new IllegalStateException("Already at end");
            }
            index++;
            return current();
        }
        
        public float previous() {
            if (!hasPrevious()) {
                throw new IllegalStateException("Already at beginning");
            }
            index--;
            return current();
        }

        public float first() {
            index = 0;
            return current();
        }

        public float last() {
            index = steps.length - 1;
            return current();
        }
        
        public float current() {
            return steps[index].floatValue();
        }
        
        public float getNextForTime(float time) {
            int newIndex = Arrays.binarySearch(steps, new Float(time));
            if (newIndex < 0) {
                // if not exact match, return next
                newIndex = -newIndex - 1;
            } else {
                // exact match, skip it and return next
                newIndex++;
            }
            if (newIndex > steps.length-1) newIndex = steps.length-1;
            index = newIndex;
            return current();
        }
        
        public float getPreviousForTime(float time) {
            int newIndex = Arrays.binarySearch(steps, new Float(time));
            if (newIndex < 0) {
                // not exact match, return previous
                newIndex = -newIndex - 2;
            } else {
                // exact match, return previous
                newIndex--;
            }
            if (newIndex < 0) newIndex = 0;
            index = newIndex;
            return current();
        }
        
    }
    
    /**
     * Mostly for unit testing purposes, does huge amounts of comparisons.
     */
    public boolean equals(Object o) {
        if(!(o instanceof DefaultDataView)) return false;
        DefaultDataView other = (DefaultDataView)o;
        if(!this.flows.equals(other.flows)) return false;
        if(!this.hosts.equals(other.hosts)) return false;
        if(!this.layers.equals(other.layers)) return false;
        if(!this.links.equals(other.links)) return false;
        if(!this.protocols.equals(other.protocols)) return false;
        if(!this.units.equals(other.units)) return false;
        return true;
    }

}