package fi.helsinki.dacopan.model;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * This class encloses the model of an encapsulation tree. It provides the services that are needed by the ENC
 * animation to properly display the encapsulation animation of a particular transfer unit.
 * 
 * The class is instantiated each time an encapsulation is to be shown, and constructed with transfer unit or the unit
 * itself.
 * 
 * Normally we wouldn't care who authors our classes, but this one is a little bit tricky. Ask Jonathan or Jarkko if
 * you need to understand this in more depth.
 */

public class ENCTreeModel {

    /**
     * This inner class encloses the concept of a single layer that's going to be drawn in the ENC animation. It
     * consists of
     * <UL>
     * <LI>A <code>List</code> of <code>DrawableTransferUnit</code> objects.</LI>
     * <LI>Information whether there are siblings to the left</LI>
     * <LI>Information whether there are siblings to the right</LI>
     * </UL>
     *  
     */
    public static class DrawableEncLayer {
        private List drawableUnits;
        private boolean siblingsToTheLeft = false;
        private boolean siblingsToTheRight = false;

        public DrawableEncLayer() {
            this.drawableUnits = new LinkedList();
        }

        /**
         * This constructor takes a List of TransferUnits, and sets up
         * the internal list to be a List of DrawableTransferUnits.
         * @param list
         */
        public DrawableEncLayer(List list) {
            this();
            for (Iterator iter = list.iterator(); iter.hasNext();) {
                TransferUnit tu = (TransferUnit) iter.next();
                drawableUnits.add(new DrawableTransferUnit(tu));
            }
        }

        /**
         * @return Returns the siblingsToTheLeft.
         */
        public boolean hasSiblingsToTheLeft() {
            return this.siblingsToTheLeft;
        }

        /**
         * @param siblingsToTheLeft
         *            The siblingsToTheLeft to set.
         */
        public void setSiblingsToTheLeft(boolean siblingsToTheLeft) {
            this.siblingsToTheLeft = siblingsToTheLeft;
        }

        /**
         * @return Returns the siblingsToTheRight.
         */
        public boolean hasSiblingsToTheRight() {
            return this.siblingsToTheRight;
        }

        /**
         * @param siblingsToTheRight
         *            The siblingsToTheRight to set.
         */
        public void setSiblingsToTheRight(boolean siblingsToTheRight) {
            this.siblingsToTheRight = siblingsToTheRight;
        }

        /**
         * @return Returns the List of DrawableTransferUnit objects.
         */
        public List getDrawableUnits() {
            return this.drawableUnits;
        }
        
        /**
         * Gets the first unit of the layer.
         */
        public DrawableTransferUnit getFirstUnit() {
            if(this.drawableUnits.size() > 0) {
                return (DrawableTransferUnit)drawableUnits.get(0);
            }
            return null;
        }

        /**
         * Gets the last unit of the layer.
         */
        public DrawableTransferUnit getLastUnit() {
            if(this.drawableUnits.size() > 0) {
                return (DrawableTransferUnit)drawableUnits.get(drawableUnits.size() - 1);
            }
            return null;
        }

    } //inner class

    /**
     * This inner class encloses the concept of a single unit that's going to be drawn in the ENC animation. It
     * contains
     * <UL>
     * <LI>A <code>TransferUnit</code> which will be draw</LI>
     * <LI>Information whether the unit continues to the left</LI>
     * <LI>Information whether the unit continues to the right</LI>
     * </UL>
     *  
     */
    public static class DrawableTransferUnit {
        private TransferUnit transferUnit;
        private boolean moreToTheLeft;
        private boolean moreToTheRight;
        private List visibleChildIds;

        public DrawableTransferUnit(TransferUnit unit) {
            this.transferUnit = unit;
            this.moreToTheLeft = false;
            this.moreToTheRight = false;
            this.visibleChildIds = new LinkedList();
        }

        /**
         * @return Returns whether there are more sibling units to the left of this unit.
         */
        public boolean hasMoreToTheLeft() {
            return this.moreToTheLeft;
        }

        /**
         * @param moreToTheLeft
         *            Whether there are more units to the left.
         */
        public void setMoreToTheLeft(boolean moreToTheLeft) {
            this.moreToTheLeft = moreToTheLeft;
        }

        /**
         * @return Returns whether there are more sibling units to the right of this unit.
         */
        public boolean hasMoreToTheRight() {
            return this.moreToTheRight;
        }

        /**
         * @param moreToTheRight
         *            Whether there are more units to the right.
         */
        public void setMoreToTheRight(boolean moreToTheRight) {
            this.moreToTheRight = moreToTheRight;
        }

        /**
         * @return Returns the transferUnit.
         */
        public TransferUnit getTransferUnit() {
            return this.transferUnit;
        }

        /**
         * @return Returns the visibleChildIds.
         */
        public List getVisibleChildIds() {
            return visibleChildIds;
        }
        
        /**
         * @param visibleChildIds The visibleChildIds to set.
         */
        public void setVisibleChildIds(List visibleChildIds) {
            this.visibleChildIds = visibleChildIds;
        }
        
        public boolean equals(Object o) {
            if(o instanceof DrawableTransferUnit) {
                return ((DrawableTransferUnit)o).transferUnit.equals(this.transferUnit);
            }
            return false;
        }
        
        // TODO hashCode depends on TransferUnit.hashCode()

    } //inner class

    /**
     * The transfer unit who's encapsulation animation we wish to model, later highlighted in the animation.
     */
    private TransferUnit selectedUnit;
    /**
     * The root unit of the selected unit. The root unit actually defines the enc tree, as the very same enc tree will
     * be constructed for every selectedUnit in that tree.
     */
    private TransferUnit rootUnit;

    //an array of Lists, array is of length getLayers().size()
    private List[] layeredEncTree;
    private DrawableEncLayer[] layeredEncTreeToDraw;

    /**
     * a class-wide reference to the dataView.getLayers()
     */
    private List layers;
    private static final int AMOUNT_TO_SHOW = 3;

    /**
     * An Enc Tree Model instance must be constructed with a particular transfer unit.
     * 
     * @param transferUnit
     *            The selected transfer unit, whos enc tree this is
     * @param dataView
     *            The current dataview instance
     */
    public ENCTreeModel(TransferUnit selectedUnit, DataView dataView) {
        this.layers = dataView.getLayers(); //set up the reference
        this.selectedUnit = selectedUnit;
        this.rootUnit = selectedUnit.getRoot(); //set root unit

        //This represents the WHOLE encapsulation tree
        layeredEncTree = new List[layers.size()];
        for (int i = 0; i < layers.size(); i++) {
            layeredEncTree[i] = new LinkedList();
        }

        /*
         * This builds an array of DrawableEncLayer objects.
         */
        layeredEncTreeToDraw = new DrawableEncLayer[layers.size()];

        //this recursive call adds everything under the rootUnit
        //to the full enc tree listArray.
        addSubTreeToListArray(layeredEncTree, rootUnit);

        //figure out which units to actually draw, and put them
        //into the draw listArray
        layeredEncTreeToDraw = findTreeToDraw(layeredEncTree, selectedUnit);
    }

    /**
     * This method adds a TransferUnit to the correct DrawableEncLayer in the passed array of DrawableEncLayers.
     * The correct layer is the one that corrensponds to the Unit's layer.
     * 
     * @param drawableTree
     * @param unit
     */
    private void addUnitToItsOwnDrawableLayerList(DrawableEncLayer[] drawableTree, TransferUnit unit) throws IllegalStateException {
        if (layers.indexOf(unit.getProtocol().getLayer()) == -1) {
            throw new IllegalStateException(
                "Holy layer, Batman."
                    + " TransferUnit's layer doesn't match any known layer"
                    + " in DataView. How did that happen?");
        }
        DrawableTransferUnit drawableUnit = new DrawableTransferUnit(unit);
        if (!drawableTree[layers.indexOf(unit.getProtocol().getLayer())].getDrawableUnits().contains(drawableUnit)) {
            drawableTree[layers.indexOf(unit.getProtocol().getLayer())].getDrawableUnits().add(drawableUnit);
        }
    }

    /**
     * This method adds a TransferUnit to the correct List in the passed array of Lists. The correct List in the array
     * is the one that corrensponds to the Unit's layer Example: if listArray looks like [ List , List , List ] then if
     * unit.layer = Transport --> listArray[2].add(unit); this means that Transport = 2nd layer (from the bottom, or
     * the start of list)
     * 
     * @param listArray
     * @param tu
     */
    private void addUnitToItsOwnLayerList(List[] listArray, TransferUnit unit) throws IllegalStateException {
        if (layers.indexOf(unit.getProtocol().getLayer()) == -1) {
            throw new IllegalStateException(
                "Holy layer, Batman."
                    + " TransferUnit's layer doesn't match any known layer"
                    + " in DataView. How did that happen?");
        }
        if (!listArray[layers.indexOf(unit.getProtocol().getLayer())].contains(unit)) {
            listArray[layers.indexOf(unit.getProtocol().getLayer())].add(unit);
        }
    }

    /**
     * This recursive method adds the passed unit and all of its children to the passed array of lists. It uses the
     * method addUnitToItsOwnLayerList() to actually add each unit.
     * 
     * @param unit
     */
    private void addSubTreeToListArray(List[] listArray, TransferUnit unit) {

        addUnitToItsOwnLayerList(listArray, unit);
        if (unit.getChildren() != null) {
            for (Iterator iter = unit.getChildren().iterator(); iter.hasNext();) {
                TransferUnit child = (TransferUnit) iter.next();
                addSubTreeToListArray(listArray, child);
            }
        }
    }

    /**
     * Creates the tree that the ENCPanel should then be able to draw.
     * 
     * @param wholeTree
     * @param selectedUnit
     * @return DrawableEncLayer[] An array of layers to draw
     */
    private DrawableEncLayer[] findTreeToDraw(List[] wholeTree, TransferUnit selectedUnit) {

        DrawableEncLayer[] treeToDraw = new DrawableEncLayer[wholeTree.length];
        for (int i = 0; i < treeToDraw.length; i++) {
            treeToDraw[i] = new DrawableEncLayer();
        }

        /*
         * We always start building the tree from the lowest layer so if the user
         * has selected a higher than lowest layer unit, we need to
         * find its "youngest" child :)
         */
        while (!selectedUnit.getChildren().isEmpty()) {
            selectedUnit = (TransferUnit) selectedUnit.getChildren().get(0);
        }

        /*
         * First create the list for the lowest level units that are selected to be drawn. In this list we try to add
         * units equally from both sides of the selected unit, but if that is not possible, just take them from the
         * other side.
         */
        LinkedList selectedLowestUnits = new LinkedList();
        selectedLowestUnits.add(selectedUnit);
        int count = 1; // Count is one as the selected unit is already in the list...

        // The lowest layer is 0
        int currentRadius = 1;
        boolean moreUnitsFound = true;

        int rightIndex;
        int leftIndex;

        // Find the lowest layer transfer units around selected
        while (count < AMOUNT_TO_SHOW && moreUnitsFound) {

            rightIndex = wholeTree[0].indexOf(selectedUnit) + currentRadius;
            leftIndex = wholeTree[0].indexOf(selectedUnit) - currentRadius;

            if (rightIndex > wholeTree[0].size() && leftIndex < 0) {
                moreUnitsFound = false;
            }

            // First try right: right index is inside the array
            if (rightIndex < wholeTree[0].size() && count < AMOUNT_TO_SHOW) {
                count++;
                selectedLowestUnits.addLast((TransferUnit) wholeTree[0].get(rightIndex));
            }
            // Then try left
            if (leftIndex >= 0 && count < AMOUNT_TO_SHOW) {
                count++;
                selectedLowestUnits.addFirst((TransferUnit) wholeTree[0].get(leftIndex));
            }

            // Search one unit further
            currentRadius++;
        }
                
        /*
         * Once the lowest units to be drawn have been selected, we just add them and their ansestors to the "to be
         * drawn" list
         */
        for (Iterator iter = selectedLowestUnits.iterator(); iter.hasNext();) {
            TransferUnit unit = (TransferUnit) iter.next();

            do {
                addUnitToItsOwnDrawableLayerList(treeToDraw, unit);

                unit = unit.getParent();
            } while (unit != null);
        }
        
        //finalize treeToDraw by iterating through each list on each layer
        //and adding boolean hasSiblings information, and then iterating through
        //each unit's children to see if they are in the list and add hasMore
        //information for each unit.
        
        int firstToDraw = 0;
        int firstWhole = 0;
        int lastToDraw = -1; //calcluated on the fly for each layer
        int lastWhole = -1;  //calcluated on the fly for each layer
        
        // Note: this loop goes from lowest layer to highest (network -> trans -> app)
        for (int i = 0; i < treeToDraw.length; i++) {
            if(!treeToDraw[i].getDrawableUnits().isEmpty()) {
                lastWhole = wholeTree[i].size()-1;
                
                DrawableTransferUnit leftMost = treeToDraw[i].getFirstUnit();
                DrawableTransferUnit rightMost = treeToDraw[i].getLastUnit();
                
                //if the first draw tree unit isn't the same as first whole tree unit,
                //there are silbings on left
                if(!leftMost.getTransferUnit().equals(wholeTree[i].get(0))){
                    treeToDraw[i].setSiblingsToTheLeft(true);
                }
                //same thing for the last units / right
                if(!rightMost.getTransferUnit().equals(wholeTree[i].get(lastWhole))){
                    treeToDraw[i].setSiblingsToTheRight(true);
                }
                
                // if not at lowest layer
                if(i > 0) {
                    if(leftMost.getTransferUnit().getChildren().size() > 0) {
                        // left most unit has children, check whether the first of them is drawn
                        DrawableTransferUnit firstChild = new DrawableTransferUnit((TransferUnit)leftMost.getTransferUnit().getChildren().get(0));
                        if(!treeToDraw[i - 1].getDrawableUnits().contains(firstChild)) {
                            leftMost.setMoreToTheLeft(true);
                        }
                    }
                    if(rightMost.getTransferUnit().getChildren().size() > 0) {
                        // rightmost unit has children, check whether the last of them is drawn
                        DrawableTransferUnit lastChild = new DrawableTransferUnit((TransferUnit)rightMost.getTransferUnit().getChildren().get(rightMost.getTransferUnit().getChildren().size() - 1));
                        if(!treeToDraw[i - 1].getDrawableUnits().contains(lastChild)) {
                            rightMost.setMoreToTheRight(true);
                        }
                    }
                    for (Iterator it = treeToDraw[i].getDrawableUnits().iterator(); it.hasNext();) {
                        DrawableTransferUnit unit = (DrawableTransferUnit) it.next();
                        int visibleUnits = 0;
                        for (Iterator children = unit.getTransferUnit().getChildren().iterator(); children.hasNext();) {
                            TransferUnit child = (TransferUnit) children.next();
                            if(treeToDraw[i - 1].getDrawableUnits().contains(new DrawableTransferUnit(child))) {
                                unit.getVisibleChildIds().add(new Long(child.getId()));
                            }
                        }
                    }
                }
                
            }
        }
        return treeToDraw;

    }

    /**
     * @return Returns the layeredEncTree.
     */
    public List[] getLayeredEncTree() {
        return layeredEncTree;
    }
    /**
     * @return Returns the layeredEncTreeToDraw.
     */
    public DrawableEncLayer[] getLayeredEncTreeToDraw() {
        return layeredEncTreeToDraw;
    }
    /**
     * @return Returns the selectedUnit to outside users of the model, mainly the ENC panel.
     */
    public TransferUnit getSelectedUnit() {
        return this.selectedUnit;
    }

    /**
     * @return Returns the root unit of the stored enc tree.
     */
    public TransferUnit getRootUnit() {
        return rootUnit;
    }

}
