package fi.helsinki.dacopan.model;

import java.util.LinkedList;
import java.util.List;

/**
 * A host present in a scenario.
 * 
 * @version $Id: Host.java,v 1.7 2004/05/07 09:19:42 aarniala-dacopan Exp $
 */
public class Host implements Identifiable {

    /** unique id of the host */
    protected long id;
    
    /** ip address */
    protected String ip;

    /** hostname */
    protected String hostname;
    
    /** list of static variables associated with this host */
    protected List staticVariables = new LinkedList();
    
    /**
     * Creates a host with the given id, ip (required) and hostname.
     */
    public Host(long id, String ip, String hostname) {
        if(ip == null) {
            throw new IllegalArgumentException("Ip must be non-null");
        }
        this.id = id;
        this.ip = ip;
        if(hostname == null) {
            this.hostname = ip;
        } else {
            this.hostname = hostname;
        }
    }

    /**
     * Creates a host with the given values and static variables.
     */
    public Host(long id, String ip, String hostname, List staticVariables) {
        this(id, ip, hostname);
        setStaticVariables(staticVariables);
    }

    public long getId() {
        return id;
    }

    /**
     * Returns the hostname of the host (may be null).
     */
    public String getHostname() {
        return hostname;
    }

    /**
     * Returns the ip address of the host (guaranteed to be non-null).
     */
    public String getIp() {
        return ip;
    }
    
    /**
     * Returns a List of ALL StaticVariables the host has in the scenario. In
     * case there will be several types of static variables, the StaticVariable
     * class will provide specific methods for more details about the static
     * variable.
     * 
     * @return non-null list of static variables
     */
    public List getStaticVariables() {
        return staticVariables;
    }

    /**
     * Sets the static variables for this host.
     */
    public void setStaticVariables(List staticVariables) {
        if(staticVariables == null) {
            throw new IllegalArgumentException("Static variables must be non-null");
        }
        this.staticVariables = staticVariables;
    }
    
    public boolean equals(Object o) {
        if(this == o) return true;
        if(!(o instanceof Host)) return false;
        Host other = (Host)o;
        if(other.id != id) return false;
        if(!other.ip.equals(ip)) return false;
        if(hostname != null ? !hostname.equals(other.hostname) : other.hostname != null) return false;
        if(!other.staticVariables.equals(staticVariables)) return false;
        return true;
    }
    
    public int hashCode() {
        int result;
        result = (int) (id ^ (id >>> 32));
        result = 29 * result + ip.hashCode();
        result = 29 * result + (hostname != null ? hostname.hashCode() : 0);
        result = 29 * result + staticVariables.hashCode();
        return result;
    }
    
    public String toString() {
        return "Host {" + id + "/" + ip + "/" + hostname + "/" + staticVariables + "}";
    }
    
}
