package fi.helsinki.dacopan.model;

import java.util.LinkedList;
import java.util.List;

import fi.helsinki.dacopan.Localization;

/**
 * Represents a network layer.
 * 
 * @version $Id: Layer.java,v 1.10 2004/05/07 09:19:42 aarniala-dacopan Exp $
 */
public class Layer implements Comparable, Identifiable {
    
    /** unique layer id */
    protected long id;

    /** an arbitrary value for ordering layers */
    protected int order;
    
    /** (non-localized) name of the layer */
    protected String name;
    
    /** protocols that operate on this layer */
    protected List protocols = new LinkedList();

    /**
     * Creates a new layer with the given name and order.
     * 
     * @param id unique identifier
     * @param name of the layer
     * @param order of the layer in the layer stack
     */
    public Layer(long id, String name, int order) {
        if(name == null) {
            throw new IllegalArgumentException("Layer must have a name");
        }
        this.id = id;
        this.name = name;
        this.order = order;
    }
    
    /**
     * Creates a layer with default parameters plus a list of protocols.
     * 
     * @see #Layer(long, String, int)
     */
    public Layer(long id, String name, int order, List protocols) {
        this(id, name, order);
        setProtocols(protocols);
    }
    
    /**
     * Returns the identifier of this layer.
     */
    public long getId() {
        return id;
    }

    /**
     * Returns the name of the layer.
     */
    public String getName() {
        return Localization.getString("model.layer." + name);
    }
    
    /**
     * Returns all the protocols on this layer.
     */
    public List getProtocols() {
        return protocols;
    }

    /**
     * Set the list of protocols on this layer.
     */
    public void setProtocols(List protocols) {
        if(protocols == null) {
            throw new IllegalArgumentException("Protocols can't be null");
        }
        this.protocols = protocols;
    }

    /**
     * Implements ordering of layers lower-layer-first.
     */
    public int compareTo(Object o) {
        return this.order - ((Layer)o).order;
    }
    
    public boolean equals(Object o) {
        if(this == o) return true;
        if(!(o instanceof Layer)) return false;
        Layer other = (Layer)o;
        if(other.id != id) return false;
        if(!other.name.equals(name)) return false;
        return true;
    }
    
    public int hashCode() {
        int result;
        result = (int) (id ^ (id >>> 32));
        result = 29 * result + name.hashCode();
        return result;
    }

    public String toString() {
        return name + " layer";
    }
    
}
