package fi.helsinki.dacopan.scenario;

import java.io.CharArrayReader;
import java.io.CharArrayWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.zip.Deflater;
import java.util.zip.ZipException;

import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.pef.ProtocolEventsReader;
import fi.helsinki.dacopan.pef.XMLProtocolEventsReader;
import fi.helsinki.dacopan.ui.ProgressIndicator;

/**
 * Implements loading and saving of scenario data, including animation
 * sequence, settings and protocol events data. Each of the objects that
 * wishes to save its state to the file should implement {@link Saveable}
 * to provide this class with the data to be saved.
 * 
 * <p>The scenario file is backed by a standard Jar archive.</p>
 * 
 * <p>
 * <h4>Usage</h4>
 * To create a new scenario file, you must provide the default constructor
 * with the target file where the information should be saved along with
 * the protocol events data that is to be saved in the file. For example:
 * <pre>
 * Reader protocolData = ...; // protocol data from some source
 * ScenarioFile file = new ScenarioFile(new File("myscenario.jar"), protocolData);
 * </pre>
 * Note: creating a new scenario file always overwrites any existing one with
 * the same name.
 * </p>
 * <p>
 * The previous step created a scenario file with nothing but the protocol
 * events data. To save other data (e.g. settings), you must provide the
 * ScenarioFile instance with a collection of {@link Saveable saveables}.
 * For example:
 * <pre>
 * // two example classes that implement Saveable
 * class MyClass implements {@link Saveable} {
 *   private String data; // the data that will be saved
 *   public Object getData() { return data }
 *   public void setData(Object o) { this.data = (String)data; }
 * }
 * class MyOtherClass implements {@link Saveable} {
 *   private Integer data;
 *   public Object getData() { return data }
 *   public void setData(Object o) { this.data = (String)data; }
 * }
 * 
 * // instances of these classes
 * MyClass myObject = new MyClass("abc");
 * MyOtherClass myOtherObject = new MyOtherClass(new Integer(123));
 * 
 * // save them to the scenario file
 * Collection toBeSaved = new LinkedList();
 * toBeSaved.add(myObject);
 * toBeSaved.add(myOtherObject);
 * file.save(toBeSaved);
 * </pre>
 * Note: the {@link #save(Collection)} -method doesn't save the Saveable
 * objects themselves, but the values returned from their {@link Saveable#getData()}
 * -methods.
 * </p>
 * <p>
 * To load a scenario file, use the {@link #ScenarioFile(File) single-parameter constructor}
 * that accepts a file. The constructor will throw an exception if the given file
 * is not a valid scenario file.
 * <pre>
 * // fails with {@link InvalidScenarioFileException} if existing is not a valid scenario file
 * ScenarioFile file = new ScenarioFile(existing); 
 * </pre>
 * </p>
 * <p>
 * To load the values saved before, you have to provide the ScenarioFile
 * with a collection of Saveables whose values are to be loaded back along
 * with a DataView that is to be populated with the saved protocol data. For example:
 * <pre>
 * DataView emptyDataView = ...;
 * Collection toBePopulated = new LinkedList();
 * toBePopulated.add(myObject);
 * toBePopulated.add(myOtherObject);
 * 
 * file.load(toBePopulated, emptyDataView);
 * </pre>
 * The {@link #load(Collection)} -method calls the {@link Saveable#setData(Object)}
 * -method of each Saveable passed to it with the saved value as a parameter.
 * </p>
 * 
 * @version $Id: ScenarioFile.java,v 1.10 2005/05/05 20:32:34 tituomin Exp $
 */
public class ScenarioFile {

    /** name of the PEF data entry in the archive */
    private static final String PEF_ENTRY = "___pef_data";

    /** is this file a settings file (false for real scenario files) */
    private boolean isSettingsFile = false;
    
    /** archive file */
    private File file;
    
    /** serializer used to save/load objects */
    private ObjectSerializer serializer = new XStreamObjectSerializer();

    /**
     * Creates (or overwrites existing) an empty scenario file with given
     * protocol data.
     * 
     * @param file in which to store the scenario data
     * @param protocolDataReader reader that provides the protocol data of the scenario
     * @throws IOException if creating the file fails
     */
    public ScenarioFile(File file, Reader protocolDataReader) throws IOException {
        if(file == null || protocolDataReader == null) {
            throw new IllegalArgumentException("Both file and protocolDataReader are required");
        }
        this.file = file;
        saveFile(protocolDataReader, Collections.EMPTY_LIST);
    }

    /**
     * Loads an existing scenario file.
     * 
     * @param file the archive file, must exist
     * @throws IOException if loading the file fails
     * @throws FileNotFoundException if the scenario file doesn't exist
     * @throws InvalidScenarioFileException if the file is not a valid scenario file
     */
    public ScenarioFile(File file) throws IOException, FileNotFoundException, InvalidScenarioFileException {
        if(!file.exists()) {
            throw new FileNotFoundException("No such file: " + file);
        }
        JarFile scenario = null;
        try {
            try {
                scenario = new JarFile(file);
            } catch(ZipException ze) {
                throw new InvalidScenarioFileException("Invalid Jar archive", ze);
            }
            if(scenario.getJarEntry(PEF_ENTRY) == null) {
            	isSettingsFile = true;
                //throw new InvalidScenarioFileException("Protocol data missing");
            }
        } finally {
            if(scenario != null) {
                scenario.close();
            }
        }
        this.file = file;
    }
    
    
    /**
     * Saves a global settings file
     * 
     * @param file The settings file
     * @param saveables Collection of settings objects
     * @throws IOException if saving fails due to I/O problems
     * @throws InvalidScenarioFileException if the file is not a valid scenariofile
     */
    public ScenarioFile(File file, Collection saveables) throws IOException, InvalidScenarioFileException {
    	this.file = file;
    	saveFile(null, saveables);
    }
    
    /**
     * Retrieves data from the given collection of {@link Saveable} objects and
     * saves them to this scenario file. Any existing data is overridden.
     * 
     * @param saveables collection of Saveable objects
     * @throws IOException if saving fails due to I/O problems
     */
    public void save(Collection saveables) throws IOException {
        if(saveables == null) {
            throw new IllegalArgumentException("Saveables null");
        }
        Set allNames = new HashSet();
        for(Iterator i = saveables.iterator(); i.hasNext(); ) {
            String type = entryNameForSaveable((Saveable)i.next());
            if(!allNames.add(type)) {
                throw new IllegalArgumentException("Can't save multiple objects of same type");
            }
        }
        JarFile existing = new JarFile(file);
        JarEntry pefEntry = existing.getJarEntry(PEF_ENTRY);
        InputStreamReader input = new InputStreamReader(existing.getInputStream(pefEntry));
        CharArrayWriter pefWriter = new CharArrayWriter((int)pefEntry.getSize());
        copyReaderToWriter(input, pefWriter);
        existing.close();
        char[] pefData = pefWriter.toCharArray();
        saveFile(new CharArrayReader(pefData), saveables);
    }
    
    /**
     * Populates the given collection of {@link Saveable} objects from
     * this scenario file.
     * 
     * @throws IOException if loading fails
     */
    public void load(Collection saveables) throws IOException {
        load(saveables, null, null);
    }
    
    /**
     * Populates the given collection of Saveables and the DataView from
     * 
     * @throws IOException if loading fails
     */
    public void load(Collection saveables, DataView view, ProgressIndicator indicator) throws IOException {
        if(saveables == null) {
            throw new IllegalArgumentException("Saveables null");
        }
        if(view != null) {
        	if (!isSettingsFile)
        		populateView(view, indicator);
            serializer.setDataView(view);
        }
        JarFile existing = new JarFile(file);
        try {
            for(Iterator i = saveables.iterator(); i.hasNext(); ) {
                Saveable saveable = (Saveable)i.next();
                JarEntry entry = existing.getJarEntry(entryNameForSaveable(saveable));
                if(entry != null) {
                    InputStreamReader reader = new InputStreamReader(existing.getInputStream(entry));
                    saveable.setData(serializer.loadObject(reader));
                    reader.close();
                }
            }
        } finally {
            existing.close();
        }
    }

    /**
     * Populates the given {@link DataView} with the protocol data
     * using {@link XMLProtocolEventsReader}.
     * 
     * @param view to be populated
     * @throws IOException if reading fails
     */
    public void populateView(DataView view) throws IOException {
        populateView(view, null);
    }
    
    /**
     * @see #populateView(DataView)
     */
    public void populateView(DataView view, ProgressIndicator indicator) throws IOException {
        if(view == null) {
            throw new IllegalArgumentException("DataView required");
        }
        ProtocolEventsReader reader = new XMLProtocolEventsReader();
        JarFile existing = new JarFile(file);
        try {
            JarEntry pefEntry = existing.getJarEntry(PEF_ENTRY);
            InputStreamReader input = new InputStreamReader(existing.getInputStream(pefEntry));
            reader.read(input, view, indicator);
        } finally {
            existing.close();
        }
    }
    
    /**
     * Returns the actual file that contains the scenario data.
     */
    public File getArchiveFile() {
        return file;
    }

    /**
     * Sets the object serializer to be used when saving objects.
     * The default is {@link XStreamObjectSerializer}.
     */
    public void setObjectSerializer(ObjectSerializer serializer) {
        this.serializer = serializer;
    }
    
    /**
     * Saves the scenario file with the protocol events data and a collection
     * of {@link Saveable saveables}.
     */
    protected void saveFile(Reader protocolDataReader, Collection saveables) throws IOException {
        JarOutputStream out = new JarOutputStream(new FileOutputStream(file));
        OutputStreamWriter writer = new OutputStreamWriter(out);
        try {
            out.setLevel(Deflater.BEST_COMPRESSION);
            if (protocolDataReader != null) {
            	out.putNextEntry(new JarEntry(PEF_ENTRY));
            	copyReaderToWriter(protocolDataReader, writer);
            	out.closeEntry();
            }
            for(Iterator i = saveables.iterator(); i.hasNext(); ) {
                Saveable saveable = (Saveable)i.next();
                out.putNextEntry(new JarEntry(entryNameForSaveable(saveable)));
                serializer.saveObject(saveable.getData(), writer);
                out.closeEntry();
            }
        } finally {
            out.close();
        }
    }
    
    /**
     * Use fully-qualified class name as the key under which each Saveable
     * is saved in the scenario file.
     */
    private String entryNameForSaveable(Saveable saveable) {
        return saveable.getClass().getName();
    }
    
    private void copyReaderToWriter(Reader r, Writer w) throws IOException {
        char[] buffer = new char[2048];
        int read;
        int total = 0;
        while((read = r.read(buffer)) != -1) {
            total += read;
            w.write(buffer, 0, read);
        }
        w.flush();
    }

}
