package fi.helsinki.dacopan.scenario;

import java.io.Reader;
import java.io.Writer;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.converters.Converter;
import com.thoughtworks.xstream.converters.MarshallingContext;
import com.thoughtworks.xstream.converters.UnmarshallingContext;
import com.thoughtworks.xstream.io.HierarchicalStreamReader;
import com.thoughtworks.xstream.io.HierarchicalStreamWriter;

import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Flow;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Identifiable;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.Link;
import fi.helsinki.dacopan.model.Note;
import fi.helsinki.dacopan.model.NoteManager;
import fi.helsinki.dacopan.model.Protocol;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * An {@link ObjectSerializer} implementation based on <a href="http://xstream.codehaus.org">XStream</a>,
 * a library which serializes object graphs as XML.
 * 
 * <p>Contains special support for {@link fi.helsinki.dacopan.model.Identifiable} objects in the model
 * package, they're not serialized as is, instead only their ids are output.</p>
 * 
 * @version $Id: XStreamObjectSerializer.java,v 1.5 2004/05/10 09:55:43 aarniala-dacopan Exp $
 */
public class XStreamObjectSerializer implements ObjectSerializer {
    
    /** used to resolve saved model objects in loadObject() */
    private DataView view;

    public void saveObject(Object o, Writer w) {
        if(o == null || w == null) {
            throw new IllegalArgumentException("Object and writer can't be null");
        }
        xs().toXML(o, w);
    }

    public Object loadObject(Reader r) {
        if(r == null) {
            throw new IllegalArgumentException("Reader can't be null");
        }
        return xs().fromXML(r);
    }
    
    public void setDataView(DataView view) {
        this.view = view;
    }

    /**
     * Returns an XStream instance with custom converters.
     */
    private XStream xs() {
        XStream xs = new XStream();
        xs.registerConverter(new FailWithUnsupportedModelObjectsConverter());
        xs.registerConverter(new IdentifiableConverter());
        return xs;
    }
    
    /**
     * For model objects with an identity, outputs the id only instead of the whole object.
     * When loading objects, gets them back by id from a DataView instance.
     */
    private final class IdentifiableConverter implements Converter {

        public boolean canConvert(Class c) {
            return Identifiable.class.isAssignableFrom(c);
        }

        public void marshal(Object o, HierarchicalStreamWriter w, MarshallingContext c) {
            w.setValue(String.valueOf(((Identifiable)o).getId()));
        }

        public Object unmarshal(HierarchicalStreamReader r, UnmarshallingContext c) {
            if(view == null) {
                throw new RuntimeException(
                        "DataView must be non-null when loading model objects with identity");
            }
            long id = Long.parseLong(r.getValue());
            Class type = c.getRequiredType();
            if(type.equals(Flow.class)) return view.getFlowById(id);
            if(type.equals(Host.class)) return view.getHostById(id);
            if(type.equals(Layer.class)) return view.getLayerById(id);
            if(type.equals(Link.class)) return view.getLinkById(id);
            if(type.equals(Protocol.class)) return view.getProtocolById(id);
            if(type.equals(TransferUnit.class)) return view.getUnitById(id);
            throw new RuntimeException("Unknown Identifiable: " + type.getName());
        }
        
    }

    /**
     * Prevents saving of model objects that have no special support.
     */
    private static final class FailWithUnsupportedModelObjectsConverter implements Converter {
        
        private static final String PKG = "fi.helsinki.dacopan.model";

        public boolean canConvert(Class c) {
            // we're only interested in the model package, fall thru
            if(!c.getName().startsWith(PKG)) {
                return false;
            }
            // some classes can be serialized as is, fall thru
            if(c.equals(NoteManager.class) ||
                    c.equals(Note.class) ||
                    c.equals(VariableDefinition.class) ||
                    c.equals(VariableDefinition.Scope.class)) {
                return false;
            }
            // for others, fail
            return true;
        }

        public void marshal(Object o, HierarchicalStreamWriter w, MarshallingContext c) {
            fail(o.getClass());
        }

        public Object unmarshal(HierarchicalStreamReader r, UnmarshallingContext c) {
            fail(c.getRequiredType());
            return null;
        }
        
        private void fail(Class type) {
            throw new RuntimeException("Saving of model object " + type.getName() + " not implemented");
        }
        
    }
    
}
