package fi.helsinki.dacopan.settings;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.scenario.Saveable;
import fi.helsinki.dacopan.ui.SettingsMSC;
import fi.helsinki.dacopan.ui.tsc.SettingsTSC;


/**
 * Class containing general settings for the animation.
 * <p>
 * The settings stored in this class are basically related to the performance of the animation.
 * These settings are the only ones that are common to all the layers shown in the different
 * animations.
 * <p>
 * Apart from these performance setting, this class has a list that stores the settings specific to
 * layer containing values to configure scale, speed, selected variables and fields to show in the
 * animations...
 * 
 * @version $Id: GeneralSettings.java,v 1.31 2005/05/05 20:32:34 tituomin Exp $
 */
public class GeneralSettings 
    implements Saveable
{

    // default values for the different attributes
    public static final int DEFAULT_REFRESH_DELAY = 25;

    public static final int MIN_REFRESH_DELAY = 5;
    public static final int MAX_REFRESH_DELAY = 100;
    public static final int MIN_REFRESH_RATE = 1000 / MAX_REFRESH_DELAY;
    public static final int MAX_REFRESH_RATE = 1000 / MIN_REFRESH_DELAY;

    /** 
     * number of miliseconds between two updates of the animation. This means
     * that for a large values, the animation will be slower and for small
     * values it will be faster
     * (TimeStampTime)
     */
    protected int refreshDelay = DEFAULT_REFRESH_DELAY;

    
    /**
     * List containing the settings specific to layer (SettingsMSC) 
     */
    protected List settingsLayersMSC;
    
    protected Layer defaultLayer;

    /**
     * Settings for the Time Sequence Chart
     */
    protected SettingsTSC settingsTSC;
    
    /**
     * true if the antialiasing option is selected
     */
    protected boolean antiAliasing;

    
    /** constructor that assign to the attributes default values */
    public GeneralSettings() {
        this.settingsLayersMSC = new LinkedList();
        this.settingsTSC = new SettingsTSC();
        this.refreshDelay = DEFAULT_REFRESH_DELAY;
        this.antiAliasing = true;
    }

    
    /** creates a new copy instance of GeneralSettings*/
    public GeneralSettings(GeneralSettings generalSettings) {
        populateFields(generalSettings);
    }
    
    /**
     * Populates all fields with copies of parameter's fields
     * 
     * @param generalSettings
     */
    private void populateFields(GeneralSettings generalSettings) {
        this.refreshDelay = generalSettings.getRefreshDelay();
        this.settingsLayersMSC = new LinkedList();
        this.antiAliasing = generalSettings.getAntiAliasing();

        for (Iterator i = generalSettings.getListSettingsMSC().iterator();i.hasNext();){
            this.addSettingsMSC(new SettingsMSC((SettingsMSC)i.next()));
        }
        this.defaultLayer = generalSettings.getDefaultLayer();
        this.settingsTSC = new SettingsTSC(generalSettings.getSettingsTSC());
    }
    
    /**
     * returns the number of miliseconds between two updates in the animation panels high values imply
     * slower animation
     * 
     * @return
     *            number of miliseconds between two updates in the animation
     */
    public int getRefreshDelay() {
        return this.refreshDelay;
    }

    /**
     * sets the number of miliseconds between two updates in the animation panels high values imply
     * slower animation
     * 
     * @param refreshDelay
     *            number of miliseconds between two updates in the animation
     */
    public void setRefreshDelay(int refreshDelay) {
        if (refreshDelay < MIN_REFRESH_DELAY
            || refreshDelay > MAX_REFRESH_DELAY) {
            throw new IllegalArgumentException("Refresh Delay out of range!");
        }
        this.refreshDelay = refreshDelay;
    }

    /**
     * returns the number of times the animation panel is updated in a second
     * 
     * @return
     *            number of times that the animation is updated in a second
     */
    public int getRefreshRate() {
        // TODO change casting by function that takes the nearest integer
        return (int) (1000 / this.refreshDelay);
    }

    
    /**
     * sets the number of times the animation panel is updated in a second
     * 
     * @param refreshRate
     *            number of times that the animation is updated in a second
     */
    /* Deprecated method
    public void setRefreshRate(int refreshRate) {
        if (refreshRate < MIN_REFRESH_RATE || refreshRate > MAX_REFRESH_RATE) {
            throw new IllegalArgumentException("Refresh Rate out of range!");
        }
        // TODO change casting by function that takes the nearest integer
        this.refreshDelay = Math.round((float)1000 / (float)this.refreshDelay);
    }
    */

    /**
     * Sets the default layer for the MSC animation
     * 
     * @param layer
     *            Layer to set as default
     */
    public void setDefaultLayer(Layer layer) {
        if(layer == null) {
            throw new IllegalArgumentException("Default layer cannot be null!");
        }
        this.defaultLayer = layer;
    }
    
    /**
     * Gets the default layer for the MSC animation
     * 
     * @return Layer default for the animation
     */
    public Layer getDefaultLayer() {
        return this.defaultLayer;
    }

    /**
     * Adds or replaces a SettingsMSC object in the settingsLayersMSC list. Replace will happen if
     * there already is a SettingsMSC object that has the same layer as the newSettings object.
     * <p>
     * If the default layer for this GeneralSettings object is null, it will be set to newSettings.
     * 
     * @param newSettings
     *            SettingsMSC instance to add to the list of layer settings
     * @return True if existing settings for the layer were found and replaced, false if new
     *         settings for a layer were added.
     */
    public boolean addSettingsMSC(SettingsMSC newSettings) {
    	if (newSettings == null) {
    		throw new IllegalArgumentException("settings cannot be null!");
    	}
    	
    	Iterator iter = settingsLayersMSC.iterator();
    	while (iter.hasNext()) {
			SettingsMSC tempSettings = (SettingsMSC) iter.next();
			Layer layer = tempSettings.getLayer();
			if (layer.equals(newSettings.getLayer()))
			{
				int existingIndex = settingsLayersMSC.indexOf(tempSettings);
				settingsLayersMSC.set(existingIndex, newSettings);
				return true;
			}
		}
    	    	
    	settingsLayersMSC.add(newSettings);
    	
    	if (getDefaultLayer() == null) {
    		setDefaultLayer(newSettings.getLayer());
    	}

    	return false;
    }


    /**
     * Sets the list of SettingsMSC objects. This method
     * is used when applying global settings.
     * 
     * @param listSettingsMSC
     */
    public void setSettingsMSC(List listSettingsMSC) {
    	settingsLayersMSC = listSettingsMSC;
    }
    
    /**
     * Clears the SettingsMSC objects. Needed for
     * global settings.
     */
    public void clearSettingsMSC() {
        this.settingsLayersMSC = new LinkedList();
    }
    
    /**
     * returns the settings for the MSC animation in a provided layer.
     * 
     * @param layer
     *            Layer to get the settings from
     * @return SettingsMSC instance with the settings for the layer specified as parameter
     */
    
    public SettingsMSC getSettingsMSC(Layer layer) {
        if (layer == null) {
            throw new IllegalArgumentException("layer cannot be null!");
        }

        for (Iterator i = settingsLayersMSC.iterator(); i.hasNext();) {
            SettingsMSC tmpSettings = (SettingsMSC) i.next();
            if (tmpSettings.getLayer().equals(layer)) {
                return tmpSettings;
            }
        }

        return null;
    }
    
    /**
     * Return the list of settings per layer
     * 
     * @return List with the list of settings for the different layers in the animation
     */
    public List getListSettingsMSC() {
        return settingsLayersMSC;
    }
    


    /**
     * Returns the anti-aliasing setting
     * 
     * @return boolean
     */
    public boolean getAntiAliasing(){
        return antiAliasing;
    }

    /**
     * Sets the anti-aliasing setting
     * 
     * @param antiAliasing
     *            boolean
     */
    public void setAntiAliasing(boolean antiAliasing){
        this.antiAliasing = antiAliasing;
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return "GeneralSettings { refreshDelay: "
            + getRefreshDelay()
            + " / defaultLayer"
            + getDefaultLayer().getName()
            + " / timeScale"
            + " / antiAliasing"
            + getAntiAliasing()
            + " }";
    }
    
    /* (non-Javadoc)
     * @see fi.helsinki.dacopan.scenario.Saveable#getData()
     */
    public Object getData() {
        return this;
    }
    
    /* (non-Javadoc)
     * @see fi.helsinki.dacopan.scenario.Saveable#setData(java.lang.Object)
     */
    public void setData(Object o) {
        populateFields((GeneralSettings)o);
    }
	/**
	 * @return Returns the settingsTSC.
	 */
	public SettingsTSC getSettingsTSC() {
		return settingsTSC;
	}
	/**
	 * @param settingsTSC The settingsTSC to set.
	 */
	public void setSettingsTSC(SettingsTSC settingsTSC) {
		this.settingsTSC = settingsTSC;
	}
}
