package fi.helsinki.dacopan.settings;

import java.util.HashMap;
import java.util.Map;

import fi.helsinki.dacopan.model.Protocol;
import fi.helsinki.dacopan.model.VariableDefinition;
import fi.helsinki.dacopan.scenario.Saveable;
import fi.helsinki.dacopan.ui.SettingsMSC;

/**
 * Class which contains the global -- not scenario- or dataview specific -- MSC settings
 * which are saved in the global settings file.
 * <p>
 * Contains the SettingsMSC for all layers and
 * a map of chosen variable fields in the MSC animation for each protocol.
 * The id of the protocol is used since the protocol objects cannot be serialized
 * with a dataview in which the do not exist
 * 
 */
public class GlobalSettingsMSC implements Saveable {

	/**
	 * Inner class for the center- and column variable list pairs stored in the map.
	 *
	 */
	public class VariableFields {
		VariableDefinition [] variablesCenter;
		VariableDefinition [] variablesColumns;
		
		/**
		 * Constructor.
		 * 
		 * @param center Variables shown on top of the unit line
		 * @param columns Variables shown in the columns
		 */
		public VariableFields (VariableDefinition[] center, VariableDefinition[] columns) {
			variablesCenter = center;
			variablesColumns = columns;
		}

		// Private class for creating the returned variables for the current dataview
		private VariableDefinition[] getVariables(VariableDefinition[] variables, Protocol protocol) {
			VariableDefinition [] returnVariables = new VariableDefinition [variables.length];
			for (int i = 0; i < variables.length; i++) {
				returnVariables[i] = new VariableDefinition(variables[i].getRawName(),
						variables[i].getScope(),
						protocol);
			}
			return returnVariables;
		}
		
		/**
		 * @param protocol The correct protocol for current dataview is needed,
		 * since it cannot be saved.
		 * @return The center variables
		 */
		public VariableDefinition[] getVariablesCenter(Protocol protocol) {
			return getVariables(variablesCenter, protocol);
		}
		
		/**
		 * @param protocol The correct protocol for current dataview is needed,
		 * since it cannot be saved.
		 * @return The column variables
		 */
		public VariableDefinition[] getVariablesColumns(Protocol protocol) {
			return getVariables(variablesColumns, protocol);
		}
		
		/**
		 * 
		 * Sets the center variables.
		 *
		 * 	@param center Array of displayable VariableDefinitions.
		 */
		public void setVariablesCenter(VariableDefinition[] center) {
			variablesCenter = center;
		}
		
		
		/**
		 * Sets the columb variables
		 * 
		 * 	@param columns Array of displayable VariableDefinitions.
		 */
		public void VariablesColumns(VariableDefinition[] columns) {
			variablesColumns = columns;
		}
		
	}
	
	
	/**
	 * Map of VariableFields, with id of protocol as key
	 */
	private Map mapSettingsMSC;
	/**
	 * Global MSC settings
	 */
	private SettingsMSC layerSettingsMSC;
	
	/**
	 * Constructor creating an empy object
	 */
	public GlobalSettingsMSC() {
		this.mapSettingsMSC = new HashMap();
		this.layerSettingsMSC = null;
	}
	
	/**
	 * Inserts the center- and column- variables for
	 * the protocol.
	 * 
	 * @param protocolId the id of the protocol, used instead of Protocol, because it's
	 * dataview-independent when serializing.
	 * @param variablesCenter the center variables
	 * @param variablesColumns the column variables
	 */
	public void putVariableFields(
				long protocolId,
				VariableDefinition[] variablesCenter,
				VariableDefinition[] variablesColumns) {

		mapSettingsMSC.put(
			(Object)(new Long(protocolId)),
			(Object)(new VariableFields(variablesCenter, variablesColumns)));
	}

	/**
	 * Gets the variableFields for the wanted protocol
	 * 
	 * @param protocolId id of the protocol
	 * @return
	 */
	public VariableFields getVariableFields(long protocolId) {
		return (VariableFields)mapSettingsMSC.get((Object)(new Long(protocolId)));
	}
	
	/**
	 * Stores the global MSC settings, other than the variable fields.
	 * 
	 * @param newSettingsMSC The new settings
	 */
	public void setSettingsMSC(SettingsMSC newSettingsMSC) {
		this.layerSettingsMSC = newSettingsMSC;
	}
	
	/**
	 * Gets the global MSC settings
	 * 
	 * @return The global SettingsMSC object
	 */
	public SettingsMSC getSettingsMSC() {
		return this.layerSettingsMSC;
	}
	
	/* (non-Javadoc)
	 * @see fi.helsinki.dacopan.scenario.Saveable#getData()
	 */
	public Object getData() {
		return this;
	}
	
	/**
	 * Sets all fields by copying them from the parameter.
	 * 
	 * @param newSettings The source of the new fields.
	 */
	private void populateFields(GlobalSettingsMSC newSettings) {
		mapSettingsMSC = new HashMap(newSettings.mapSettingsMSC);
		layerSettingsMSC = new SettingsMSC(newSettings.layerSettingsMSC);
	}
	
	/* (non-Javadoc)
	 * @see fi.helsinki.dacopan.scenario.Saveable#setData(java.lang.Object)
	 */
	public void setData(Object data) {
		populateFields((GlobalSettingsMSC)data);
	}

}
