package fi.helsinki.dacopan.ui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.font.TextLayout;
import java.util.Iterator;
import java.util.List;

import javax.swing.JPanel;

import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * This panel is used to draw the actual animation inside the UFO panel.
 */
public class ChannelPanel extends JPanel {

    private final static int PADDING = 5;
    private final static Color CHANNEL_COLOR_ARROW = new Color(0xDDDDDD);
    private final static float CHANNEL_PROPORTION = 0.3f;
    private final static int MARGIN = 5;
    private final static int MARGIN_TOP = 25;
    private final static Font CHANNEL_TITLE_FONT = new Font("Verdana",Font.BOLD,12);
    private final static Font VARIABLE_FIELD_FONT = new Font("Arial",Font.PLAIN,12);
    private static final int TITLE_AREA_HEIGHT = 20; // in pixels
    private static final float TITLE_AREA_MARGIN_RATIO = 0.2f; // ratio    
    private static final int HOST_AREA_WIDTH = 20; // in pixels    
    private static final BasicStroke HOST_AREA_LINE_STROKE =
        new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, new float[] {10.0f}, 0.0f);
    private static final Color HOST_AREA_LINE_COLOR = Color.GRAY;
        
    public final static int LEFT_TO_RIGHT = 0;
    public final static int RIGHT_TO_LEFT = 1;
    
    private final static int MIN_VARIABLE_FIELD_HEIGHT = 150;

    private static final int MAX_DRAWABLE_VALUE_LENGTH = 150;
    
    // The hosts
    private Host leftHost;
    private Host rightHost;
    
    // The units to be drawn
    private List unitsInFlight;
    
    private float currentTime;
    private boolean antialias;

    private Rectangle defaultClipRectangle;
    
    // The selected transfer units
    private TransferUnit currentlyActive[];
    
    private int currentY;
    private CalcYCoord coord;

    /**
     * 
     * @param hosts
     */
    public ChannelPanel(List hosts) {
        leftHost = (Host)hosts.get(0);
        rightHost = (Host)hosts.get(1);
        currentlyActive = new TransferUnit[2];
        currentlyActive[0] = null;
        currentlyActive[1] = null;
        antialias = true;
        
        defaultClipRectangle = 
            new Rectangle(0,0,getWidth(),getHeight());
    }

    /**
     * Draws the units moving in the channels
     * 
     * @param g
     */
    public void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        
        g2.setClip(0,0,getWidth(),getHeight());
        g2.clearRect(0, 0, getWidth(), getHeight());
        
        // Draw the two channels
        drawChannels(g2);
        
        // Draw the two variable fields
        drawVariableFields(g2);
        
        // Remove clipping rectangle
        g2.setClip(null);
    }
    
    /**
     * Tells the panel what it should draw
     * 
     * @param   units       The units that are currently on the now line
     * @param   currentTime Now time
     */
    public void drawUnits(List units, float currentTime, boolean antialias, CalcYCoord coordinates, boolean isInPlayMode) {
        unitsInFlight = units;

        coord = coordinates;
        
        if (coord != null) {
	        if (isInPlayMode) {
	        	currentY = coord.getLinearYCoordForTime(currentTime);
	        } else {
	        	currentY = coord.getYCoordForTime(currentTime);
	        }
        } else {
        	currentY = 0;
        }
        // Check the currently active units 
        findCurrentlyActive(unitsInFlight, rightHost, LEFT_TO_RIGHT, getWidth() - MARGIN - 10 -1);
        findCurrentlyActive(unitsInFlight, leftHost, RIGHT_TO_LEFT, getWidth() - MARGIN - 10 -1);
        
        this.currentTime = currentTime;
        this.antialias = antialias;
        
        repaint();
    }
    
    /**
     * Returns the currently active transfer unit.
     * 
     * @param direction
     * @return The currently active unit
     */
    public TransferUnit getCurrentlyActive(int direction)
    {
        return currentlyActive[direction];
    }
    
    /**
     * Draws the channels in their right sizes.
     * 
     * @param g
     * @param panelBounds
     */
    private void drawChannels(Graphics2D g2)
    {        
        int channelWidth = getWidth() - (MARGIN + 10);
        int availableChannelSpace = calculateVerticalChannelSpace();
        
        // Calculate the height for one channel
        int channelHeight = (availableChannelSpace - PADDING) / 2;
                    
        // First draw the host names
        g2.setFont(CHANNEL_TITLE_FONT);
        
        if(leftHost.getHostname() != null) {
            g2.drawString(leftHost.getHostname(), MARGIN, MARGIN + 10);
        }
        
        // Calculate the length of the host name
        if(rightHost.getHostname() != null) {        
            TextLayout textLayout = new TextLayout(rightHost.getHostname(), 
                g2.getFont(), g2.getFontRenderContext());
            g2.drawString(rightHost.getHostname(), 
                (int)(getWidth() - (MARGIN + textLayout.getBounds().getWidth())), 
                MARGIN + 10);
        }
        
        // First channel         
        int leftX = MARGIN;
        int topY = 25;
        drawChannel(g2, leftX, topY, channelWidth, channelHeight, LEFT_TO_RIGHT);
        drawUnitsInChannel(g2, rightHost, LEFT_TO_RIGHT, leftX + 1, topY + 1, channelWidth - 1, channelHeight - 1);
        
        // Second channel
        leftX = MARGIN;
        topY += channelHeight + 5;
        drawChannel(g2, leftX, topY, channelWidth, channelHeight, RIGHT_TO_LEFT);
        drawUnitsInChannel(g2, leftHost, RIGHT_TO_LEFT, leftX + 1, topY + 1, channelWidth - 1, channelHeight - 1);
    }

    /**
     * Draws the variable fields in right place and with the right size
     * 
     * @param g
     * @param panelBounds
     */
    private void drawVariableFields(Graphics2D g2)
    {    	
        g2.setFont(VARIABLE_FIELD_FONT);
        
        int fieldWidth = (getWidth() - 4 * MARGIN) / 2;
        int fieldHeight = calculateVerticalFieldSpace();
        
        int topY = getHeight() - (fieldHeight + PADDING);
        
        // First variable field
        int leftX = MARGIN;
        drawVariableField(g2, LEFT_TO_RIGHT, leftX, topY, fieldWidth, fieldHeight);
       
        // Second variable field
        leftX += fieldWidth + MARGIN;
        drawVariableField(g2, RIGHT_TO_LEFT, leftX, topY, fieldWidth, fieldHeight);
    }
    
    /**
     * Draws an arrow with the given direction and coordinates
     * 
     * @param g
     * @param x
     * @param y
     * @param width
     * @param height
     * @param direction
     */
    private void drawArrow(Graphics2D g2, int x, int y, 
			int width, int height, int direction) 
    {
        Rectangle arrow = new Rectangle();
        Polygon arrowPolygon = new Polygon();
        Rectangle arrowRectangle = new Rectangle(); 
        
        // And then create the actual arrow shape, i.e. calculate 
        // how the head and the "line" should be drawn.
        
        int arrowHeadWidth = (int)(width * 0.2f);
                
        arrowRectangle.height = (int)(height * 0.5f);
        arrowRectangle.y = (int)(y + 
            (height - arrowRectangle.height) / 2);
        arrowRectangle.width = width - arrowHeadWidth;
        
        // The direction the arrow is pointing to affects the way the 
        // arrow is drawn (on which side the head is)                   
        if(direction == LEFT_TO_RIGHT) {
            arrowRectangle.x = x;
            
            // top-left of the arrow head
            arrowPolygon.addPoint(arrowRectangle.x + arrowRectangle.width, y);
            // bottom-left of the arrow head
            arrowPolygon.addPoint(arrowRectangle.x + arrowRectangle.width, y + height);                               
            // head
            arrowPolygon.addPoint(arrowRectangle.x + width, y + height / 2);

        }
        else{
            arrowRectangle.x = x + arrowHeadWidth;

            // top-right of the arrow head
            arrowPolygon.addPoint(x + arrowHeadWidth, y);
            // bottom-right of the arrow head
            arrowPolygon.addPoint(x + arrowHeadWidth, y + height);                               
            // head
            arrowPolygon.addPoint(x, y + height / 2);
        }

        g2.setColor(CHANNEL_COLOR_ARROW);
        g2.fill(arrowRectangle);
        g2.fill(arrowPolygon);    

    }
    
    
    /**
     * Draws a channel with the given coordinates. The direction tells which way 
     * the arrow on the channel should be pointing.
     * 
     * @param g
     * @param x
     * @param y
     * @param width
     * @param height
     * @param direction
     */
    private void drawChannel(Graphics2D g2, int x, int y, 
                             int width, int height, int direction)
    {
        if(antialias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }
        else {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);            
        }

        // Draw a shadow
        g2.setColor(Color.LIGHT_GRAY);
        g2.fillRect(x+3, y+3, width, height);
        
        // Draw the box
        g2.setColor(Color.WHITE);
        g2.fillRect(x, y, width, height);
        
        g2.setColor(Color.BLACK);
        g2.drawRect(x, y, width, height);
        
        // inside the box draw a "host" area for showing parts of 
        // packets that still are waiting to be sent
        Stroke oldStroke = g2.getStroke();
        Color oldColor = g2.getColor();
        g2.setStroke(HOST_AREA_LINE_STROKE);
        g2.setColor(HOST_AREA_LINE_COLOR);
        
        g2.drawLine(x + HOST_AREA_WIDTH, y, x + HOST_AREA_WIDTH, y + height);
        g2.drawLine(x + width - HOST_AREA_WIDTH, y, x + width - HOST_AREA_WIDTH, y + height);
        
        g2.setColor(oldColor);
        g2.setStroke(oldStroke);
        
        // And in the middle, draw the arrow pointing to the correct direction
        int arrowWidth = (int)(width * 0.5f);
        int arrowHeight = (int)(height * 0.8f);
        int arrowX = x + (width - arrowWidth) / 2;
        int arrowY = y + (height - arrowHeight) / 2;
        drawArrow(g2, arrowX, arrowY, arrowWidth, arrowHeight, direction);
    }

    /**
     * Draws a variable field with the given parameters.
     * 
     * @param g
     * @param x
     * @param y
     * @param width
     * @param height
     */
    private void drawVariableField(Graphics2D g2, int direction, 
    		int x, int y, int width, int height) 
    {    	
    	// Draw the shadow
    	g2.setColor(Color.LIGHT_GRAY);
    	g2.fillRoundRect(x+3,y+3,width,height,10,10);
    	    	
    	// Draw the rectangle
        g2.setColor(Color.WHITE);
    	g2.fillRoundRect(x,y,width,height,10,10);
       
    	g2.setColor(Color.BLACK);
    	g2.drawRoundRect(x,y,width,height,10,10);

        // Draw the title bar
        drawVariableHeading(g2, x, y, width, TITLE_AREA_HEIGHT, direction);
        
    	// Draw the variables
        Shape oldClip = g2.getClip();
        
    	g2.setClip(x,y,width,height);
    	g2.setColor(Color.BLACK);
        
    	TransferUnit unit = currentlyActive[direction];
        if(unit != null) {
            List variables = currentlyActive[direction].getProtocol().getVariables();
            Iterator i = variables.iterator();
            int lineHeight = g2.getFontMetrics().getHeight();
            
            x += MARGIN;
            y += TITLE_AREA_HEIGHT + lineHeight;
            
            while(i.hasNext()) {
                VariableDefinition var = (VariableDefinition)i.next();
                String value = unit.getValue(var);
                if(value != null) {
                    if (value.length() > MAX_DRAWABLE_VALUE_LENGTH) {
                        value = value.substring(0, MAX_DRAWABLE_VALUE_LENGTH);
                    }

                    g2.drawString(var.getName() +": "+ value, x, y);
                    y += lineHeight;
                }
            }
        }
        g2.setClip(oldClip);
    }
    
    
    private void drawVariableHeading(Graphics2D g2, int x, int y, 
            int width, int height, int direction)
    {
    	// Draw a line to separate this part from the variables
        g2.setColor(Color.BLACK);
        g2.drawLine(x, y + height, x + width, y + height);
        
        // Calculate the correct place for the arrow.
        int marginX = (int)(width * TITLE_AREA_MARGIN_RATIO);
        int marginY = (int)(height * TITLE_AREA_MARGIN_RATIO);
        
        int arrowX = x + marginX;
        int arrowY = y + marginY;
        
        drawArrow(g2, arrowX, arrowY, width-(2 * marginX), height-(2 * marginY), direction);
    }
    
    
    /**
     * Calculates the vertical space that will be allocated to the 
     * channels.
     * 
     * @param panelBounds
     * @return
     */
    private int calculateVerticalChannelSpace()
    {
        int availableChannelSpace = (int)(getHeight() * CHANNEL_PROPORTION);
        int maxChannelSpace = getHeight() - (MARGIN_TOP + PADDING + MIN_VARIABLE_FIELD_HEIGHT);
        if(availableChannelSpace > maxChannelSpace) {
            availableChannelSpace = maxChannelSpace;
        }
        
        return availableChannelSpace;
    }
    
    /**
     * Calculates the vertical space that is going to be allocated to the 
     * variable fields.
     * 
     * @param panelBounds
     * @return
     */
    private int calculateVerticalFieldSpace() {        
        int channelSpace  = calculateVerticalChannelSpace();
        int fieldHeight = getHeight() - (channelSpace + MARGIN_TOP + PADDING * 2);
        
        return fieldHeight;        
    }

    private void drawDroppedSign(Graphics2D g2, int x, int y, 
            int width, int height)
    {
        g2.setColor(Color.RED);
        Stroke currentStroke = g2.getStroke();
        g2.setStroke(new BasicStroke(2.0f));
        g2.drawLine(x-4, y, x + width + 4, y + height);
        g2.drawLine(x-4, y + height, x + width + 4, y); //To get wider X than packet.
        g2.setStroke(currentStroke);
        
    }
    
    /**
     * Draws the units on the channel going to the given destination
     *
     * The coordinates should be considered without the borders.
     *
     * @param g
     * @param destination
     * @param direction
     * @param x
     * @param y
     * @param width
     * @param height
     */
    private void drawUnitsInChannel(Graphics2D g2, Host destination, int direction,
    int x, int y, int width, int height) {
        
        // Set the clipping rectangle so that we can draw outside the channel
        Shape oldClip = g2.getClip();
        g2.setClip(x, y, width, height);
        
        if(unitsInFlight != null) {
            // Remove the host areas from the width
            width -= HOST_AREA_WIDTH * 2;
            
            Iterator i = unitsInFlight.iterator();
            while(i.hasNext()) {
                TransferUnit unit = (TransferUnit)i.next();

                if (getYCoordForTime(unit.getSendStart()) <= currentY
                		&& getYCoordForTime(unit.getReceiveEnd()) >= currentY
						&& !unit.isDropped()) {
	                // If the unit is going to the destination we want draw it
	                if(unit.getDestination() == destination) {
	                    
	                    
	                    // Calculate position in the channel
	                    float xPosition = 0f;
	                    
	                    // For larger packets use a more complex approach
	                    if(unit.getChildren().size() > 1) {
	                        drawLargeUnit(g2, unit, currentTime, x, y, width, height, direction);
	                    }
	                    else {
	                        drawSingleUnit(g2, unit, currentTime, x, y, width, height, direction);
	                    }
	                }
                } else if (unit.isDropped() && getYCoordForTime(unit.getSendStart()) <= currentY 
                		&& getYCoordForTime(unit.getSendStart()) + width / 2 >= currentY
						&& unit.getDestination() == destination) {
                 	drawSingleUnit(g2, unit, currentTime, x, y, width, height, direction);
                }
            }
        }
        
        // Remove the clipping rectangle
        g2.setClip(oldClip);
        
    }

    private void drawSingleUnit(Graphics2D g2, TransferUnit unit,
        float currentTime, int x, int y, int width, int height, int direction) 
    {
        int size = UFOPanel.DACOPAN_UFO_UNIT_SIZE;
        float xPosition = 0f;
        float unitSent = unit.getSendStart();
        float unitReceived = unit.getReceiveEnd();

        if(unit.isDropped()) {
            unitReceived += (unitReceived - unitSent);
        }
        
        //xPosition = (currentTime - unitSent) * (width + size) / (unitReceived - unitSent);
        int sendY = getYCoordForTime(unit.getSendStart());
        int receiveY = getYCoordForTime(unitReceived);
        
        if (unit.isDropped()) {
        	receiveY = sendY + width;
        }
        if (sendY < receiveY) {
        	xPosition = (int)((currentY - sendY) * width) / (receiveY - sendY);
        } else {
        	xPosition = width / 2;
        }
        // Draw the unit
        if(direction == LEFT_TO_RIGHT) {
            xPosition -= size;
            g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_1);
            g2.fillRect((int)(x + xPosition + HOST_AREA_WIDTH), y, size, height);
            
            if(unit.isDropped()) {
                drawDroppedSign(g2, x + (int)xPosition + HOST_AREA_WIDTH, y, size, height);
            }
            
            // if the unit is active, draw a border around it
            if(unit == currentlyActive[direction]) {
                Stroke oldStroke = g2.getStroke();
                g2.setStroke(new BasicStroke(1.0f));
                //g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_1.darker());
                g2.setColor(Color.RED);
                g2.drawRect((int)(x + xPosition + HOST_AREA_WIDTH), y, size, height);
                g2.setStroke(oldStroke);
            }
            else {
                g2.setColor(Color.BLACK);
                g2.drawRect((int)(x + xPosition + HOST_AREA_WIDTH), y, size, height);
            }
        }
        else {
            xPosition = (HOST_AREA_WIDTH + width) - xPosition;
            g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_2);
            g2.fillRect((int)(x + xPosition), y, size, height);
            
            if(unit.isDropped()) {
                drawDroppedSign(g2, x + (int)xPosition, y, size, height);
            }
            
            // if the unit is active, draw a border around it
            if(unit == currentlyActive[direction]) {
                Stroke oldStroke = g2.getStroke();
                g2.setStroke(new BasicStroke(1.0f));
                //g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_2.darker());
                g2.setColor(Color.RED);
                g2.drawRect((int)(x + xPosition), y, size, height);
                g2.setStroke(oldStroke);
            }
            else {
                g2.setColor(Color.BLACK);
                g2.drawRect((int)(x + xPosition), y, size, height);
            }
        }
        
    }


    /**
     * @param unit
     * @param size
     * @return
     */
    private void drawLargeUnit(Graphics2D g2, TransferUnit unit,
        float currentTime, int x, int y, int width, int height, int direction) 
    {
        float startX;
        float endX;
        
        startX = width * (currentTime - unit.getSendStart()) /
        (unit.getReceiveStart() - unit.getSendStart());
        
        endX = width * (currentTime - unit.getSendEnd()) /
        (unit.getReceiveEnd() - unit.getSendEnd());
        
        // Draw the unit
        if(direction == LEFT_TO_RIGHT) {
            float newStartX = HOST_AREA_WIDTH + endX;
            endX = HOST_AREA_WIDTH + startX;
            startX = newStartX;
            g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_1);
        }
        else {
            startX = (HOST_AREA_WIDTH + width) - startX;
            endX = (HOST_AREA_WIDTH + width) - endX;
            g2.setColor(UFOPanel.DACOPAN_UFO_UNIT_COLOR_2);
        }
        
        // Do the drawing
        
        g2.fillRect((int)(x + startX), y, (int)(endX - startX), height);
        
        if(unit.isDropped()) {
            drawDroppedSign(g2, x + (int)startX, y, (int)(endX - startX), height);
        }
        
        // if the unit is active, draw a border around it
        if(unit == currentlyActive[direction]) {
            Stroke oldStroke = g2.getStroke();
            g2.setStroke(new BasicStroke(4.0f));
            g2.setColor(g2.getColor().darker());
            g2.drawRect((int)(x + startX), y, (int)(endX - startX), height);
            g2.setStroke(oldStroke);
        }
        else {
            g2.setColor(Color.BLACK);
            g2.drawRect((int)(x + startX), y, (int)(endX - startX), height);
        }
    }

    private void findCurrentlyActive(List units, Host destination, int direction, int width) {
    	
        // Go through all units and select the unit with the 
        // latest sendstart time. If the user has selected some
        // unit to be the active one, let it be until it goes
        // out of the screen. 
        
        int unitsInChannel = 0;
        currentlyActive[direction] = null;
        
        if(units != null)
        {                      
            if(!units.isEmpty()) {
                Iterator i = units.iterator();
                while(i.hasNext()) {
                    TransferUnit unit = (TransferUnit)i.next();
                    
                    if(unit.getDestination() == destination) {
                        
                        unitsInChannel++;
                                                
                        if (getYCoordForTime(unit.getSendStart()) <= currentY 
                        		&& ((!unit.isDropped() 
                        		        && getYCoordForTime(unit.getReceiveEnd()) >= currentY)
                        		|| (unit.isDropped() 
                        		        && getYCoordForTime(unit.getSendStart()) + width / 2 >= currentY))) {
                        	unitsInChannel++;
                        	if(currentlyActive[direction] == null) {
	                            currentlyActive[direction] = unit;
	                        }
	                    	else if(unit.getSendStart() > currentlyActive[direction].getSendStart())
	                        {
	                            currentlyActive[direction] = unit;
	                        }
                        } 
                    }
                }        
            }
        }
        
        // If there is no unit going this way, clear selection
        if(unitsInChannel == 0) {
            currentlyActive[direction] = null;
        }
    }
    
    private int getYCoordForTime(float time) {
    	if (coord == null) {
    		return 0;
    	} else {
    		return coord.getYCoordForTime(time);
    	}
    }
}
