package fi.helsinki.dacopan.ui;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.GeneralPath;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JPanel;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.ENCTreeModel;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * The ENCPanel is what displays the encapsulation animation of a particular
 * selected TransferUnit. It shows a picture of the layers and the various units
 * that can be directly linked to the selected unit through the encapsulation tree.
 * See also ENCTreeModel for a look at the internal representation of the
 * encapsulation tree and it's construction.
 * 
 * TODO if active unit also has more to left/right, it looks messed up
 * 
 * @version $Id: ENCPanel.java,v 1.25 2005/05/04 11:58:45 tituomin Exp $
 * @see fi.helsinki.dacopan.model.ENCTreeModel
 */
public class ENCPanel extends AbstractAnimationPanel {

    /****  CONSTANTS ****/
    
    /** margin on all sides of the enc tree (px) */
    static final int PANEL_MARGIN = 15;
    
    /** horizontal gap between units (px) */
    static final int UNIT_HGAP = 30;
    
    /** minimum width of a unit (px) */
    static final int UNIT_WIDTH_MIN = 100;

    /** minimum height of a unit (px) */
    static final int UNIT_HEIGHT_MIN = 40;
    
    /** ratio of the payload width to unit width */
    static final float RATIO_PAYLOAD_TO_UNIT = 0.15f;

    /** top and bottom margin of a layer (px) */
    static final int LAYER_MARGIN = 25;

    /** minimum height of a layer (px) */
    static final int LAYER_HEIGHT_MIN = UNIT_HEIGHT_MIN + 2 * LAYER_MARGIN;

    /** width of the symbol area on each side (px) */
    static final int SYMBOL_AREA_WIDTH = 20;
    
    /** vertical gap between unit variables (px) */
    static final int VARIABLE_VGAP = -3;
    
    /** left padding of variable values (px) */
    static final int VARIABLE_PADDING_LEFT = 5;
    
    /** top padding of variable values (px) */
    static final int VARIABLE_PADDING_TOP = 2;
    
    /** background gradient from -color */
    static final Color BACKGROUND_GRADIENT_FROM = new Color(180, 180, 180);
    
    /** background gradient to -color */
    static final Color BACKGROUND_GRADIENT_TO = Color.WHITE;
    
    /** default paint color */
    static final Color PAINT_COLOR = Color.BLACK;
    
    /** color of the border of the active unit */
    static final Color BORDER_ACTIVE_COLOR = Color.RED;
    
    /** color of the cross drawn over dropped units */
    static final Color DROPPED_UNIT_COLOR = new Color(255, 0, 0, 100);
    
    /** background color of the payload */
    static final Color PAYLOAD_BACKGROUND_COLOR = new Color(140, 140, 140);
    
    /** unit background color */
    static final Color UNIT_BACKGROUND_COLOR = Color.WHITE;
    
    /** number of edges to draw to the symbol indicating a partial unit */
    static final int PARTIAL_UNIT_EDGES = 4;

    /** width of an edge (percentage of symbol area width) */
    static final float PARTIAL_UNIT_EDGE_WIDTH = 0.75f;
    
    /** transparency value for sibling units partially drawn on screen [0.0..1.0] */
    static final float SIBLING_OPACITY = 0.3f;
    
    /** stroke style for the line that separates layers */
    static final BasicStroke LAYER_SEPARATOR_STROKE =
        new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, new float[] {10.0f}, 0.0f);

    /** stroke style for the lines that connect units */
    static final BasicStroke UNIT_CONNECTION_STROKE =
        new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 2.0f, new float[] {2.0f}, 0.0f);
    
    /** stroke style used to highlight the active unit */
    static final BasicStroke ACTIVE_UNIT_STROKE = new BasicStroke(3.0f);

    /** stroke style used to paint a cross over dropped units */
    static final BasicStroke DROPPED_UNIT_STROKE = new BasicStroke(3.0f);
    
    // Instance variables
    
    /** the source of all the goodies */
    private ENCTreeModel etm;

    // Global widgets
    private GridBagLayout gridBag;
    private TitlePanel encTitle;
       
    /**
     * Encapsulation panels are constructed with a reference to the
     * mainframe, and an encapsulation model object.
     * 
     * @param mainFrame
     * @param tree
     */
    public ENCPanel(MainFrame mainFrame, ENCTreeModel encTreeModel) {
        super(mainFrame);
        this.etm = encTreeModel;
        setLayout(new BorderLayout());
        addWidgets();
    }
    
    /**
     * This method adds all widgets to the main panel
     */
    private void addWidgets() {
        // add all widgets
        encTitle = new TitlePanel(Localization.getString("panel.enc.title"), null); //$NON-NLS-1$
        add(encTitle, BorderLayout.NORTH);

        EncDrawingPanel drawingPanel = new EncDrawingPanel();
        add(drawingPanel, BorderLayout.CENTER);
        
        JPanel southPanel = new JPanel(new GridBagLayout());
        southPanel.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
        GridBagConstraints c = new GridBagConstraints();
        
        JButton quitEncButton = new JButton(Localization.getString("panel.enc.quit")); //$NON-NLS-1$
        UserInterface.setDefaultButtonFont(quitEncButton);        
        quitEncButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                mainFrame.continueAnimationModeMSC();
            }
        });
        
        c.gridx = 0;
        c.gridy = 0;
        c.insets = new Insets(5,5,5,5);       
        southPanel.add(quitEncButton, c);

        add(southPanel, BorderLayout.SOUTH);
        
    }
    
    

    public void stepTo(float nowTime) { }
    
    //inner class EncDrawingPanel
    
    public class EncDrawingPanel extends JPanel {
        
        /**
         * Paints the whole ENC panel.
         */
        protected void paintComponent(Graphics g) {
            
            Graphics2D g2 = (Graphics2D)g;
            
            // since ENC panel is not animated, let's get fancy and use anti-aliasing
            if(mainFrame.getSettings().getAntiAliasing()) {
            	g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            }
            else {
                g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);                
            }

            // paint the background of the whole panel with a gradient
            Paint originalPaint = g2.getPaint();
            GradientPaint gradient =
                new GradientPaint(0, 0, BACKGROUND_GRADIENT_FROM,
                                  0, getBounds().height, BACKGROUND_GRADIENT_TO);
            g2.setPaint(gradient);
            g2.fillRect(0, 0, getWidth(), getHeight());
            g2.setPaint(originalPaint);
            
            Stroke regular = g2.getStroke();
            
            g2.setColor(Color.BLACK);
            
            int drawableHeight = getHeight();
            int drawableWidth = getWidth() - PANEL_MARGIN * 2;
            
            ENCTreeModel.DrawableEncLayer[] layers = etm.getLayeredEncTreeToDraw();
            
            // make sure layers aren't shrunk too much
            int layerHeight = drawableHeight / layers.length;
            if(layerHeight < LAYER_HEIGHT_MIN) {
                layerHeight = LAYER_HEIGHT_MIN;
            }
            
            /*
             * When lower level units capture payload from higher level units,
             * we need to visualize these relations by drawing lines between
             * the units. This map stores the coordinates of the parent unit
             * to which the lines should be drawn from the children. For each
             * child unit X, the coordinates are available under the key
             * X.id (long) as a int[] array with the length of 4 (x1, x2, y1, y2).
             */
            Map unitCoords = new HashMap();

            int currentLayerTop = 0;
            for (int i = layers.length - 1; i >= 0; i--) {
                // calculate the area each layer should be painted on
                Rectangle layerArea = new Rectangle(
                    PANEL_MARGIN,
                    currentLayerTop + LAYER_MARGIN,
                    drawableWidth,
                    layerHeight - 2 * LAYER_MARGIN
                );
                paintLayer(g2, layers[i], layerArea, unitCoords);
                currentLayerTop += layerHeight;
                if(i > 0) {
                    // if there are more layers, draw the layer separator line
                    g2.setStroke(LAYER_SEPARATOR_STROKE);
                    g2.drawLine(0, currentLayerTop - 1, getWidth(), currentLayerTop - 1);
                    g2.setStroke(regular);
                }
            }
            
        }
        
        /**
         * Paints a single layer.
         */
        private void paintLayer(Graphics2D g2, ENCTreeModel.DrawableEncLayer encLayer, Rectangle area, Map unitCoords) {

            if(encLayer.getDrawableUnits().isEmpty()) {
                // TODO handle empty layers separately?
                return;
            }

            Stroke regular = g2.getStroke();
            
            // calculate width of each unit
            int numUnits = encLayer.getDrawableUnits().size();
            int unitWidth = (area.width - ((numUnits - 1) * UNIT_HGAP) - 2 * SYMBOL_AREA_WIDTH) / numUnits;
            if(unitWidth < UNIT_WIDTH_MIN) {
                unitWidth = UNIT_WIDTH_MIN;
            }
            UnitDimensions d = new UnitDimensions(0, area.y, unitWidth, area.height);

            // draw each unit on the layer starting from the leftmost
            int unitX = area.x + SYMBOL_AREA_WIDTH;
            for(Iterator it = encLayer.getDrawableUnits().iterator(); it.hasNext(); ) {

                ENCTreeModel.DrawableTransferUnit unit = (ENCTreeModel.DrawableTransferUnit)it.next();
                d.x = unitX;

                // if we have visible children, count the coordinates they should draw lines to
                if(unit.getVisibleChildIds().size() > 0) {
                    int childWidth = unitWidth / unit.getVisibleChildIds().size();
                    int xCoord = d.x;
                    int yCoord = d.y + d.height;
                    for (Iterator iter = unit.getVisibleChildIds().iterator(); iter.hasNext();) {
                        int[] coords = new int[] { xCoord, yCoord, xCoord + childWidth, yCoord };
                        // store the coordinates in the map under the id of the child
                        Long childId = (Long) iter.next();
                        unitCoords.put(childId, coords);
                        xCoord += childWidth;
                    }
                }
                // if this unit has a parent, draw connections to it using the
                // pre-calculated coordinates of the parent unit
                if(unit.getTransferUnit().getParent() != null) {
                    long myId = unit.getTransferUnit().getId();
                    int[] coords = (int[])unitCoords.get(new Long(myId));
                    g2.setStroke(UNIT_CONNECTION_STROKE);
                    g2.drawLine(d.x, d.y, coords[0], coords[1]);
                    g2.drawLine(d.x + d.payloadWidth, d.y, coords[2], coords[3]);
                    g2.setStroke(regular);
                }
                
                paintUnit(g2, unit, d, 1.0f);
                
                unitX += unitWidth + (it.hasNext() ? UNIT_HGAP : 0);

            }

            // if partial unit with more to left, draw a symbol indicating this
            if(encLayer.getFirstUnit().hasMoreToTheLeft()) {
                paintPartialUnitEdges(g2, d, true);
            }

            // if partial unit with more to right, draw a symbol indicating this
            if(encLayer.getLastUnit().hasMoreToTheRight()) {
                paintPartialUnitEdges(g2, d, false);
            }

            // draw dots representing left siblings, if any
            if(encLayer.hasSiblingsToTheLeft()) {
                paintSiblings(g2, d, true);
            }

            // draw dots representing right siblings, if any
            if(encLayer.hasSiblingsToTheRight()) {
                paintSiblings(g2, d, false);
            }
            
        }
        
    }
    
    /**
     * Paints a single unit to the specified coordinates.
     * 
     * @param unit to paint
     * @param d dimensions of the unit
     * @param opacity transparency value (1.0f = fully visible, 0.0f = invisible)
     */
    private void paintUnit(Graphics2D g2,
            ENCTreeModel.DrawableTransferUnit unit, UnitDimensions d, float opacity) {
        
        Color payloadColor = PAYLOAD_BACKGROUND_COLOR;
        Color unitColor = UNIT_BACKGROUND_COLOR;
        Color borderColor = PAINT_COLOR;
        
        // if transparency specified, create colors with alpha
        if(opacity < 1.0f) {
            payloadColor = copyColorWithAlpha(payloadColor, opacity);
            unitColor = copyColorWithAlpha(unitColor, opacity);
            borderColor = copyColorWithAlpha(borderColor, opacity);
        }
        
        Stroke regular = g2.getStroke();
        
        // payload rectangle
        g2.setColor(payloadColor);
        g2.fillRect(d.x, d.y, d.payloadWidth, d.height);
        g2.setColor(borderColor);
        g2.drawRect(d.x, d.y, d.payloadWidth, d.height);
        
        // protocol values rectangle
        g2.setColor(unitColor);
        g2.fillRect(d.x + d.payloadWidth, d.y, d.width - d.payloadWidth, d.height);
        g2.setColor(borderColor);
        Rectangle variableRect = new Rectangle(d.x + d.payloadWidth, d.y, d.width - d.payloadWidth, d.height);
        g2.draw(variableRect);
        
        // if this unit is selected, highlight it with a thicker border
        if(etm.getSelectedUnit().equals(unit.getTransferUnit())) {
            Rectangle wholeUnitRect = new Rectangle(d.x, d.y, d.width, d.height); 
            g2.setStroke(ACTIVE_UNIT_STROKE);
            g2.setColor(BORDER_ACTIVE_COLOR);
            g2.draw(wholeUnitRect);
            g2.setStroke(regular);
            g2.setColor(PAINT_COLOR);
        }
        
        // add the correct padding for the variables and paint them, if not a dummy unit
        if(unit.getTransferUnit() != null) {
            variableRect.x += VARIABLE_PADDING_LEFT;
            variableRect.y += VARIABLE_PADDING_TOP;
            variableRect.width -= VARIABLE_PADDING_LEFT;
            variableRect.height -= VARIABLE_PADDING_TOP;
            paintUnitVariables(g2, unit, variableRect);
            // if dropped, make sure everybody notices this
            if(unit.getTransferUnit().isDropped()) {
                g2.setColor(DROPPED_UNIT_COLOR);
                g2.setStroke(DROPPED_UNIT_STROKE);
                // draw \
                g2.drawLine(d.x + 1, d.y + 1, d.x + d.width - 2, d.y + d.height - 2);
                // draw /
                g2.drawLine(d.x + d.width - 2, d.y + 1, d.x + 1, d.y + d.height - 2);
                g2.setStroke(regular);
            }
        }
        
        
        g2.setColor(PAINT_COLOR);
    }
    
    /**
     * Creates a copy of the color with the given transparency value.
     */
    private Color copyColorWithAlpha(Color color, float alpha) {
        return new Color(color.getRed(), color.getGreen(), color.getBlue(), (int)(alpha * 255));
    }

    /**
     * Paints the symbols indicating that there are more siblings.
     * 
     * @param g2 graphics object
     * @param d unit dimensions
     * @param directionLeft if true, draw to left-hand size
     */
    private void paintSiblings(Graphics2D g2, UnitDimensions d, boolean directionLeft) {
        // a dummy unit, depends on paintUnit having a null check for the unit
        ENCTreeModel.DrawableTransferUnit unit = new ENCTreeModel.DrawableTransferUnit(null);
        int x = directionLeft ?
                PANEL_MARGIN - d.width :
                d.x + d.width + SYMBOL_AREA_WIDTH;
        UnitDimensions dd = new UnitDimensions(x, d.y, d.width, d.height);
        paintUnit(g2, unit, dd, SIBLING_OPACITY);
    }
    
    /**
     * Paints the edges indicating a partial unit.
     * 
     * @param d dimensions of the unit
     * @param directionLeft direction of edges is to the left
     */
    private void paintPartialUnitEdges(Graphics2D g2, UnitDimensions d, boolean directionLeft) {
        int edgeWidth = (int)(SYMBOL_AREA_WIDTH * PARTIAL_UNIT_EDGE_WIDTH);
        if(!directionLeft) {
            edgeWidth *= -1;
        }
        int edgeHeight = d.height / (PARTIAL_UNIT_EDGES * 2);
        int currY = d.y;
        int baseX = (directionLeft ? d.x + 1 : d.x + d.width);
        GeneralPath partial = new GeneralPath(GeneralPath.WIND_EVEN_ODD);
        partial.moveTo(baseX, currY);
        for(int i = 0; i < PARTIAL_UNIT_EDGES; i++) {
            boolean lastEdge = (i + 1 == PARTIAL_UNIT_EDGES);
            currY += edgeHeight;
            partial.lineTo(baseX - edgeWidth, currY);
            if(lastEdge) {
                currY = d.y + d.height + 1; // make sure lowest edge matches unit's bottom
            } else {
                currY += edgeHeight + 1;
            }
            partial.lineTo(baseX + (lastEdge ? 0 : (directionLeft ? -1 : 1 )), currY);
        }
        g2.setColor(directionLeft ? PAYLOAD_BACKGROUND_COLOR : UNIT_BACKGROUND_COLOR);
        g2.fill(partial);
        g2.setColor(PAINT_COLOR);
    }

    /**
     * Paints the variables of a unit.
     * 
     * @param g2 graphics object
     * @param unit to be drawn
     * @param area where to draw the unit
     */
    private void paintUnitVariables(Graphics2D g2, ENCTreeModel.DrawableTransferUnit unit, Rectangle area) {
        Shape oldClip = g2.getClip();
        g2.setClip(area);
        TransferUnit tu = unit.getTransferUnit();
        Font oldFont = g2.getFont();
        Font variableFont = oldFont;
        Font unitHeaderFont = variableFont.deriveFont(Font.BOLD);
        g2.setFont(unitHeaderFont);
        int lineHeight = g2.getFontMetrics().getHeight();
        int lineBase = area.y + lineHeight;
        g2.drawString(tu.getProtocol().getName() + " " + tu.getId(), area.x, lineBase); //$NON-NLS-1$
        lineBase += lineHeight + VARIABLE_VGAP;
        g2.setFont(variableFont);
        lineHeight = g2.getFontMetrics().getHeight();
        for (Iterator it = tu.getProtocol().getVariables().iterator(); it.hasNext();) {
            VariableDefinition def = (VariableDefinition) it.next();
            String value = tu.getValue(def);
            if(value != null && value.length() > 0) {
                g2.drawString(def.getName() + ": " + value, area.x, lineBase); //$NON-NLS-1$
                lineBase += lineHeight + VARIABLE_VGAP;
            }
        }
        g2.setClip(oldClip);
        g2.setFont(oldFont);
    }
    
    /**
     * For wrapping the dimensions of a unit. Makes method signatures look nice.
     */
    private static final class UnitDimensions {
        int x, y, width, height, payloadWidth;
        private UnitDimensions(int x, int y, int width, int height) {
            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
            this.payloadWidth = (int)(RATIO_PAYLOAD_TO_UNIT * width);
        }
    }
    
}
