/*
 * MSCPanel.java
 *
 * Created on 29. huhtikuuta 2004, 16:24
 */

package fi.helsinki.dacopan.ui;


import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelListener;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.JPanel;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Note;
import fi.helsinki.dacopan.model.StaticVariable;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;
import fi.helsinki.dacopan.ui.MSCPanel.MSCColumnModel.MSCColumn;

/**
 * MSCPanel draws the MSC animation type. As it exteds AbstractAnimationPanel, its most important methods
 * are those inherited from AbstractAnimationPanel, i.e. <code>stepTo()</code> and <code>advance()</code>.
 * 
 */

/* (non-javadoc)
 * The implementation of MSCPanel is quite tricky because of AnimationTimeState logic that is not designed for
 * this use. Completely rewrite of MSCPanel and AnimationTimeState is recomended.
 */
public class MSCPanel extends fi.helsinki.dacopan.ui.AbstractAnimationPanel {
    
    private static final int MSC_VERTICAL_MARGIN = 16; //16;
    private static final int MSC_SCROLLBAR_MULTIPLIER = 1; // 10000;
    private static final int MSC_SCROLLBAR_UNIT_INCREMENT_RATIO = 10; // 10;
    
    private static final int WIDTH_COLUMN_NOTES = 15;
    private static final int WIDTH_COLUMN_PER_CHARACTER = 8;
    private static final int WIDTH_DRAWING_AREA_MINIMUM = 100;
    
    private static final int MAX_DRAWABLE_VALUE_LENGTH = 150;
    
    /** Disables using antialias for drawing 'fat' units (fat is defined as unit.isOneUnit()
     * returning false).*/
    private static final boolean DISABLE_ANTIALIAS_FOR_FAT_UNITS = true;
    
    /** Uses optimized drawing in play mode (advanceMode == true). The optimized drawing means
     * caching image data in a buffer and then drawing slices on screen as the now line progresses.
     * The buffer is only refreshed when the screen needs to 'nudge'.
     * 
     * <p>Note that currently the optimization scheme is rather simple. The buffer data is
     * refreshed completely, instead of utilizing the 2/3 of it that would be usable. Also,
     * the buffer might also be used to optimize drawing in scrolling mode. However, these
     * ideas are not currently implemented.
     *
     * <p>Note also that the optimized drawing changes slightly the way in which column values
     * and notes appear on screen. While stepping, the values for current event appear even if
     * they have to be drawn below the now line. This is not true in optimized play mode, because
     * the image is simply cropped at now line. However, as the now line is moving, noticing
     * the difference in play mode would normally be extremely difficult.
     */
    private static final boolean USE_OPTIMIZED_PLAYMODE_DRAWING = true;
        
    private static final Color COLOR_DROPPED_CROSS = Color.red;
    private static final int DROPPED_CROSS_WIDTH = 16;
    
    private static final Color COLOR_COLUMN_BORDERS = Color.gray;
    private static final Color COLOR_NOWLINE = Color.blue;
    private static final Color COLOR_COLUMN_VALUES = Color.black;
    private static final Color COLOR_EVENT_LINES = new Color(0x33, 0x33, 0x33);
    private static final Color COLOR_EVENT_AREA = new Color(204, 204, 75);
    private static final Color COLOR_EVENT_VALUES = Color.black;
    
    /** The margin left between the line representing a unit on MSC, and the corresponding
     * text presenting the header field values for that unit. This is in absolute pixels on
     * absolute, regardless of the angle of the text.
     */
    private static final int EVENT_VALUE_YDIFF = 3;
    
    private static final String NOTE_ICON_PATH = "/icon_note.png";
    
    private static final int MSC_ARROW_HEIGHT = 5;
    private static final int MSC_ARROW_LENGTH = 5;
    
    private static final int DROP_POINT = 24; 

    private BufferedImage noteImage = null;
    {
        try {
            noteImage = ImageIO.read(getClass().getResourceAsStream(NOTE_ICON_PATH));
        } catch (IOException e) {
            System.out.println("MSC Panel trying to load note icon:");
            System.out.println(e);
            System.out.flush();
        }
    }
    
    private final SettingsMSC settings;
    private final DataView dataView;
    
    private final MSCPanel.MSCDrawingPanel drawingPanel;
    
    private final boolean drawBelowNowLine = false;
    
    private final boolean useAntiAlias;
    
    private boolean advanceMode = false;
    private boolean isScrolling = false;
    private boolean isSteppingBackwards = false;
    private boolean isStepping = true;
    private boolean forcedRedraw = false;
    private boolean useRealTime = true;
    
    private MainFrame mainFrame;
    
    /* true -> false: when jScrollBar1.setValue is called
     * false -> true: jScrollBar1AdjustmentValueChanged method beginning, and
     * return from the method.
     *
     * effect: when the value is set programmatically, the first call to
     * jScrollBar1AdjustmentValueChanged will not be interpreted as user action
     */
    private boolean scrollBarValueSet = false;
    
    private float timeNow;
    
    private MSCColumnModel columnModel;
    
    private CalcYCoord coordinates = null;
    
    private int fontSize;
    
    private boolean printTimestamp; 
    
    private int verticalMargin;

    /** Creates new form MSCPanel */
    public MSCPanel(
        MainFrame mainFrame, 
        SettingsMSC settings, 
        DataView dataView,
        float nowTime) 
    {
        super(mainFrame);
        
        this.mainFrame = mainFrame;
        
        if (settings == null) throw new IllegalArgumentException("Settings can't be null!");
        this.settings = settings;
        
        if (dataView == null) throw new IllegalArgumentException("DataView can't be null!");
        this.dataView = dataView;

        if (nowTime < 0f) throw new IllegalArgumentException("nowTime can't be negative!");
        this.timeNow = nowTime;
        
        useAntiAlias = mainFrame.getSettings().getAntiAliasing();
        
        drawingPanel = new MSCDrawingPanel();
        
        columnModel = new MSCColumnModel(getDrawableWidth(), settings.getVariablesColumns());

        List units = dataView.getUnitsForLayer(settings.getLayer());
        coordinates = new CalcYCoord(dataView, settings);

        fontSize = settings.getFontSize(); 
        
        printTimestamp = settings.showTimeStamps();
        
        verticalMargin = Math.max(MSC_VERTICAL_MARGIN, 
        		settings.getVariablesColumns().length 
					* (int)(new Font("verdana", Font.PLAIN, fontSize)).getMaxCharBounds(
        				new FontRenderContext(null, false, false)).getHeight());
        
        initComponents();

        jPanel2.setBorder(BorderFactory.createLineBorder(Color.gray));
        
        jScrollBar1.setMinimum(-verticalMargin);

        jScrollBar1.setMaximum(coordinates.getLastYCoord() - coordinates.getFirstYCoord() - drawingPanel.getDrawableHeight() 
        		+ verticalMargin);
        
        drawingPanel.addMouseWheelListener(drawingPanel);
        jScrollBar1.addMouseWheelListener(drawingPanel);
        
        // set to true to prevent scrollbar from yanking to top
        scrollBarValueSet = true;
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     * 
     * DaCoPAn2 note: Contents have been changed from what they were. No NOT
     * use old .form-files or Form Editor to change layout!
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        java.awt.GridBagConstraints gridBagConstraints;

        jPanel1 = new javax.swing.JPanel();
        jPanelTitle = new TitlePanel(Localization.getString("panel.msc.title"), settings.getLayer().getName());
        jPanelHosts = new javax.swing.JPanel();
        jLabelHostLeft = new javax.swing.JLabel();
        jLabelHostRight = new javax.swing.JLabel();
        jTableHostVariablesLeft = new javax.swing.JTable();
        jTableHostVariablesRight1 = new javax.swing.JTable();
        jPanel2 = new javax.swing.JPanel();
        jPanelColumnHeadingPanel = new ColumnHeadingPanel();
        jPanelDrawingPanel = drawingPanel;
        jScrollBar1 = new javax.swing.JScrollBar();

        setLayout(new java.awt.BorderLayout());

        jPanel1.setLayout(new java.awt.GridBagLayout());

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        jPanel1.add(jPanelTitle, gridBagConstraints);

        jPanelHosts.setLayout(new BorderLayout());

        jLabelHostLeft.setText(Localization.getString("panel.msc.host_a")+settings.getLeftHost().getIp());
        jPanelHosts.add(jLabelHostLeft, BorderLayout.WEST);

        jLabelHostRight.setText(Localization.getString("panel.msc.host_b")+settings.getRightHost().getIp());
        jPanelHosts.add(jLabelHostRight, BorderLayout.EAST);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.insets = new java.awt.Insets(2, 2, 2, 2);
        jPanel1.add(jPanelHosts, gridBagConstraints);

        jPanel2.setLayout(new java.awt.GridBagLayout());

        jPanelColumnHeadingPanel.setMinimumSize(new java.awt.Dimension(100, 20));
        jPanelColumnHeadingPanel.setPreferredSize(new java.awt.Dimension(100, 20));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.SOUTH;
        gridBagConstraints.weightx = 1.0;
        jPanel2.add(jPanelColumnHeadingPanel, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        jPanel2.add(jPanelDrawingPanel, gridBagConstraints);

        jScrollBar1.addAdjustmentListener(new java.awt.event.AdjustmentListener() {
            public void adjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {
                jScrollBar1AdjustmentValueChanged(evt);
            }
        });

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.VERTICAL;
        gridBagConstraints.weighty = 1.0;
        jPanel2.add(jScrollBar1, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new java.awt.Insets(4, 4, 4, 4);
        jPanel1.add(jPanel2, gridBagConstraints);

        add(jPanel1, java.awt.BorderLayout.CENTER);

    }//GEN-END:initComponents

    private void jScrollBar1AdjustmentValueChanged(java.awt.event.AdjustmentEvent evt) {//GEN-FIRST:event_jScrollBar1AdjustmentValueChanged
        if (scrollBarValueSet) {
            scrollBarValueSet = false;
            return;
        }
        
        if (advanceMode) {
            drawingPanel.setScrollBarValue();
        } else {
            isScrolling = true;
            drawingPanel.repaint();
        }
    }//GEN-LAST:event_jScrollBar1AdjustmentValueChanged
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel jLabelHostLeft;
    private javax.swing.JLabel jLabelHostRight;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanel2;
    private javax.swing.JPanel jPanelColumnHeadingPanel;
    private javax.swing.JPanel jPanelDrawingPanel;
    private javax.swing.JPanel jPanelHosts;
    private javax.swing.JPanel jPanelTitle;
    private javax.swing.JScrollBar jScrollBar1;
    private javax.swing.JTable jTableHostVariablesLeft;
    private javax.swing.JTable jTableHostVariablesRight1;
    // End of variables declaration//GEN-END:variables
    
    
    private final int getDrawableWidth() {
        return drawingPanel.getWidth() - drawingPanel.getInsets().left
            - drawingPanel.getInsets().right;
    }
    
    public void stepTo(float nowTime) {
    	isStepping = true;
        advanceMode = false;
        isScrolling = false;
        if (this.timeNow > nowTime) isSteppingBackwards = true;
        else isSteppingBackwards = false;
        this.timeNow = nowTime;
        drawingPanel.repaint();
        forcedRedraw = false;
    }
    
    public void advance(float step, float nowTime) {
        isScrolling = false;
        isSteppingBackwards = false;
        isStepping = false;
        
        if ((!advanceMode) || (!USE_OPTIMIZED_PLAYMODE_DRAWING) ||
        		drawingPanel.yDrawingBottomChanges(nowTime) ||
        		(!drawingPanel.isBufferedImageSizeValid()) ||
				forcedRedraw) 
        {
            advanceMode = true;
            this.timeNow = nowTime;
            drawingPanel.repaint();
            forcedRedraw = false;
        } else {
            drawingPanel.advanceDrawingTo(this.timeNow, nowTime);
            this.timeNow = nowTime;
        }
        
    }
    
    public void toPlayMode() {
    	if (useRealTime) {
    		float nowTime;
    		if (drawingPanel != null && coordinates != null)
    			nowTime = drawingPanel.getAllwaysLinearTimeForYCoord(coordinates.getYCoordForTime(timeNow));
    		else
    			nowTime = 0f;
    		mainFrame.getAnimationTimeState().setNowTime(nowTime);
    		forcedRedraw = true;
            useRealTime = false;
    	}
        advanceMode = true;
        isScrolling = false;
        isSteppingBackwards = false;
        isStepping = false;

    }
    
    public void toPauseMode() {
        if (advanceMode) {
            advanceMode = false;
            drawingPanel.setScrollBarValue();
        }
        if (!useRealTime) {
        	useRealTime = true;
	        float nowTime;
	        if (coordinates != null) {
	        	nowTime = coordinates.getTimeForYCoord(coordinates.getLinearYCoordForTime(timeNow));
	        } else {
	        	nowTime = 0.0f;
	        }
	        mainFrame.getAnimationTimeState().setNowTime(nowTime);
	        forcedRedraw = true;
        }
    }
    
    public MSCColumnModel getColumnModel() {
        if (getDrawableWidth() != columnModel.getTotalWidth()) {
            columnModel = new MSCColumnModel(getDrawableWidth(), settings.getVariablesColumns());
            return columnModel;
        } else {
            return columnModel;
        }
    }

    /**
     * Returns animators "play"-time for given animation time (aka the real time).
     * Uses getAnimatorTime(float time) from MSCDrawingPanel.
     * @see MSCDrawingPanel
     * @param events time
     * @return animator time for given event time
     */
    
    public float getAnimatorTime(float eventTime) {
    	if (drawingPanel != null)
    		return drawingPanel.getAnimatorTime(eventTime);
    	else
    		return 0.0f;
    }
    
    /**
     * Returns the real time (aka the animation time) that the animator is currently playing.
     * Uses getRealTime() from MSCDrawingPanel.
     * @see MSCDrawingPanel
     * @return real time used by animator just right now
     */
    public float getRealTime() {
    	
    	if (drawingPanel != null)
    		return drawingPanel.getRealTime();
    	else
    		return 0.0f;
    }
    
private class MSCDrawingPanel extends JPanel implements MouseWheelListener {

    private float timeDrawingTop;
    private float timeDrawingBottom;

    /** The previously drawn time interval measured as the interval length on time axis from
     * the top of the drawing area to the bottom of the drawing area, regardless of the location
     * of the now-line.
     */
    private float prevDrawInterval;
    private float currentDrawInterval;
    
    private int previousDrawableWidth;
    private int previousDrawableHeight;

    private BufferedImage bufferedImage = null;
    
    private boolean isMouseDragging = false;
    
    /* Used to prevent handling events in a wrong order. */
    private boolean isMouseDraggingFinished = false;
    
    /* Used to prevent stupid behaviour if user touches >1 mouse buttons during drag.
     * The drag ends exactly when the button that was pressed first is released. */
    private int mouseDraggingButton = MouseEvent.NOBUTTON;

    private int paintCount = 0; // used only for debugging output
    
    
    private int yDrawingTop;
    private int yDrawingBottom;
    private int prevDrawingYInterval;
    private int currentDrawingYInterval;

    // These are used to get good angle for dropped units. It preserves last non-dropped units angle.
    private double lastGoodLeftAngle = 0;
    private double lastGoodRightAngle = 0;
    

    public MSCDrawingPanel() {
        this.timeDrawingTop = timeNow;
        this.timeDrawingBottom = timeNow + ((float)getDrawableHeight() / settings.getVisualScale());
        this.prevDrawInterval = (float)getDrawableHeight() / settings.getVisualScale();
        setOpaque(true);
        this.setBackground(Color.WHITE);
        this.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent evt) {
                if (!advanceMode && (mouseDraggingButton == MouseEvent.NOBUTTON)) {
                    mouseDraggingButton = evt.getButton();
                    isMouseDragging = true;
                    isMouseDraggingFinished = false;
                    doMouseSelecting(evt.getY());
                }
            }
            
            public void mouseReleased(MouseEvent evt) {
                if (evt.getButton() == mouseDraggingButton) {
                    isMouseDraggingFinished = true;
                    mouseDraggingButton = MouseEvent.NOBUTTON;
                }
            }
        });
        this.addMouseMotionListener(new MouseMotionAdapter() {
            public void mouseDragged(MouseEvent evt) {
                if (!advanceMode && isMouseDragging && !isMouseDraggingFinished) {
                    doMouseSelecting(evt.getY());
                }
            }
        });
        
        if (coordinates != null) {
        	this.yDrawingTop = getLinearYCoordForTime(timeNow);
        	this.yDrawingBottom = getLinearYCoordForTime(timeNow) + getDrawableHeight();
        	this.prevDrawingYInterval = getDrawableHeight();
        } else {
        	this.yDrawingTop = 0;
        	this.yDrawingBottom = getDrawableHeight();
        	this.prevDrawingYInterval = getDrawableHeight();
        }
    }

    /**
     * Return the animator "play" time for given animation (aka real) time.
	 * @param events time
	 * @return animator time for given event time
	 */
	public float getAnimatorTime(float eventTime) {
		return getAllwaysLinearTimeForYCoord(getYCoordForTime(eventTime));
	}

	/**
	 * Returns the "real" time, aka the time in animation.
	 * @return real time used by animator just now
	 */
	public float getRealTime() {
		if (useRealTime) {
			return timeNow;
		} else {
			return getTimeForYCoord(getAllwaysLinearYCoordForTime(timeNow));
		}
	}

	protected final int getNewScrollBarValue() {
        int newValue;
        if (yDrawingTop < -verticalMargin) {
            newValue = -verticalMargin;
        } else if (advanceMode){
			newValue = yDrawingTop; // TODO why?
        } else {
        	newValue = yDrawingTop; 
        }

        return newValue;
    }
    
    protected final void setScrollBarValue() {
    	if (isScrolling) return;

        int newValue = getNewScrollBarValue();
        
        scrollBarValueSet = true;
        if (newValue != jScrollBar1.getValue()) {
            jScrollBar1.setValue(newValue);
        }
        scrollBarValueSet = false;
    }

    private final int getDrawableHeight() {
        return getHeight() - getInsets().top - getInsets().bottom;
    }
    
    /** Returns the new y-coordinate representing the bottom of drawing area. Method is parameterized
     * to take the conceptual 'now time' as the parameter newTime, but otherwise the method relies
     * in some instance variables, such as advanceMode, isScrolling, currentDrawInterval and
     * prevDrawInterval. Method doesn't have any side-effects.
     * 
     * @param newTime
     * @return newYCoordinate
     */
    
    
    private int getYDrawingBottom(float newTime) {
        int newYDrawingBottom = yDrawingBottom;
        int timePlace = getLinearYCoordForTime(newTime);
        
         if (isMouseDragging && !isMouseDraggingFinished) {
            // do nothing, we don't want to move display while dragging
         	if (timePlace > yDrawingBottom) {
         		newYDrawingBottom = timePlace;
         	} else if (timePlace < yDrawingTop) {
         		newYDrawingBottom = timePlace + currentDrawingYInterval;
         	}
        } else if (prevDrawingYInterval != currentDrawingYInterval) {
            // resize has happened
            if (timePlace > yDrawingTop + currentDrawingYInterval - verticalMargin) {
                newYDrawingBottom = timePlace + verticalMargin;
            } else {
                newYDrawingBottom = yDrawingTop + currentDrawingYInterval - verticalMargin;
            }
        } else if (advanceMode) {
            // we are in advance (play) mode
            if (timePlace > yDrawingBottom) {
                // is step the size of one 'nudge'?
                if (timePlace < yDrawingBottom + (currentDrawingYInterval / 3)) {
                    // TODO implement sliding
                    newYDrawingBottom += currentDrawingYInterval / 3;
                } else {
                    newYDrawingBottom = timePlace + (currentDrawingYInterval / 3);
                }
            } else if (timePlace < yDrawingTop) {
                // trying to play while now line is way above display, yank display to it
                newYDrawingBottom = timePlace + (currentDrawingYInterval / 3);
            } else {
                // do nothing
            }
        } else if (isScrolling) {
        	newYDrawingBottom = jScrollBar1.getValue() + currentDrawingYInterval;
        } else { // no advance mode, no resize, just stepping
        	if (timePlace < yDrawingTop) {
        		// We're above window
        		if (isSteppingBackwards) {
        			newYDrawingBottom = Math.max(
        					(int)(timePlace + 2.0f  / 3.0f * currentDrawingYInterval - verticalMargin),
							currentDrawingYInterval - verticalMargin);
        		} else {
        			newYDrawingBottom = Math.max(timePlace + currentDrawingYInterval - verticalMargin,
        					currentDrawingYInterval - verticalMargin);
        		}
        	} else {
        		// We're below the top of the window
        		if (timePlace < yDrawingBottom) {
        			// We're within the window
        			// do nothing
        		}
        		else {
        			// We're below the bottom of the window
        			newYDrawingBottom = (int)(timePlace + 1.0f / 3.0f * currentDrawingYInterval 
        					- verticalMargin);
        		}
        	}
        }
        
        int lastY = coordinates.getLastYCoord();
        if (lastY > yDrawingTop + currentDrawingYInterval - (2*verticalMargin) &&
            newYDrawingBottom > lastY + verticalMargin) 
        {
            newYDrawingBottom = lastY + verticalMargin;
        }
        
        return newYDrawingBottom;
    }
    
    
    protected final boolean yDrawingBottomChanges(float newTime) {
    	return yDrawingBottom != getYDrawingBottom(newTime);
    }
    
    
    protected final boolean isBufferedImageSizeValid() {
        if ((bufferedImage == null) || 
            (bufferedImage.getWidth() != getDrawableWidth()) ||
            (bufferedImage.getHeight() != getDrawableHeight()))
        {
            return false;
        } else {
            return true;
        }
    }
    
    protected void updateYDrawingTopBottomAndInterval(float nowTime) {
    	currentDrawingYInterval = getDrawableHeight();
    	yDrawingBottom = getYDrawingBottom(nowTime);
    	yDrawingTop = yDrawingBottom - currentDrawingYInterval;
    }
    

    /** Gets new scrollbar extent based on currentDrawInterval. */
    protected int getNewScrollBarExtent() {
        int newExtent;

        if (coordinates.getLastYCoord() < getDrawableHeight()) {
        	newExtent = jScrollBar1.getMaximum();
        } else {
        	newExtent = Math.round((currentDrawingYInterval * jScrollBar1.getMaximum()) / 
				(float)(coordinates.getLastYCoord() - coordinates.getFirstYCoord() + verticalMargin));
        }
        
        
        return newExtent;
    }
    
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);

        if (getDrawableWidth() == 0 || getDrawableHeight() == 0) return;

        Graphics2D g2Component = (Graphics2D)g;
        Graphics2D g2Buffer = null;
        
        if (advanceMode && USE_OPTIMIZED_PLAYMODE_DRAWING) {
            if (bufferedImage == null || !isBufferedImageSizeValid())
            {
                bufferedImage = g2Component.getDeviceConfiguration()
                    .createCompatibleImage(getDrawableWidth(), getDrawableHeight());
            }
        
            Graphics bufG = bufferedImage.getGraphics();
            bufG.setColor(this.getBackground());
            bufG.fillRect(0, 0, bufferedImage.getWidth(), bufferedImage.getHeight());
            
            g2Buffer = (Graphics2D)bufferedImage.getGraphics();
        }

        final MSCColumnModel columnModel = getColumnModel();

        updateYDrawingTopBottomAndInterval(timeNow);
        
        // test if area has changed
        if (prevDrawingYInterval != currentDrawingYInterval) {
        	// come out of scrolling
        	isScrolling = false;
        	isSteppingBackwards = false;
        	
        	// change scrollbar extent property
        	int newExtent = getNewScrollBarExtent();
            
            if (jScrollBar1.getValue() + newExtent > jScrollBar1.getMaximum()) {
                jScrollBar1.setValue(jScrollBar1.getMaximum() - newExtent);
            }
            jScrollBar1.setVisibleAmount(newExtent);
            
            jScrollBar1.setUnitIncrement(jScrollBar1.getVisibleAmount() / 
                MSC_SCROLLBAR_UNIT_INCREMENT_RATIO);
            jScrollBar1.setBlockIncrement(jScrollBar1.getVisibleAmount());
        }
        
        int bottomCropY;
        int yNow = getLinearYCoordForTime(timeNow);
        if (drawBelowNowLine) {
        	bottomCropY = yDrawingBottom;
        } else if (isScrolling) {
        	if (yNow > yDrawingBottom) {
        		bottomCropY = yDrawingBottom;
        	} else if (yNow >= yDrawingTop && yNow <= yDrawingBottom) {
        		bottomCropY = yNow;
        	} else {
        		paintColumnBorders(g2Component, columnModel);
        		return;
        	}
        } else {
        	bottomCropY = yNow;
        }
    
        
        if (advanceMode && USE_OPTIMIZED_PLAYMODE_DRAWING) {
            paintColumnBorders(g2Buffer, columnModel);
            
            paintAllEventData(g2Buffer, yDrawingTop, yDrawingBottom, yDrawingBottom, 
            		columnModel);
            
            Shape oldClip = g2Component.getClip();
            g2Component.setClip(0, 0, getDrawableWidth(), bottomCropY - yDrawingTop);
            g2Component.drawImage(bufferedImage, null, 0, 0);
            g2Component.setClip(oldClip);

            paintColumnBorders(g2Component, columnModel);

            paintNowLine(g2Component, timeNow);
        } else {
            paintColumnBorders(g2Component, columnModel);
            paintAllEventData(g2Component, yDrawingTop, bottomCropY, yDrawingBottom, 
            		columnModel);
            paintNowLine(g2Component, timeNow);
        }

        setScrollBarValue();

        // DEBUGGING STUFF - just uncomment what you need
        //System.out.println("-----------------------------------------");
        //System.out.println(paintCount+" advanceMode: "+advanceMode);
        //System.out.println(paintCount+" isScrolling: "+isScrolling);
        //System.out.println(paintCount+" isSteppingBackwards: "+isSteppingBackwards);
        //System.out.println(paintCount+" scrollBarValueSet: "+scrollBarValueSet);
        //System.out.println(paintCount+" isMouseDragging: "+isMouseDragging);
        //System.out.println(paintCount+" isMouseDraggingFinished: "+isMouseDraggingFinished);
        //System.out.println(paintCount+" currentDrawInterval: "+currentDrawInterval);
        //System.out.println(paintCount+" timeDrawingTop: "+timeDrawingTop);
        //System.out.println(paintCount+" timeDrawingBottom: "+timeDrawingBottom);
        //System.out.println(paintCount+" timeNow: "+timeNow);
        //System.out.println(paintCount+" bottomCropTime: "+bottomCropTime);
        //System.out.println(paintCount+" scrollbar value: "+jScrollBar1.getValue());
        //System.out.println(paintCount+" scrollbar extent: "+jScrollBar1.getVisibleAmount());
        //System.out.flush();

        if (isMouseDraggingFinished) {
            isMouseDraggingFinished = false;
            isMouseDragging = false;
        }
        
        previousDrawableWidth = getDrawableWidth();
        previousDrawableHeight = getDrawableHeight();
        paintCount++;
        
        prevDrawingYInterval = currentDrawingYInterval;
    }
    

    private final void paintAllEventData(
            Graphics2D g2, 
            int topY, 
            int cropY,
            int bottomY,
            MSCColumnModel columnModel) 
        {
            // if nothing to draw, just return (otherwise would crash...)
            if (topY > cropY) return;
            
            // TODO a good point for optimizing code. Also dropped packets disappear if this is used as is.
            float topTime = getTimeForYCoord(topY); 
            float cropTime = getTimeForYCoord(cropY);
            

            List units = dataView.getUnitsForLayer(settings.getLayer(), topTime, cropTime);        

            Rectangle clipRectDA = new Rectangle(columnModel.getDrawingArea().getLeftEdge(), 0,
                columnModel.getDrawingArea().getWidth(), cropY - yDrawingTop);
            paintDrawingArea(g2, units, clipRectDA);

            Rectangle clipRectNC = new Rectangle(columnModel.getNotesCol().getLeftEdge(), 0, 
                columnModel.getNotesCol().getWidth(), cropY - yDrawingTop);
            Collection notes = mainFrame.getNoteManager()
                .getNotesInTimeRange(settings.getLayer(), topTime, cropTime);
            paintNotesColumn(g2, columnModel, notes, clipRectNC);

            g2.setColor(COLOR_COLUMN_VALUES);

            paintColumnValues(g2, units, columnModel, bottomY - 
            		yDrawingTop, cropY - yDrawingTop);
            
        }
    
    protected final void advanceDrawingTo(float oldTime, float newTime) {
        Graphics2D g2 = (Graphics2D)getGraphics();
        if (g2 == null) return;
        int rightEdge = getDrawableWidth()-1;
        int sliceTop = getLinearYCoordForTime(oldTime) - yDrawingTop;
        int sliceBottom = getLinearYCoordForTime(newTime) - yDrawingTop;

        g2.drawImage(bufferedImage, 
            0, sliceTop, rightEdge, sliceBottom,
            0, sliceTop, rightEdge, sliceBottom,
            null
        );
        
        paintNowLine(g2, newTime);
    }
    
    private final void paintNowLine(Graphics2D g2, float time) {
        g2.setColor(COLOR_NOWLINE);
        int y = getLinearYCoordForTime(time) - yDrawingTop;
        g2.drawLine(getInsets().left, y, getWidth()-getInsets().right-2, y);
    }

    private final void paintDrawingArea(Graphics2D g2, List units, Rectangle clipRect) {
        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }

        Shape oldClip = null;
        if (!drawBelowNowLine) {
            oldClip = g2.getClip();
            g2.setClip(clipRect);
        }

        Stroke oldStroke = g2.getStroke();
        g2.setStroke(new BasicStroke(2.0f));
        paintEventLines(g2, units, clipRect);
        g2.setStroke(oldStroke);

        g2.setColor(COLOR_EVENT_VALUES);
        paintEventValues(g2, units, clipRect);

        if (!drawBelowNowLine) {
            g2.setClip(oldClip);
        }

        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }
    }

    private final void paintEventLines(Graphics2D g2, List units, Rectangle area) {
        Iterator iter = units.iterator();
        while (iter.hasNext()) {
            TransferUnit element = (TransferUnit)iter.next();
            if (element.isOneUnit()) {
                if (element.isDropped()) {
                    g2.setColor(COLOR_EVENT_LINES);
                    drawDroppedUnit(g2, element, area, units);
                } else {
                    g2.setColor(COLOR_EVENT_LINES);
                    drawNormalUnit(g2, element, area);
                }
            } else {
                g2.setColor(COLOR_EVENT_AREA);
                drawFatUnit(g2, element, area);
            }
        }
    }

    private final void paintEventValues(Graphics2D g2, List units, Rectangle area) {
        Iterator iter = units.iterator();
        while (iter.hasNext()) {
            TransferUnit element = (TransferUnit)iter.next();
            drawUnitFieldValue(g2, element, area);
        }
    }

    /** Paints the event values inside the columns. Note that this method is
     * fully parameterized, it doesn't use any instance field values, only
     * values from parameters. This makes it flexible for drawing both directly on the
     * component or on a buffered image.
     *
     * @param g2 Graphics2D context where to draw.
     * @param units A list of TransferUnits whose values will be drawn.
     * @param columnModel The columnModel for the columns.
     * @param bottomCoordinate The absolute bottom coordinate for drawable area.
     * @param currentTime The imaginary 'timeNow' that will be used so that no
     * values for events happening after this time will be drawn. 
     */
    private final void paintColumnValues(
        Graphics2D g2, 
        List units, 
        MSCColumnModel columnModel,
        int bottomCoordinate,
        int currentY)
    {
        Shape oldClip = g2.getClip();

        MSCColumnModel.MSCColumn[] colsLeft = columnModel.getLeftCols();
        MSCColumnModel.MSCColumn[] colsRight = columnModel.getRightCols();

        int rowCount = 0;
        
        for (int i=0; i < colsLeft.length; i++) {
        	if (printTimestamp && i == 0) {
        		Iterator iter = units.iterator();
        		while (iter.hasNext()) {
        			TransferUnit element = (TransferUnit)iter.next();
        			paintTimestampColumnValue(g2, colsLeft[i], colsRight[i], element, currentY, bottomCoordinate);
        		}
        	}
        	else {
        		Iterator iter = units.iterator();
        		while (iter.hasNext()) {
        			TransferUnit element = (TransferUnit)iter.next();

        			if (element.getSource().equals(settings.getLeftHost())) {
        				paintSingleColumnValue(g2, colsLeft[i], element, currentY, bottomCoordinate, true, rowCount);
        				paintSingleColumnValue(g2, colsRight[i], element, currentY, bottomCoordinate, false, rowCount);
        			} else if (element.getSource().equals(settings.getRightHost())) {
        				paintSingleColumnValue(g2, colsLeft[i], element, currentY, bottomCoordinate, false, rowCount);
        				paintSingleColumnValue(g2, colsRight[i], element, currentY, bottomCoordinate, true, rowCount);
        			} else {
                    // hosts don't match, do nothing
        			} 
        		} // end while
        		rowCount++;
        	} // end if-else
        } // end for

        g2.setClip(oldClip);

    } // end method paintColumns
    
    
    /**
     * Paints the timestamp for current event.
     * 
	 * @param g2
	 * @param column
	 * @param column2
	 * @param element
	 * @param currentTime
	 * @param bottomCoordinate
	 */
	private void paintTimestampColumnValue(
			Graphics2D g2, 
			MSCColumn leftColumn, 
			MSCColumn rightColumn, 
			TransferUnit unit, 
			int currentYCoord, 
			int bottomCoordinate) {
		
		String sendStamp, receiveStamp;
		
        Font oldFont = g2.getFont();
        Font newFont = new Font(oldFont.getName(), oldFont.getStyle(), fontSize);
        g2.setFont(newFont);

        // print timestamp for sendStart
        if (getYCoordForTime(unit.getSendStart()) <= currentYCoord || drawBelowNowLine) {
        	NumberFormat form = new DecimalFormat();
        	form.setMinimumFractionDigits(6);
        	form.setMaximumFractionDigits(6);
        	sendStamp = form.format(unit.getSendStart());
        	Rectangle2D bounds = g2.getFontMetrics().getStringBounds(sendStamp, g2);
        	int startY = getYCoordForTime(unit.getSendStart()) + (int)(bounds.getHeight() / 2);
        	int startX;
        	if (unit.getSource().equals(settings.getLeftHost())) {
        		g2.setClip(leftColumn.getLeftEdge(), 0, leftColumn.getWidth(), bottomCoordinate);
                if (bounds.getWidth() > leftColumn.getWidth()) {
                    // if value doesn't fit completely, align beginning to left edge
                    startX = leftColumn.getLeftEdge();
                } else {
                    // if fits, align end of string with right edge
    				startX = leftColumn.getRightEdge() - (int)bounds.getWidth();
                }	
        	} else {
        		g2.setClip(rightColumn.getLeftEdge(), 0, rightColumn.getWidth(), bottomCoordinate);
        		if (bounds.getWidth() > rightColumn.getWidth()) {
                    // if value doesn't fit completely, align beginning to left edge
                    startX = rightColumn.getLeftEdge();
                } else {
                    // if fits, align end of string with right edge
    				startX = rightColumn.getRightEdge() - (int)bounds.getWidth();
                }
        	}
        	g2.drawString(sendStamp, startX, startY);
        }
        
        // print timestamp for receive end
        if ((getYCoordForTime(unit.getReceiveEnd()) <= currentYCoord || drawBelowNowLine) && !unit.isDropped()) {
        	NumberFormat form = new DecimalFormat();
        	form.setMinimumFractionDigits(6);
        	form.setMaximumFractionDigits(6);
        	receiveStamp = form.format(unit.getReceiveEnd());
        	Rectangle2D bounds = g2.getFontMetrics().getStringBounds(receiveStamp, g2);
        	int startY = getYCoordForTime(unit.getReceiveEnd()) + (int)(bounds.getHeight() / 2);
        	int startX;
        	if (unit.getSource().equals(settings.getLeftHost())) {
        		g2.setClip(rightColumn.getLeftEdge(), 0, rightColumn.getWidth(), bottomCoordinate);
        		if (bounds.getWidth() > rightColumn.getWidth()) {
        			startX = rightColumn.getLeftEdge();
        		} else {
        			startX = rightColumn.getRightEdge() - (int)bounds.getWidth();
        		}
        	} else {
        		g2.setClip(leftColumn.getLeftEdge(), 0, leftColumn.getWidth(), bottomCoordinate);
                if (bounds.getWidth() > leftColumn.getWidth()) {
                    // if value doesn't fit completely, align beginning to left edge
                    startX = leftColumn.getLeftEdge();
                } else {
                    // if fits, align end of string with right edge
    				startX = leftColumn.getRightEdge() - (int)bounds.getWidth();
                }
        	}
        	g2.drawString(receiveStamp, startX, startY);
        }
        g2.setFont(oldFont);
	}

	private final void paintSingleColumnValue(
        Graphics2D g2, 
        MSCColumnModel.MSCColumn column,
        TransferUnit unit,
        float currentTime,
        int bottomCoordinate,
        boolean isSending,
		int rowCount) 
    {
        float time;
        String value = null;
        
        Font oldFont = g2.getFont();
        Font newFont = new Font(oldFont.getName(), oldFont.getStyle(), fontSize);
        g2.setFont(newFont);
        
        int topY = 0;
        
        topY = yDrawingTop; 
        
        if (isSending) {
            time = unit.getSendStart();
            if (unit.getValueSend(column.getVariable()) != null)
            	value = column.getVariable().getName() + ": " + unit.getValueSend(column.getVariable());
        } else {
            time = unit.getReceiveStart();
            if (unit.getValueReceive(column.getVariable()) != null)
            	value = column.getVariable().getName() + ": " + unit.getValueReceive(column.getVariable());
        }
        
        if (value != null && (drawBelowNowLine || (time <= currentTime))) {
            // don't try to process more than necessary
            if (value.length() > MAX_DRAWABLE_VALUE_LENGTH) {
                value = value.substring(0, MAX_DRAWABLE_VALUE_LENGTH);
            }
            
            // get exact bounds for the values in this context
            // (needed for right-aligning the values)
            Rectangle2D bounds = g2.getFontMetrics().getStringBounds(value, g2);

            int offset = g2.getFontMetrics().getHeight() * rowCount;
            
            g2.setClip(column.getLeftEdge(), 0, 
                column.getWidth(), bottomCoordinate);

            
            int startX;
            if (bounds.getWidth() > column.getWidth()) {
                // if value doesn't fit completely, align beginning to left edge
                startX = column.getLeftEdge();
            } else {
                // if fits, align end of string with right edge
				startX = column.getRightEdge() - (int)bounds.getWidth();
            }
            
            /* Draw value inside the column so that for sender 
             * the top edge of the visible string is aligned with event time and for 
             * receiver the bottom edge of string is aligned with event time.
             */
            int startY;
            if (isSending) {
            	startY = getYCoordForTime(time) + offset + (int)(bounds.getHeight() / 2);
            } else {
            	startY = getYCoordForTime(time) + offset + (int)(bounds.getHeight() / 2); 
            }
            
            g2.drawString(value, startX, startY);
        }
        g2.setFont(oldFont);
    }

    private final void paintColumnBorders(Graphics2D g2, MSCColumnModel columnModel) {
        g2.setColor(COLOR_COLUMN_BORDERS);
        paintBordersForColumn(g2, columnModel.getNotesCol());
        for (int i=0; i < columnModel.getLeftCols().length; i++) {
            paintBordersForColumn(g2, columnModel.getLeftCols()[i]);
        }
        paintBordersForColumn(g2, columnModel.getDrawingArea());
        for (int i=0; i < columnModel.getRightCols().length; i++) {
            paintBordersForColumn(g2, columnModel.getRightCols()[i]);
        }
    }
    
    private final void paintBordersForColumn(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn column) {
        g2.drawLine(column.getLeftEdge()-1, getInsets().top,
        column.getLeftEdge()-1, getHeight() - getInsets().bottom);
        g2.drawLine(column.getRightEdge()+1, getInsets().top,
        column.getRightEdge()+1, getHeight() - getInsets().bottom);
    }

    private final void paintNotesColumn(Graphics2D g2, MSCPanel.MSCColumnModel columnModel, 
        Collection notes, Rectangle area) 
    {
        Iterator iterator = notes.iterator();
        while (iterator.hasNext()) {
            Note element = (Note)iterator.next();
            g2.drawImage(noteImage, null, area.x + 2, getYCoordForTime(element.getTime()));
        }
    }


	
	private final void drawNormalUnit(Graphics2D g2, TransferUnit unit, Rectangle area) {
		int leftY, rightY, yArrowPoint, xArrowPoint, arrowDirection;
		int topY;
		topY = yDrawingTop; 
		if (unit.getSource().equals(settings.getLeftHost())) {
			leftY = getYCoordForTime(unit.getSendStart());
			rightY = getYCoordForTime(unit.getReceiveEnd());
			yArrowPoint = rightY;
			xArrowPoint = area.x + area.width;
			arrowDirection = -1;
		} else {
			leftY = getYCoordForTime(unit.getReceiveEnd());
			rightY = getYCoordForTime(unit.getSendStart());
			yArrowPoint = leftY;
			xArrowPoint = area.x;
			arrowDirection = 1;
		}
		g2.drawLine(area.x, leftY, area.x + area.width, rightY);
		
		// arrow heads
		int xArrowHead[] = new int[3];
		int yArrowHead[] = new int[3];
		xArrowHead[0] = 0; yArrowHead[0] = 0;
		xArrowHead[1] = arrowDirection * MSC_ARROW_LENGTH;
		yArrowHead[1] = MSC_ARROW_HEIGHT;
		xArrowHead[2] = xArrowHead[1];
		yArrowHead[2] = -MSC_ARROW_HEIGHT;
		
		double angle = -arrowDirection * Math.atan((double)(getYCoordForTime(unit.getReceiveEnd()) 
				- getYCoordForTime(unit.getSendStart())) / area.width);
		
		g2.translate(xArrowPoint, yArrowPoint);
		g2.rotate(angle);
		g2.fillPolygon(xArrowHead, yArrowHead, 3);
		g2.rotate(-angle);
		g2.translate(-xArrowPoint, -yArrowPoint);
	}


    private final void drawFatUnit(Graphics2D g2, TransferUnit unit, Rectangle area) {
        if (DISABLE_ANTIALIAS_FOR_FAT_UNITS) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
        }

        int leftYStart, leftYEnd, rightYStart, rightYEnd;

        if (unit.getSource().equals(settings.getLeftHost())) {
            leftYStart = getYCoordForTime(unit.getSendStart());
            leftYEnd = getYCoordForTime(unit.getSendEnd());
            rightYStart = getYCoordForTime(unit.getReceiveStart());
            rightYEnd = getYCoordForTime(unit.getReceiveEnd());
        } else {
            leftYStart = getYCoordForTime(unit.getReceiveStart());
            leftYEnd = getYCoordForTime(unit.getReceiveEnd());
            rightYStart = getYCoordForTime(unit.getSendStart());
            rightYEnd = getYCoordForTime(unit.getSendEnd());
        }

        int[] xPoints = new int[5];
        int[] yPoints = new int[5];
        xPoints[0] = area.x; yPoints[0] = leftYStart;
        xPoints[1] = area.x + area.width; yPoints[1] = rightYStart;
        xPoints[2] = area.x + area.width; yPoints[2] = rightYEnd;
        xPoints[3] = area.x; yPoints[3] = leftYEnd;
        xPoints[4] = area.x; yPoints[4] = leftYStart;

        g2.fillPolygon(xPoints, yPoints, 5);

        if (useAntiAlias) {
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        }
    }

	private final void drawDroppedUnit(Graphics2D g2, TransferUnit unit, Rectangle area, List units) {

		int point[] = getDroppingPoint(g2, unit, area);
		if (unit.getSource().equals(settings.getLeftHost())) {
			g2.drawLine(area.x, getYCoordForTime(unit.getSendStart()), point[0], point[1]);
		} else {
			g2.drawLine(area.x + area.width, getYCoordForTime(unit.getSendStart()), point[0], point[1]);
		}
		drawDroppedCross(g2, point[0], point[1], DROPPED_CROSS_WIDTH);
	}

    private final void drawDroppedCross(Graphics2D g2, int x, int y, int width) {
        Color oldColor = g2.getColor();
        g2.setColor(COLOR_DROPPED_CROSS);
        g2.drawLine(x-width/2, y-width/2, x+width/2, y+width/2);
        g2.drawLine(x-width/2, y+width/2, x+width/2, y-width/2);
        g2.setColor(oldColor);
    }

    private final int[] getDroppingPoint(Graphics2D g2, TransferUnit unit, Rectangle area) {
    	int[] point = new int[2];
    	
    	if (!unit.isDropped()) {
    		if (unit.getSource().equals(settings.getLeftHost())) {
    			point[0] = area.x;
    		} else {
    			point[0] = area.x + area.width;
    		}
    		point[1] = getYCoordForTime(unit.getReceiveStart());
    		return point;
    	}
    	
    	// Need to find nearest packet that is send from same side... 
		List new_units = dataView.getUnitsForLayer(settings.getLayer());
		LinkedList nearest = new LinkedList();
		Iterator iter = new_units.iterator();
		ListIterator liter;
		TransferUnit temp, next=null, prev=null;
		
		while (iter.hasNext()) {
			temp = (TransferUnit)iter.next();
			if (temp.getSource().equals(unit.getSource()))
				nearest.add(temp);
		}
		
		Collections.sort(nearest); // Just in case
		return reallyGetDroppingPoint(g2, unit, area, nearest);
    }

    private final int[] reallyGetDroppingPoint(Graphics2D g2, TransferUnit unit, Rectangle area, 
    		LinkedList nearest) {
    	int[] point = new int[2];
    	boolean useLastGoodAngle = false;
    	
    	if (!unit.isDropped()) {
    		if (unit.getSource().equals(settings.getLeftHost())) {
    			point[0] = area.x;
    		} else {
    			point[0] = area.x + area.width;
    		}
    		point[1] = getYCoordForTime(unit.getReceiveStart());
    		return point;
    	}
    	
		ListIterator liter;
		TransferUnit temp, next=null, prev=null;
		
		int unitIndex = nearest.indexOf(unit);
		if (unitIndex == -1) {
    		if (unit.getSource().equals(settings.getLeftHost())) {
    			point[0] = area.x;
    		} else {
    			point[0] = area.x + area.width;
    		}
    		point[1] = getYCoordForTime(unit.getReceiveStart());
     		return point;
 		}
		liter = nearest.listIterator(unitIndex);
		if (liter.hasPrevious()) {
			prev = (TransferUnit)liter.previous();
			liter.next();
		}
		liter.next();
		// liter should now point after unit. 
		if (liter.hasNext()) {
			next = (TransferUnit)liter.next();
		}

		int nearestStartY, nearestStopY;
		nearestStartY = nearestStopY = 0;
		if (prev != null && next != null) {
			if (Math.abs(getYCoordForTime(next.getSendStart()) - getYCoordForTime(unit.getSendStart())) 
					< Math.abs(getYCoordForTime(unit.getSendStart()) - getYCoordForTime(prev.getSendStart()))) {
				if (!next.isDropped()) {
					nearestStartY = getYCoordForTime(next.getSendStart());
					nearestStopY = getYCoordForTime(next.getReceiveEnd());
				} else { // use prev as nearest packet
					nearestStartY = getYCoordForTime(prev.getSendStart());
					if (!prev.isDropped()) {
						nearestStopY = getYCoordForTime(prev.getReceiveEnd());
					} else {
						// We should use lastGoodAngle
						useLastGoodAngle = true;
						
					}
				}
			} else {
				nearestStartY = getYCoordForTime(prev.getSendStart());
				if (!prev.isDropped()) {
					nearestStopY = getYCoordForTime(prev.getReceiveEnd());
				} else {
					useLastGoodAngle = true;
				}
			}
		} else if (prev != null) {
			nearestStartY = getYCoordForTime(prev.getSendStart());
			if (!prev.isDropped()) {
				nearestStopY = getYCoordForTime(prev.getReceiveEnd());
			} else {
				useLastGoodAngle = true;
			}
		} else if (next != null) {
			if (!next.isDropped()) {
				nearestStartY = getYCoordForTime(next.getSendStart());
				nearestStopY = getYCoordForTime(next.getReceiveEnd());
			} else {
				nearestStartY = getYCoordForTime(unit.getSendStart());
				nearestStopY = nearestStartY; 
			}
		} else {
			nearestStartY = getYCoordForTime(unit.getSendStart());
			nearestStopY = nearestStartY; 
		}
		
		// Find out how much space do we need for text over line

        StringBuffer sb = new StringBuffer();
        VariableDefinition[] variables = settings.getVariablesCenter();
        for (int i = 0; i < variables.length; i++) {
            if (i>0) sb.append("  ");
            sb.append(variables[i].getName()+"="+unit.getValue(variables[i]));
        }

        Font oldFont = g2.getFont();
        Font newFont = new Font(oldFont.getName(), oldFont.getStyle(), fontSize);
        g2.setFont(newFont);
        Rectangle2D bounds = g2.getFontMetrics().getStringBounds(sb.toString(), g2);
        g2.setFont(oldFont);
        
        // Calculate hypotenusa and angle
        // Now you here, maybe we want this thing to put dropped cross 2*DROPPED_CROSS_WIDTH away from starting location.
        double hypotenusa, angle;
        if (useLastGoodAngle) {
        	angle = (unit.getSource().equals(settings.getLeftHost())) ? lastGoodLeftAngle : lastGoodRightAngle;
        	hypotenusa = Math.max(bounds.getWidth() + DROPPED_CROSS_WIDTH, 2 * DROPPED_CROSS_WIDTH); 
        } else {
	        hypotenusa = Math.max(bounds.getWidth() + DROPPED_CROSS_WIDTH, 2 * DROPPED_CROSS_WIDTH);
	        angle = Math.atan((float)(nearestStopY - nearestStartY) / area.width);
	        if (unit.getSource().equals(settings.getLeftHost())) {
	        	lastGoodLeftAngle = angle;
	        } else {
	        	lastGoodRightAngle = angle;
	        }
        }
        // Calculate xDiff and yDiff 
        int xDiff = Math.round(Math.round(hypotenusa * Math.sin(Math.PI / 2 - angle)));
        int yDiff = Math.round(Math.round(hypotenusa * Math.sin(angle)));
        
        point[1] = getYCoordForTime(unit.getSendStart()) + yDiff;
        if (unit.getSource().equals(settings.getLeftHost())) {
        	point[0] = area.x + xDiff;
        } else {
        	point[0] = area.x + area.width - xDiff;
        }
                
        // should NOT draw anythin below end and over the edges.
        if (point[1] + yDrawingTop > coordinates.getLastYCoord()) {
        	point[1] = coordinates.getLastYCoord() - yDrawingTop;
        }
        if (point[0] < area.x + DROPPED_CROSS_WIDTH / 2) {
        	point[0] = area.x + DROPPED_CROSS_WIDTH / 2;
        } else if (point[0] > area.x + area.width - DROPPED_CROSS_WIDTH / 2) {
        	point[0] = area.x + area.width - DROPPED_CROSS_WIDTH / 2;
        }
                
        return point;
    }

    
    // TODO dropped packets that are sent from right host have texts too near to dropping cross.
    private final void drawUnitFieldValue(Graphics2D g2, TransferUnit unit, Rectangle area) {
        int leftY, rightY, centY;
        int[] point = null;
        double xDiff;
        Font oldFont = g2.getFont();
        Font newFont = new Font(oldFont.getName(), oldFont.getStyle(), fontSize);
        g2.setFont(newFont);
        if (!unit.isDropped()) {
        	if (unit.getSource().equals(settings.getLeftHost())) {
        		leftY = getYCoordForTime(unit.getSendStart());
        		rightY = getYCoordForTime(unit.getReceiveStart());
        	} else {
        		leftY = getYCoordForTime(unit.getReceiveStart());
        		rightY = getYCoordForTime(unit.getSendStart());
        	}
            xDiff = area.width;        	
        } else {

        	point = getDroppingPoint(g2, unit, area);

        	if (unit.getSource().equals(settings.getLeftHost())) {
        		leftY = getYCoordForTime(unit.getSendStart());        		
        		rightY = point[1]; 
        	} else {
        		rightY = getYCoordForTime(unit.getSendStart());
        		leftY = point[1];

        	}
        	xDiff = (unit.getSource().equals(settings.getLeftHost())) ? point[0] - area.x : area.x + area.width - point[0];
        }
        
        double yDiff = rightY-leftY;
        
        double angle = Math.atan(yDiff / xDiff);

        StringBuffer sb = new StringBuffer();
        VariableDefinition[] variables = settings.getVariablesCenter();
        for (int i = 0; i < variables.length; i++) {
            if (i>0) sb.append("  ");
            sb.append(variables[i].getName()+"="+unit.getValue(variables[i]));
        }

        Rectangle2D bounds = g2.getFontMetrics().getStringBounds(sb.toString(), g2);
        
        int stringLeftEdge = (int)(area.x + area.width / 2.0 - bounds.getWidth() / 2.0);
        if (stringLeftEdge < area.x) stringLeftEdge = area.x;
        
        if (unit.isDropped() && point != null) {
        	if (unit.getSource().equals(settings.getRightHost()))
            	stringLeftEdge = point[0];
        	else 
        		stringLeftEdge = area.x;
        }


        int stringStartY;
        if (unit.isDropped()) {
        	if (unit.getSource().equals(settings.getLeftHost())) {
        		int stringStartYCorrection = (int)(Math.tan(angle) * (double)(stringLeftEdge - area.x));
        		stringStartY = leftY + stringStartYCorrection - EVENT_VALUE_YDIFF;
        	} else {
        		stringStartY = leftY - EVENT_VALUE_YDIFF;
        	}
        } else {
            int stringStartYCorrection = (int)(Math.tan(angle) * (double)(stringLeftEdge-area.x));
            stringStartY = leftY + stringStartYCorrection - EVENT_VALUE_YDIFF;
        }

        AffineTransform oldTransform = g2.getTransform();
        g2.transform(AffineTransform.getRotateInstance(angle, stringLeftEdge, stringStartY));

        g2.drawString(sb.toString(), stringLeftEdge, stringStartY);

        g2.setTransform(oldTransform);
        g2.setFont(oldFont);
    }

    private final int getYCoordForTime(float time) {
    	return coordinates.getYCoordForTime(time) - yDrawingTop;
    }
    
    private final int getLinearYCoordForTime(float time) {
    	if (useRealTime) {
     		return coordinates.getYCoordForTime(time) ;
    	} else {
     		return coordinates.getLinearYCoordForTime(time);
    	}
    }
    
    private final int getAllwaysLinearYCoordForTime(float time) {
    	return coordinates.getLinearYCoordForTime(time);
    }
    
    private final float getAllwaysLinearTimeForYCoord(int y) {
    	return coordinates.getLinearTimeForYCoord(y);
    }
    
    private final float getTimeForYCoord(int y) {
    	return coordinates.getTimeForYCoord(y);
    }
    
    private final float getLinearTimeForYCoord(int y) {
    	if (useRealTime) {
    		return coordinates.getTimeForYCoord(y);
    	} else {
    		return coordinates.getLinearTimeForYCoord(y);
    	}
    }
    
    private final void doMouseSelecting(int y) {
    	isMouseDragging = true;
    	float time = getLinearTimeForYCoord(y + yDrawingTop);
    	mainFrame.getAnimationTimeState().setNowTime(time);
    }
    
    public void mouseWheelMoved(java.awt.event.MouseWheelEvent evt) {
        jScrollBar1.setValue(jScrollBar1.getValue() + 
            (evt.getUnitsToScroll() * jScrollBar1.getUnitIncrement()));
    }
    
} // end inner class MSCDrawingPanel


/** MSCColumnModel encapsulates the information about different columns drawn in
 * the MSCDrawingPanel. MSCColumnModel has four main parts, from left to right:
 * note column, left variable columns, drawing area, right variable columns.
 * Note column and drawing area are always represented by exactly one MSCColumn.
 * Left and right variable columns are represented by 0..2 MSCColumn objects
 * each.
 *
 * <p>Currently MSCColumnModel is implemented so that the same columns always
 * exist in left and in right, and the order of columns is 'centric' in
 * relation to the drawing are: the same column is always closest to the
 * drawing area both in left and in right.
 * 
 * <p>The actual MSCColumn objects always represent the usable area between
 * the column border lines: the getLeftEdge returns the first pixel that can
 * be used for plotting and getRightEdge returns the last. MSCColumnModel takes
 * care of reserving one pixel between each column (and on the left and right
 * sides of the left- and rightmost columns).
 *
 *<p>MSCColumnModel has actually two different kind of columns as variable columns.
 * They are timestamp column and variable column. Not other columns are currently
 * not supported as variable columns. Timestamp column is drawed if 
 * <code>printTimestamp == true</code>. Variable column is drawed if user has select
 * any variables to shown (e.g. <code>!columnVariables.isEmpty()</code>).
 * 
 * <p>To calculate the width of character MSCColumnModel uses the width of '8' at
 * plain Verdana with selected font size as reference. The width of timestamp column,
 * on the other hand, is calculated as a width of string "888.888888" using same font
 * as above. <em>The width of timestamp column is never reduced.</em> For the variable
 * column the width is calculated so that all values and their captions can be written
 * to this field. Should this width cause drawing area to be smaller than 
 * WIDTH_DRAWING_AREA_MINIMUM, the available width is given to variable host. That is,
 * the variable columns are shrunk as small as needed to get drawing area its minimum size.
 * The widht of notes column is WIDTH_COLUMN_NOTES and it's allways included to the
 * overall widht (as notes column is allways drawed).
 */
public class MSCColumnModel {

    private final MSCColumn columnNotes;

    /* Index 0 is closest to drawing area, i.e. right-most of left columns */
    private MSCColumn[] columnsLeft;
    private final MSCColumn columnDrawingArea;

    /* Index 0 is closest to drawing area, i.e. left-most of right columns */
    private MSCColumn[] columnsRight;

    private final int totalWidth;

    public MSCColumnModel(int width, VariableDefinition[] columnVariables) {
        this.totalWidth = width;
        int columnDemand = 0;
        int timestampWidth = 0;
        int headerWidth = 0;
        int columnWidth = 0;
        
        // Calculate the width of current font size.
        // TODO does not know what the font is, thus this one must be changed if font name or type is changed.
        Font font = new Font("verdana", Font.PLAIN, fontSize);
        
        // Width is calculated from '8' as default font render context.
        // TODO is '8' correct character to calculate width for numbers? 
        int fontWidth = (int)font.getStringBounds("8", new FontRenderContext(null, false, false)).getWidth(); 
        for (int i=0; i < columnVariables.length; i++) {
        	headerWidth = (int)font.getStringBounds(columnVariables[i].getName()+": ", 
        			new FontRenderContext(null, false, false)).getWidth();
            columnDemand = Math.max(2 * (1 + fontWidth * dataView.getValueMaxLength(columnVariables[i])
            		+ headerWidth), columnDemand);
            columnWidth = Math.max(fontWidth * dataView.getValueMaxLength(columnVariables[i]) + headerWidth,
            		columnWidth);
        }

        /* Calculate the space needed by timestamp column, ugly way.
         * We presume that the maximum precision of timestamp is 6 digits and no of them has more than three
         * digits left from decimal (that is all timestamps are smaller than 999.999999 seconds).
         */
        if (printTimestamp) {
        	timestampWidth = (int)font.getStringBounds("888.888888", new FontRenderContext(null, false, false)).getWidth();
        	columnDemand += 2 * timestampWidth;
        }
        
        float columnWidthFactor;
        int drawingAreaWidth;

        if (columnDemand > 0 &&
            WIDTH_COLUMN_NOTES + 2 + WIDTH_DRAWING_AREA_MINIMUM + columnDemand > width)
        {
        	// Timestamp column is never shrinked!
        	int availableForColumns = width - WIDTH_DRAWING_AREA_MINIMUM - WIDTH_COLUMN_NOTES - 2;
            columnWidthFactor = (float)(availableForColumns - 2 * timestampWidth) / (float)columnDemand;
            drawingAreaWidth = WIDTH_DRAWING_AREA_MINIMUM;
        } else {
            columnWidthFactor = 1.0f;
            drawingAreaWidth = width - WIDTH_COLUMN_NOTES - columnDemand - 2;
        }

        columnNotes = new MSCColumn(null, 1, WIDTH_COLUMN_NOTES-1);

        float previousRightEdge = WIDTH_COLUMN_NOTES-1;

        columnsLeft = new MSCColumn[columnVariables.length];
        for (int i = columnVariables.length-1; i >= 0; i--) {
            float currentColumnWidth = dataView.getValueMaxLength(columnVariables[i])
                * fontWidth * columnWidthFactor;

            float newLeftEdge = WIDTH_COLUMN_NOTES + 1f; 
            float newRightEdge = newLeftEdge + columnWidth * columnWidthFactor + 1f;

            MSCColumn tempColumn = new MSCColumn(
                columnVariables[i],
                Math.round(newLeftEdge),
                Math.round(newRightEdge)
            );

            columnsLeft[i] = tempColumn;
            previousRightEdge = newRightEdge;
        }

        // Timestamp column
        if (printTimestamp) {
        	MSCColumn[] temp = new MSCColumn[columnVariables.length];
        	for (int i = 0; i < columnVariables.length; i++)
        		temp[i] = columnsLeft[i];
        	columnsLeft = new MSCColumn[columnVariables.length + 1];
        	columnsLeft[0] = new MSCColumn(null, Math.round(previousRightEdge + 2f), 
        			Math.round(previousRightEdge + 1f + timestampWidth));
        	previousRightEdge += 1f + timestampWidth;
        	for (int i = 0; i < columnVariables.length; i++)
        		columnsLeft[i+1] = temp[i];
        }
        
        columnDrawingArea = new MSCColumn(
            null,
            Math.round(previousRightEdge + 2f),
            Math.round(previousRightEdge + 2f + drawingAreaWidth-1f)
        );

        previousRightEdge = previousRightEdge + 2f + drawingAreaWidth - 1f;
        
        // Timestamp column
        MSCColumn tempTimestamp = null;
        if (printTimestamp) {
        	float newLeftEdge = previousRightEdge + 2f;
        	float newRightEdge = previousRightEdge + 1f + timestampWidth;
        	if (newLeftEdge < width && Math.round(newRightEdge) > width - 2) {
        		newRightEdge = width - 2;
        	}
        	tempTimestamp = new MSCColumn(null, Math.round(newLeftEdge),
        			Math.round(newRightEdge));
        	previousRightEdge = newRightEdge;
        }

        float rightColumnBegins = previousRightEdge + 2f;
        columnsRight = new MSCColumn[columnVariables.length];
        for (int i=0; i < columnVariables.length; i++) {
            float currentColumnWidth = dataView.getValueMaxLength(columnVariables[i])
                * fontWidth * columnWidthFactor;

            float newLeftEdge = rightColumnBegins; 
            float newRightEdge = newLeftEdge + columnWidth * columnWidthFactor + 1f; 
            
            /* Final correction for the rightmost edge of the whole column model.
             * Because of precision issues we will never get this right without
             * a manual correction... */
            if (newLeftEdge < width && Math.round(newRightEdge) > width-2) {
                newRightEdge = width-2;
            }

            MSCColumn tempColumn = new MSCColumn(
                columnVariables[i],
                Math.round(newLeftEdge),
                Math.round(newRightEdge)
            );

            columnsRight[i] = tempColumn;
            previousRightEdge = newRightEdge;
        }
        
        // Timestamp column
        if (printTimestamp) {
        	MSCColumn[] temp = new MSCColumn[columnVariables.length];
        	for (int i = 0; i < columnVariables.length; i++)
        		temp[i] = columnsRight[i];
        	columnsRight = new MSCColumn[columnVariables.length + 1];
        	columnsRight[0] = tempTimestamp;
        	for (int i = 0; i < columnVariables.length; i++)
        		columnsRight[i+1] = temp[i];
        }

    }

    /** Getter for property columnNotes.
     * @return Value of property columnNotes.
     *
     */
    public MSCColumnModel.MSCColumn getNotesCol() {
        return columnNotes;
    }

    /** Getter for property columnsLeft.
     * @return Value of property columnsLeft.
     *
     */
    public MSCColumnModel.MSCColumn[] getLeftCols() {
        return this.columnsLeft;
    }

    /** Getter for property columnDrawingArea.
     * @return Value of property columnDrawingArea.
     */
    public MSCColumnModel.MSCColumn getDrawingArea() {
        return columnDrawingArea;
    }

    /** Getter for property columnsRight.
     * @return Value of property columnsRight.
     */
    public MSCColumnModel.MSCColumn[] getRightCols() {
        return this.columnsRight;
    }

    public int getTotalWidth() {
        return totalWidth;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer("MSCColumnModel (total "+getTotalWidth()+")");

        if (columnNotes != null) {
            sb.append(", notes "+columnNotes.toString());
        }
        if (columnsLeft.length > 0) {
            sb.append(", left ");
            for (int i=0; i < columnsLeft.length; i++) {
                sb.append(columnsLeft[i].toString());
            }
        }

        if (columnDrawingArea != null) {
            sb.append(", center "+columnDrawingArea.toString());
        }

        if (columnsRight.length > 0) {
            sb.append(", right ");
            for (int i=0; i < columnsRight.length; i++) {
                sb.append(columnsRight[i].toString());
            }

        }

        return sb.toString();
    }


    /* A simple class to express the left and right edges of one column in the column model,
     * including also the reference to VariableDefinition in case of variable columns.
     * The left and right edge values are both included in the drawable inside area of the
     * column.
     */
    public class MSCColumn {

        private final VariableDefinition variable;
        private final int leftEdge;
        private final int rightEdge;

        public MSCColumn(VariableDefinition variable, int leftEdge, int rightEdge) {
            this.variable = variable;
            this.leftEdge = leftEdge;
            this.rightEdge = rightEdge;
        }

        public VariableDefinition getVariable() {
            return variable;
        }

        public int getLeftEdge() {
            return leftEdge;
        }

        public int getRightEdge() {
            return rightEdge;
        }

        public int getWidth() {
            return rightEdge - leftEdge + 1;
        }

        public String toString() {
            return "("+leftEdge+","+rightEdge+")";
        }
    }
}


private class HostVariableTableModel extends javax.swing.table.AbstractTableModel {

    private String[][] tableData;

    public HostVariableTableModel(Host host) {
        List staticVariables = host.getStaticVariables();
        tableData = new String[2][staticVariables.size()];
        int rowIndex = 0;

        Iterator iter = staticVariables.iterator();
        while (iter.hasNext()) {
            StaticVariable element = (StaticVariable)iter.next();
            tableData[0][rowIndex] = element.getName();
            tableData[1][rowIndex] = element.getValue();
            rowIndex++;
        }
    }

    public int getColumnCount() {
        return 2;
    }

    public int getRowCount() {
        return tableData[0].length;
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        return tableData[columnIndex][rowIndex];
    }

}

private class ColumnHeadingPanel extends JPanel {

    public ColumnHeadingPanel() {
        setOpaque(false);
    }

    protected void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        MSCPanel.MSCColumnModel columnModel = getColumnModel();

        g2.setColor(COLOR_COLUMN_BORDERS);
        paintBorders(g2, columnModel.getNotesCol());
        for (int i=0; i < columnModel.getLeftCols().length; i++) {
            paintBorders(g2, columnModel.getLeftCols()[i]);
        }

        paintBorders(g2, columnModel.getDrawingArea());
        for (int i=0; i < columnModel.getRightCols().length; i++) {
            paintBorders(g2, columnModel.getRightCols()[i]);
        }

        g2.drawLine(getInsets().left, getHeight() - getInsets().bottom - 1,
            getWidth() - getInsets().right, getHeight() - getInsets().bottom - 1);

        g2.setColor(COLOR_COLUMN_VALUES);

        paintLabels(g2, columnModel.getLeftCols());
        paintLabels(g2, columnModel.getRightCols());
    }

    private void paintLabels(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn[] columns) {
        for (int i=0; i < columns.length; i++) {
            g2.setClip(columns[i].getLeftEdge(), 0,
            		columns[i].getWidth(), getHeight());
            
            String label;
            if (printTimestamp && i == 0) {
            	label = Localization.getString("panel.msc.timestamp_label");
            	Rectangle2D bounds = g2.getFontMetrics().getStringBounds(label, g2);
            	g2.drawString(label, columns[0].getLeftEdge()+1, (int)bounds.getHeight());
            } else {
            	label = Localization.getString("panel.msc.variables_label");
            	Rectangle2D bounds = g2.getFontMetrics().getStringBounds(label, g2);
            	g2.drawString(label, columns[i].getLeftEdge()+1, (int)bounds.getHeight());
            }
        }
    }

    private void paintBorders(Graphics2D g2, MSCPanel.MSCColumnModel.MSCColumn column) {
        g2.drawLine(column.getLeftEdge()-1, getInsets().top,
        column.getLeftEdge()-1, getHeight() - getInsets().bottom);
        g2.drawLine(column.getRightEdge()+1, getInsets().top,
        column.getRightEdge()+1, getHeight() - getInsets().bottom);
    }
}
    
}
