
package fi.helsinki.dacopan.ui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.Note;
import fi.helsinki.dacopan.model.TransferUnit;

/**
 */
public class NotePanel extends AbstractAnimationPanel {
    
    //Instance Variables
    /**
     * This is the layer that we are currently showing in the MSC
     * and UFO panels. If this is set to a real layer instance,
     * then the TransferUnit should be null.
     */
    private final Layer layer;
    /**
     * This is the root unit of an encapsulation tree. The NoteManager
     * can get a note text for this NotePanel if this is
     * set to a non-null TransferUnit instance. However, nothing
     * guarantees that there is a note for this rootUnit. Nevertheless,
     * if this is non-null, then the layer should be null.
     */
    private final TransferUnit rootUnit;
    /**
     * The actual Note object to display.
     */
    private Note note;
    
    /**
     * True if we're showing enc notes.
     */
    private final boolean inEncMode;

    
    // Various widgets in the Note Panel    
    private JScrollPane noteTextScrollPane;
    private JTextArea noteTextArea;
    private TitlePanel noteTitle;

    private JButton editNoteButton;
    private JButton deleteNoteButton;
    private JButton addNoteButton;
    private float currentTime = 0.0f;

    // Reference to this panel itself
    private NotePanel notePanelReference;
	
    
    /**
     * This is the main (private) constructor for a NotePanel. It sets up the
     * panel in the proper way. Note that only one of the two params
     * layer and transferUnit can be non-null at the same time.
     * 
     * @param mainFrame
     * @param layer
     * @param transferUnit
     */
    private NotePanel(MainFrame mainFrame, Layer layer, TransferUnit transferUnit) {
        super(mainFrame);
        notePanelReference = this;
        if(layer==null && transferUnit==null){
        	throw new IllegalArgumentException("Layer and TransferUnit cannot be null at the same time!");
        }
        this.layer = layer;
        this.rootUnit = transferUnit;
        if(layer==null){
            inEncMode = true;
        }else{
            inEncMode = false;
        }
        setLayout(new BorderLayout());
        addWidgets();
        
        currentTime = mainFrame.getMSCPanel().getRealTime();
    }
    
    /**
     * This constructs a NotePanel for showing notes for MSC. The layer
     * tells what layer we want the notes to come from, in other words
     * what layer we are looking at in the MSC.
     * 
     * @param mainFrame
     * @param layer
     */
    public NotePanel(MainFrame mainFrame, Layer layer) {
        this(mainFrame, layer, null);
    }

    /**
     * This constructs a NotePanel for showing encapsulation notes.
     * It fetches the note from the NoteManager based on the transferUnit.
     * See NoteManager.
     * 
     * @param mainFrame
     * @param transferUnit
     */
    public NotePanel(MainFrame mainFrame, TransferUnit transferUnit) {
        this(mainFrame, null, transferUnit);
        this.note = mainFrame.getNoteManager().getNoteEnc(transferUnit);
        
        refreshNoteTextArea();
        
        // enable/disable the buttons depending on
        // if there is a note available or not
        refreshButtons();        
    }
    
    /**
     * Adds all the necessary widgets to the panel.
     */
    protected void addWidgets() {
        
        if(inEncMode){
            noteTitle = new TitlePanel(Localization.getString("panel.note.title_enc_mode"), null); //$NON-NLS-1$
        }else{
            noteTitle = new TitlePanel(Localization.getString("panel.note.title_msc_mode"), layer.getName());             //$NON-NLS-1$
        }
        
        noteTextArea = new JTextArea();
        noteTextArea.setEditable(false);       
        noteTextScrollPane = new JScrollPane(noteTextArea);
        noteTextScrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        noteTextArea.setMargin(new Insets(5,10,5,25));
        noteTextArea.setLineWrap(true);
        
        JPanel buttonsPanel = new JPanel(new GridBagLayout());        
        
        GridBagConstraints c = new GridBagConstraints();      
        c.gridx = GridBagConstraints.RELATIVE; 
        c.gridy = GridBagConstraints.RELATIVE; 
        c.insets = new Insets(2,2,2,2);               
        
        addNoteButton = new JButton(Localization.getString("panel.note.add_note")); //$NON-NLS-1$
        //addNoteButton.setMinimumSize(new Dimension(85,25));
        UserInterface.setDefaultButtonFont(addNoteButton);
        addNoteButton.addActionListener(new ActionListener(){

            public void actionPerformed(ActionEvent arg0) {
                EditNoteDialog dialog = 
                    new EditNoteDialog(mainFrame, notePanelReference, EditNoteDialog.NOTE_EDIT_MODE_ADD, note);
                dialog.pack();
                dialog.setModal(true);
                dialog.setLocationRelativeTo(mainFrame);
                dialog.setVisible(true);
                
                refreshNoteTextArea();
                refreshButtons();
            }
            
        });
        buttonsPanel.add(addNoteButton,c);
        
        editNoteButton = new JButton(Localization.getString("panel.note.edit_note")); //$NON-NLS-1$
        UserInterface.setDefaultButtonFont(editNoteButton);
        editNoteButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent arg0) {
				EditNoteDialog dialog = 
                    new EditNoteDialog(mainFrame, notePanelReference, EditNoteDialog.NOTE_EDIT_MODE_EDIT, note);
                dialog.pack();
                dialog.setModal(true);
                dialog.setLocation(0,0);
				dialog.setLocationRelativeTo(mainFrame);
                dialog.setVisible(true);                
                
                refreshNoteTextArea();
                mainFrame.getFileInformation().setModified(true);
                mainFrame.refreshUI();
			}            
        });
        buttonsPanel.add(editNoteButton,c);
        
        deleteNoteButton = new JButton(Localization.getString("panel.note.delete_note"));         //$NON-NLS-1$
        UserInterface.setDefaultButtonFont(deleteNoteButton);
        deleteNoteButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent arg0) {
				deleteCurrentNote();
                refreshNoteTextArea();
			}            
        });
        buttonsPanel.add(deleteNoteButton,c);
        
        // widgets constructed, now adding..
        add(noteTitle,BorderLayout.NORTH);
        add(noteTextScrollPane,BorderLayout.CENTER);
        add(buttonsPanel,BorderLayout.SOUTH);       
        
        refreshButtons();
    }

    public void stepTo(float nowTime) {
        
        if(!inEncMode) {
            currentTime = nowTime;
            
            // Get the correct note
            
            float realTime = (mainFrame.isInPlayMode()) ? mainFrame.getMSCPanel().getRealTime() : nowTime;
            Note newNote = mainFrame.getNoteManager().getNoteTimeLayerPrev(layer, realTime, true);
            
            if (!this.equals(newNote)) {
                this.note = newNote;
                refreshNoteTextArea();

                // enable/disable the buttons depending on
                // if there is a note available or not
                refreshButtons();
            }
        }
    }

    /**
     * Adds the given note to the current time and layer
     * 
     * @param noteText
     */
    public void addNoteToCurrentPosition(String noteText) {
        if(noteText != null && noteText.length() > 0) {
            Note newNote;
            
            if(inEncMode) {            
                newNote = new Note(mainFrame.getCurrentENCUnit(), noteText);
                mainFrame.getNoteManager().addNoteEnc(newNote);                
            }
            else {
                newNote = new Note(mainFrame.getCurrentLayer(),
                    /*mainFrame.getMSCPanel().getRealTime()*/currentTime, noteText);
                mainFrame.getNoteManager().addNoteTimeLayer(newNote);
            }
                        
            mainFrame.setFileModified(true);
            mainFrame.refreshUI();
            note = newNote;
        }
    }

    /**
     * Deletes the note that is currently active
     */
    protected void deleteCurrentNote() {
        if(note != null) {
            mainFrame.getNoteManager().deleteNote(note);
            note = null;
            
            mainFrame.setFileModified(true);
            refreshButtons();
            refreshNoteTextArea();
            mainFrame.refreshUI();
        }
    }
    
    /**
     *
     */
    protected void refreshNoteTextArea() {
        if(note != null && !note.getText().equals(noteTextArea.getText())){
            noteTextArea.setText(note.getText());
        }
        else if(note == null) {
            noteTextArea.setText("");
        }
    }
    
    /**
     * Updates the buttons to their correct states:
     * if there is no note before the current now line
     * edit and delete buttons are disabled. Otherwise all
     * buttons are enabled. Add note is always enabled in
     * MSC mode. In ENC mode add note is enabled only if
     * no note for the current transfer unit exists.
     */
    private void refreshButtons() {
        if (note != null) {
            editNoteButton.setEnabled(true);
            deleteNoteButton.setEnabled(true);
            
            if (inEncMode) {
                addNoteButton.setEnabled(false);
            }
        } else {
            editNoteButton.setEnabled(false);
            deleteNoteButton.setEnabled(false);
            
            if(inEncMode) {
                addNoteButton.setEnabled(true);
            }
        }
    }
}
