package fi.helsinki.dacopan.ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.ListSelectionModel;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.settings.GeneralSettings;

/**
 * The dialog window which handles the displaying of the settings
 * panels for the different animation modes.
 */
public class SettingsDialog extends javax.swing.JDialog {
	
    /** Settings to use temporary in the settings panel and later update the ones in MainFrame instance */
    private GeneralSettings tempGeneralSettings;
    private SettingsMSC tempScenarioItemSettings;
    private MainFrame mainFrame;
    private JPanel settingsPanel;
    private SettingsPanel settingsPanelMSC;
    private SettingsPanelTSC settingsPanelTSC;
	private SettingsPanelGeneral settingsPanelGeneral;
    private JList settingsModesList;
    
    
    /** Mode for changing MSC settings */
    public static final int MODE_MSC = 0;
    /** Mode for changing TSC settings */
    public static final int MODE_TSC = 1;
    /** Mode for changing General settings */
    public static final int MODE_GENERAL = 2;
    
    /**
     * Constructor.
     * 
     * @param mainFrame Current mainframe
     * @param title Title of the window
     * @param modal Is the dialog modal
     * @param mode The settings panel which should be activated when displaying the window
     */
    public SettingsDialog(MainFrame mainFrame, String title, boolean modal, int mode) {
    	this(mainFrame, title, modal);
   		this.settingsModesList.setSelectedIndex(mode);

   		// update the active settings panel
    	settingsModeChanged();
    }
    
    /**
     * Creates the dialog making the settings for the current animation type active
     */
    public SettingsDialog(MainFrame mainFrame, String title, boolean modal) {

    	super(mainFrame, title, modal);
    	this.mainFrame = mainFrame;

        /* Create the dialog that will contain the settings and buttons for acept and cancel */
        this.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
        this.setPreferredSize(new Dimension(700,400));
        this.getContentPane().setLayout(new BorderLayout());
        
        /* check if the panel can be created or display just a message of error */
        if (mainFrame.getSettings() == null){
            JOptionPane.showMessageDialog(mainFrame, Localization.getString("ui.msg_no_file_loaded"),  //$NON-NLS-1$
                            Localization.getString("ui.title_error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
            return;
        }
        
        /* temporal copy of the General Settings to work with it and update changes at the end */
        tempGeneralSettings = new GeneralSettings(mainFrame.getSettings());
        
        /* Creates a panel that contains the accept and cancel buttons */
        JPanel buttonsthis = new JPanel(new java.awt.FlowLayout());
        
        buttonsthis.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
        
        JButton buttonsthisAccept = new JButton(Localization.getString("ui.button_accept")); //$NON-NLS-1$
        JButton buttonsthisCancel = new JButton(Localization.getString("ui.button_cancel")); //$NON-NLS-1$
        JButton buttonsthisApply = new JButton(Localization.getString("ui.button_apply"));
        
        buttonsthisAccept.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent arg0) {
                acceptSettingsButtonActionPerformed(false);
            }
        });
        
        buttonsthisCancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent arg0) {
                cancelSettingsButtonActionPerformed();
            }
        });

        buttonsthisApply.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent arg0) {
        		acceptSettingsButtonActionPerformed(true);
        	}
        });
        
        buttonsthis.add(buttonsthisAccept);
        buttonsthis.add(buttonsthisApply);
        buttonsthis.add(buttonsthisCancel);
        
        /* initialize the panel for browsing between different settings*/
        this.getContentPane().add(initModeSelector(), BorderLayout.WEST);
        this.getContentPane().add(buttonsthis, BorderLayout.SOUTH);
        // update the active settings panel
        settingsModeChanged();
    }

    /* Creates a panel for choosing between General, MSC, and TSC settings */
    private JPanel initModeSelector() {
        GridBagConstraints c;
        JPanel settingsModeSelector = new JPanel (new GridBagLayout());
        settingsModeSelector.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);

        String[] modes = new String[3];
        modes[MODE_MSC] = Localization.getString("panel.msc.title");
        modes[MODE_TSC] = Localization.getString("panel.tsc.title");
        modes[MODE_GENERAL] = Localization.getString("ui.settings_general");
        settingsModesList = new JList(modes);
        settingsModesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        if (mainFrame.isInMSCMode())
        	settingsModesList.setSelectedIndex(0);
        else if (mainFrame.isInTSCMode())
        	settingsModesList.setSelectedIndex(1);
        else
        	settingsModesList.setSelectedIndex(2);

        settingsModesList.addListSelectionListener(
        		new ListSelectionListener() {
        			public void valueChanged(ListSelectionEvent e) {
        				if (!e.getValueIsAdjusting())
        					settingsModeChanged();
        			}
        		});
        

        BorderLayout layout = new BorderLayout();
        layout.setVgap(5);
        layout.setHgap(5);
        settingsModeSelector.setLayout(layout);
        
        settingsModesList.setBorder(new BevelBorder(BevelBorder.LOWERED));
        settingsModeSelector.add(settingsModesList, BorderLayout.CENTER);
        settingsModeSelector.setBorder(new EmptyBorder(5,5,5,5));
        
        return settingsModeSelector;
    }
    
    /**
     * Sets the current temporary settings object, called
     * when e.g. the settings have been loaded from file
     * 
     * @param generalSettings New settings for this dialog
     */
    public void setSettings(GeneralSettings generalSettings) {
    	this.tempGeneralSettings = generalSettings;
    	settingsModeChanged();
    }
    
    /**
     * Takes in use in the Animator the current temporary settings
     */
    public void applySettings () {
    	acceptSettingsButtonActionPerformed(true);
    }
    

    /*
     * Implements the functionality for Applying and Accepting the
     * current temporary settings of this dialog.
     * 
     * @param dialogVisible Should the dialog remain open after applying settings
     */
    private void acceptSettingsButtonActionPerformed(boolean dialogVisible){
        if (mainFrame.isInScenarioMode()) {
            mainFrame.getCurrentMSCSettings().setAllFields(tempScenarioItemSettings);
            mainFrame.refreshScenarioSettings();
        } else {
        	// Make sure that the MSC settings which are dependent of each other
        	// are in sync
            settingsPanelMSC.updateSettings();

            mainFrame.setSettings(this.tempGeneralSettings);
        }
        // Make sure the chosen language in the general settings is applied
    	settingsPanelGeneral.updateLanguage();
    	
        mainFrame.refreshSettings();
        // Settings dialog is closed when Accept button was pressed,
        // and remains open if the Apply button was pressed.
        this.setVisible(dialogVisible);
    }
    
    private void cancelSettingsButtonActionPerformed(){
        this.setVisible(false);
    }    
    
    private void saveAllSettingsButtonActionPerformed(){
    	acceptSettingsButtonActionPerformed(true);
    	mainFrame.saveAllSettings();
    }

    // Make active the settings the user has chosen to change
    private void settingsModeChanged() {
    	int newMode = this.settingsModesList.getSelectedIndex();

        /* Create panel for MSC settings */
    	if (mainFrame.isInScenarioMode()) {
            tempScenarioItemSettings = mainFrame.getCurrentMSCSettings();
            settingsPanelMSC = new SettingsPanel(mainFrame.getDataView(), mainFrame.getCurrentMSCSettings());
        } else {
            settingsPanelMSC = new SettingsPanel(mainFrame.getDataView(), this.tempGeneralSettings);
        }
        
    	// Create the other panels
        this.settingsPanelTSC = new SettingsPanelTSC(mainFrame.getDataView(), tempGeneralSettings, this);
        this.settingsPanelGeneral = new SettingsPanelGeneral(tempGeneralSettings, mainFrame, this);
    	
        // Remove old active settings panel
        if (settingsPanel != null)
    		this.getContentPane().remove(settingsPanel);
    	
        // Add the correct panel
        if (newMode == MODE_MSC)
        	this.settingsPanel = this.settingsPanelMSC;
        else if (newMode == MODE_TSC) {
        	this.settingsPanel = this.settingsPanelTSC;
        	if (!mainFrame.isTSCEnabled())
        		this.settingsPanel.setEnabled(false);
        }
        else
        	this.settingsPanel = this.settingsPanelGeneral;
    	
    	this.getContentPane().add(settingsPanel, BorderLayout.CENTER);
    	this.settingsPanel.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
    	this.settingsPanel.repaint();
    	this.pack();
    }
}