package fi.helsinki.dacopan.ui;

import java.awt.BorderLayout;
import java.util.Iterator;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.settings.GeneralSettings;


/**
 * <b>SettingsPanel </b> allows to change the MSC animation settings.
 * <p>
 * The settings panel depend on the animation mode running and show a different
 * set of options that can be changed depending on that.
 * </p>
 * <p>  
 *  In the explore mode is in which more settings are shown to be tuned. These
 *  options are the following:
 *      <ul>
 *          <li>Scale options: allow to change values related with the scale used to
 *              represent the data contained in the protocol events file or scenario file.
 *          </li>
 *          <li>Performance options: change values that affect the performance of the
 *              animation and thus the animations can be run in slower machines
 *          </li>
 *          <li>Layer options: a panel per layer is shown to select the variables and
 *              protocol header fields that can be visualized in the animation
 *          </li>
 *      </ul>
 * </p>
 * <p>
 * In scenario mode only is shown information about the variables and protocol
 * event fields to be selected for display in the animation. Also the value for the 
 * visual scale for the active layer can be set.
 * </p>
 * 
 * @version $Id: SettingsPanel.java,v 1.43 2005/05/05 20:32:34 tituomin Exp $
 */
public class SettingsPanel extends javax.swing.JPanel {
   
    /**
     * Creates a new panel to be used in EXPLORE mode
     * 
     * @param dataView
     *            reference to a DataView object to access to some data contained in the data
     *            structures
     * @param generalSettings
     *            reference to the GeneralSettings object that stores the current settings
     */
    public SettingsPanel(DataView dataView, GeneralSettings generalSettings) {
        if (dataView == null || generalSettings == null) {
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        
        List layers;
        this.dataView=dataView;
        this.generalSettings = generalSettings;
        this.auxSettingsMSC = generalSettings.getSettingsMSC(generalSettings.getDefaultLayer());

        this.setLayout(new java.awt.BorderLayout());
        this.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
        
        /* initialize the JTabPane that will contain the different settings panels */ 
        initSettingsJTabPane();

        /* initialize the Performance Settings and the Scale Settings to be used in all layers */
        initGeneralSettingsPanel();

        /* initialize the JPanels that contain the settings that can be modified in every layer */
        initAllLayersSettings();

        /* initialize the information of the last tabbed panel which was displayed */
		lastActivePanel = (JPanel) tabSettingsPanel.getSelectedComponent();
        
    }

    /*
     * Creates a new panel to be used in SCENARIO mode
     * 
     * @param dataView
     *            reference to a DataView object to access to some data contained in the data
     *            structures
     * @param generalSettings
     *            reference to the GeneralSettings object that stores the current settings
     * @param layer
     *            current layer to be modified
     */
/* Commented out as this constructor is never used
   
    public SettingsPanel(DataView dataView, GeneralSettings generalSettings, Layer layer) {
        
        if (dataView == null || generalSettings == null || layer == null) {
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        
        this.dataView = dataView;
        SettingsMSC settingsMSC = generalSettings.getSettingsMSC(layer);

        this.setLayout(new java.awt.BorderLayout());
*/        
        /* initialize the JTabPane that will contain the different settings panels */
/*        initSettingsJTabPane();
*/
        /* create and add the settings panel used to change the scale settings of the layer
         * passed as parameter */
/*       this.scaleSettingsPanel = new SettingsPanelScale(generalSettings, layer);
        this.tabSettingsPanel.addTab(
                Localization.getString("panel.settings.tab_title_scale"),
                this.scaleSettingsPanel);
*/        
        /* create and add the settings panel for the layers */
/*        JPanel settingsPanelLayer = new SettingsPanelLayer(dataView, settingsMSC ,layer, true);
        this.tabSettingsPanel.addTab(layer.getName(), settingsPanelLayer); //$NON-NLS-1$
    }
*/    
    /**
     * Creates a new panel to be used in SCENARIO mode receiving as parameter a reference to a
     * SettingMSC object
     * 
     * @param dataView
     *            reference to a DataView object to access to some data contained in the data
     *            structures
     * @param settingsMSC
     *            SettingsMSC object to create the panel
     */
    public SettingsPanel(DataView dataView, SettingsMSC settingsMSC) {
        
        if (dataView==null || settingsMSC==null){
            throw new IllegalArgumentException("Null parameters not accepted!");
        }
        this.dataView = dataView;

        this.setLayout(new java.awt.BorderLayout());
        
        /* initialize the JTabPane that will contain the different settings panels */        
        initSettingsJTabPane();

        /* create and add the settings panel used to change the scale settings of the layer
         * passed as parameter */        
        this.auxSettingsMSC = settingsMSC;
        
        JPanel settingsPanelLayer = new SettingsPanelLayer(
            dataView, 
            settingsMSC,
            settingsMSC.getLayer(), 
            true);

        /* create and add the settings panel for the layers */
        initButtonPanel();
        this.tabSettingsPanel.addTab(
            settingsMSC.getLayer().getName(), 
            settingsPanelLayer); //$NON-NLS-1$
    }
    
    
    /**
     * Initializes the panels that are specific for the EXPLORE mode. <br>
     * These are basically the panels with the performance settings and scale settings that are
     * changed for all the layers.
     */
    private void initGeneralSettingsPanel(){
        
        initButtonPanel();
   
        /* initialize and add a Performance Settings Panel */        
        this.performanceSettingsPanel = new SettingsPanelPerformance(generalSettings);
        this.tabSettingsPanel.addTab(
                Localization.getString("panel.settings.tab_title_performance"),
                this.performanceSettingsPanel); //$NON-NLS-1$        
        
    }

    // Initializes the button for switching between linear and non-linear modes.
    // the actual settings panel contained in the button panel depends on the
    // mode
    private void initButtonPanel() {

    	if (buttonPanel == null)
    		buttonPanel = new JPanel();

    	int i = this.tabSettingsPanel.indexOfComponent(buttonPanel);
        if (i != -1) {
        	tabSettingsPanel.remove(buttonPanel);
            buttonPanel = new JPanel();
        	this.tabSettingsPanel.add(buttonPanel, i);
        }
        else {
        	this.tabSettingsPanel.add(buttonPanel);
        	i = this.tabSettingsPanel.indexOfComponent(buttonPanel);
        }
        
        buttonPanel.setLayout(new BorderLayout());
        JButton buttonChangeMode;

        if (!auxSettingsMSC.isLinearMode()) {
        	// Scenario mode
        	if (generalSettings == null) {
        		this.steppingSettingsPanel = new SettingsPanelStepping(auxSettingsMSC);
        	}
        	// Explore mode
        	else {
        		this.steppingSettingsPanel = new SettingsPanelStepping(generalSettings);
        	}
            buttonPanel.add(this.steppingSettingsPanel, BorderLayout.CENTER);
        	buttonChangeMode = new JButton(
        		Localization.getString("panel.settings.button_linearmode"));
        	buttonChangeMode.setActionCommand("linear");

        	tabSettingsPanel.setTitleAt(i, Localization.getString("panel.settings.tab_title_stepping"));

        }
        else {
        	if (generalSettings == null) {
        		this.scaleSettingsPanel = new SettingsPanelScale(auxSettingsMSC);
        	}
        	else {
        		this.scaleSettingsPanel = new SettingsPanelScale(generalSettings);
        	}
            buttonPanel.add(this.scaleSettingsPanel, BorderLayout.CENTER);
            buttonChangeMode = new JButton(
            		Localization.getString("panel.settings.button_steppingmode"));
            buttonChangeMode.setActionCommand("stepping");
            tabSettingsPanel.setTitleAt(i, Localization.getString("panel.settings.tab_title_scale"));

        }

        buttonPanel.add(buttonChangeMode,
			BorderLayout.SOUTH);
        tabSettingsPanel.setSelectedIndex(i);
       
        buttonChangeMode.addActionListener(new java.awt.event.ActionListener () {
        	public void actionPerformed (java.awt.event.ActionEvent e) {
        		if ("linear".equals(e.getActionCommand())) {
        	    	auxSettingsMSC.setLinearMode(true);
        			initButtonPanel();
        		}
        		else {
        	    	auxSettingsMSC.setLinearMode(false);
        			initButtonPanel();
        		}
        		updateSettings();
        	}
        });
    }

    // Makes sure all the settings objects are in sync with each other.
	public void updateSettings () {
		// explore mode, update linearmode settings for all layers
		if (generalSettings != null)
			for (Iterator i= generalSettings.getListSettingsMSC().iterator(); i.hasNext();){
				SettingsMSC tempSettingsMSC = (SettingsMSC)i.next();
				tempSettingsMSC.setLinearMode(auxSettingsMSC.isLinearMode());
			}
		// stepping settings (minimum stepping) may depend on changed variable field settings
		// and may have to be updated
        if (steppingSettingsPanel != null)
        	steppingSettingsPanel.updateSettings();
	}    
		/**
	}
     * Creates as many SettingsPanelLayer as layers are. These panels are added to the
     * tabSettingsPanel. <br>This method is meant to be used in EXPLORE mode.
     */
    private void initAllLayersSettings(){
        List layers = this.dataView.getLayers();
        Layer auxLayer;
        SettingsPanelLayer auxSettingPanelLayer; 
        int index = 0;
        
        /* loop to create the panels and add them to the JTabPane */ 
        this.layerSettingsPanels = new SettingsPanelLayer[layers.size()];
        for (Iterator i = layers.iterator();i.hasNext();){
        	auxLayer=(Layer)i.next();
            this.layerSettingsPanels[index]=new SettingsPanelLayer(this.dataView, generalSettings.getSettingsMSC(auxLayer), auxLayer, false);
            this.tabSettingsPanel.addTab(auxLayer.getName(),this.layerSettingsPanels[index]);
            index++;
        }
    }
    
    /** 
     * initializes the JTabPane that will contain the different settings panels. After that the JTabPane is added to the
     * class (extends JFRAME) by calling the add method for the panel
     */
    private void initSettingsJTabPane() {
        tabSettingsPanel = new JTabbedPane();
        
        tabSettingsPanel.setBackground(UserInterface.DACOPAN_COLOR_PANE_BG);
        tabSettingsPanel.setForeground(UserInterface.DACOPAN_COLOR_PANE_TEXT);
        
        // Listener for making sure that the sliders in the stepping settings
        // panel are in sync with the selected variables in the layer settings.
        // Also makes sure that the settings updated in the stepping settings
        // are applied to all layers.
        tabSettingsPanel.addChangeListener(new ChangeListener() {
        	public void stateChanged(ChangeEvent e) {
        		if (lastActivePanel == null) {
        			return;
        		}
        		if (!auxSettingsMSC.isLinearMode()) {
	        		// Stepping settings were just changed
	        		if (lastActivePanel == buttonPanel) {
	        			steppingSettingsPanel.updateSettings();
	        		}
	        		else {
		        		// Stepping settings may have to be updated.
	        			// Were layer settings just changed?
	        			boolean wasLayerSettings = false;
		        		for (int i = 0; i<layerSettingsPanels.length; i++)
		        			if (lastActivePanel == layerSettingsPanels[i])
		        				wasLayerSettings = true;
		        		// If they were, update stepping settings
		        		if (wasLayerSettings) {
		        			steppingSettingsPanel.updateSliders();
		        		}
	        		}
        		}
        		lastActivePanel = (JPanel) tabSettingsPanel.getSelectedComponent();
        	}
        });
        
        this.add(tabSettingsPanel, BorderLayout.CENTER);
    }
    
    // When changing panels, the information of last changed settings
    // has to be stored in order to keep all the settings in sync
    private JPanel lastActivePanel;
    
    /** Contains the different panels for the settings that can be accessd by changing the tabs*/
    private JTabbedPane tabSettingsPanel;

    /** JPanel with the settings for the scale of the animation. <br> 
     * Used in EXPLORE and SCENARIO modes.*/
    private SettingsPanelScale scaleSettingsPanel;
    private JPanel buttonPanel;
    
    private SettingsPanelStepping steppingSettingsPanel;
 
    /** JPanel with the settings for the performance of the animation. <br> 
     * Used in EXPLORE mode.*/
    private SettingsPanelPerformance performanceSettingsPanel;

    /** contains the panel with the settings for the settings that can be set for a layer. <br> 
     * Used in EXPLORE mode (several) and SCENARIO MODE (a single one).*/
    private SettingsPanelLayer [] layerSettingsPanels;

    /** GeneralSettings instance that will be used to access to the data that will be 
     * displayed in the different panels. */
    private GeneralSettings generalSettings;
    
    private SettingsMSC auxSettingsMSC;
    
    /** DataView instance used to access to some data stored in the data structures that
     * represent the connections */
    private final DataView dataView;

    // TODO maybe the attributes defined to store the panels are not necessary because they are
    // contained in the JTabPane. Make local variables and add those ones to the JTabPane
}
