package fi.helsinki.dacopan.ui;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.font.FontRenderContext;
import java.util.Iterator;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.settings.GeneralSettings;

/**
 * A JPanel to change the parameters of the non-linear mode in MSC animation.
 * <p>
 * In explore mode, these settings are applied to all layers. This happens
 * by calling the UpdateSettings method. In scenario mode they are changed
 * for current layer only.
 */
public class SettingsPanelStepping extends JPanel {

	/**
	 * Constructor for explore mode
	 * @param generalsettings Temporary settings to modify
	 */
	public SettingsPanelStepping (GeneralSettings generalsettings) {
		this.generalSettings=generalsettings;
		this.scenarioMode = false;
		this.auxSettingsMSC = generalSettings.getSettingsMSC(generalSettings.getDefaultLayer());
		initComponents();
	}
	
	/**
	 * Constructor for scenario mode.
	 * @param settingsMSC Temporary SettingsMSC to modify
	 */
	public SettingsPanelStepping (SettingsMSC settingsMSC) {
		this.auxSettingsMSC = settingsMSC;
		this.generalSettings = null;
		this.scenarioMode = true;
		initComponents();
	}

	
	/**
	 * Updates the changed settings for all layers.
	 * Also makes sure the sliders are in sync with each other.
	 * Used in explore mode.
	 */
	public void updateSettings () {
    	updateSliders();
        if (!scenarioMode){
            for (Iterator i= generalSettings.getListSettingsMSC().iterator(); i.hasNext();){
                SettingsMSC tempSettingsMSC = (SettingsMSC)i.next();
                tempSettingsMSC.setFontSize(auxSettingsMSC.getFontSize());
                tempSettingsMSC.setMinStepping(auxSettingsMSC.getMinStepping());
                tempSettingsMSC.setMaxStepping(auxSettingsMSC.getMaxStepping());
                tempSettingsMSC.setProgressSpeed(auxSettingsMSC.getProgressSpeed());
            }
        }
	}
	
	// Calculates the maximum number of lines needed for the variables column.
	// Used to calculate the minimal minimum stepping.
	private int getMaxVariableAmount() {
        int maxVariableAmount = 1;
		if (!scenarioMode){
            for (Iterator i= generalSettings.getListSettingsMSC().iterator(); i.hasNext();){
                SettingsMSC tempSettingsMSC = (SettingsMSC)i.next();
                if (tempSettingsMSC.getVariablesColumns().length > maxVariableAmount)
                	maxVariableAmount = tempSettingsMSC.getVariablesColumns().length;
            }
		}
		else {
			maxVariableAmount = auxSettingsMSC.getVariablesColumns().length;
		}
        return maxVariableAmount;
	}
	
	// Initializes the swing components
	private void initComponents() {

		JPanel slidersPanel = new JPanel();
		
        labelMinStepping = new JLabel();
        sliderMinStepping = new JSlider();
        tfMinStepping = new JTextField();

        labelMaxStepping = new JLabel();
        sliderMaxStepping = new JSlider();
        tfMaxStepping = new JTextField();

        labelProgressSpeed = new JLabel();
        sliderProgressSpeed = new JSlider();
        tfProgressSpeed = new JTextField();

        labelFontSize = new JLabel();
        sliderFontSize = new JSlider();
        tfFontSize = new JTextField();
        
		GridBagConstraints c = new GridBagConstraints();
        slidersPanel.setLayout(new java.awt.GridBagLayout());
		
		c.weightx = 1.0;
		c.weighty = 0.0;
		c.insets = new Insets(2,2,2,2);
		labelMinStepping.setText(Localization.getString("panel.settings.stepping.label_minstepping"));
		labelMinStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_minstepping_tooltip"));
		slidersPanel.add(labelMinStepping, c);
		
		c.fill = GridBagConstraints.HORIZONTAL;
            
		slidersPanel.add(sliderMinStepping, c);

		c.fill = GridBagConstraints.NONE;
		c.gridwidth = GridBagConstraints.REMAINDER;
		tfMinStepping.setEditable(false);
		tfMinStepping.setColumns(6);
		tfMinStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_minstepping_tooltip"));
		tfMinStepping.setHorizontalAlignment(javax.swing.JTextField.TRAILING);
		slidersPanel.add(tfMinStepping, c);

		c.gridwidth = 1;
		labelMaxStepping.setText(Localization.getString("panel.settings.stepping.label_maxstepping"));
		labelMaxStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_maxstepping_tooltip"));
		slidersPanel.add(labelMaxStepping, c);
		
		c.fill = GridBagConstraints.HORIZONTAL;
		slidersPanel.add(sliderMaxStepping, c);

		c.fill = GridBagConstraints.NONE;		
		c.gridwidth=GridBagConstraints.REMAINDER;
		tfMaxStepping.setEditable(false);
		tfMaxStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_maxstepping_tooltip"));
		tfMaxStepping.setColumns(6);
		tfMaxStepping.setHorizontalAlignment(javax.swing.JTextField.TRAILING);
		slidersPanel.add(tfMaxStepping, c);
		
		c.gridwidth = 1;
		labelProgressSpeed.setText(Localization.getString("panel.settings.stepping.label_progressspeed"));
		labelProgressSpeed.setToolTipText(Localization.getString("panel.settings.stepping.label_progressspeed_tooltip"));
		slidersPanel.add(labelProgressSpeed, c);
		
		c.fill = GridBagConstraints.HORIZONTAL;
		slidersPanel.add(sliderProgressSpeed, c);

		c.fill = GridBagConstraints.NONE;		
		c.gridwidth=GridBagConstraints.REMAINDER;
		tfProgressSpeed.setEditable(false);
		tfProgressSpeed.setToolTipText(Localization.getString("panel.settings.stepping.label_progressspeed_tooltip"));
		tfProgressSpeed.setColumns(6);
		tfProgressSpeed.setHorizontalAlignment(javax.swing.JTextField.TRAILING);
		slidersPanel.add(tfProgressSpeed, c);		
		
		c.gridwidth = 1;
		labelFontSize.setText(Localization.getString("panel.settings.stepping.label_fontsize"));
		labelFontSize.setToolTipText(Localization.getString("panel.settings.stepping.label_fontsize_tooltip"));
		slidersPanel.add(labelFontSize, c);
		
		c.fill = GridBagConstraints.HORIZONTAL;
		slidersPanel.add(sliderFontSize, c);

		c.fill = GridBagConstraints.NONE;		
		c.gridwidth=GridBagConstraints.REMAINDER;
		tfFontSize.setEditable(false);
		tfFontSize.setColumns(6);
		tfFontSize.setToolTipText(Localization.getString("panel.settings.stepping.label_fontsize_tooltip"));
		tfFontSize.setHorizontalAlignment(javax.swing.JTextField.TRAILING);
		slidersPanel.add(tfFontSize, c);

		sliderFontSize.setToolTipText(Localization.getString("panel.settings.stepping.label_fontsize_tooltip"));
		sliderFontSize.setMinimum(SettingsMSC.MIN_FONT_SIZE);
		sliderFontSize.setMaximum(SettingsMSC.MAX_FONT_SIZE);
		sliderMinStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_minstepping_tooltip"));
		sliderMinStepping.setMaximum(SettingsMSC.MAX_MIN_STEPPING);
		sliderMaxStepping.setToolTipText(Localization.getString("panel.settings.stepping.label_maxstepping_tooltip"));
		sliderMaxStepping.setMaximum(SettingsMSC.MAX_MAX_STEPPING);
		sliderProgressSpeed.setToolTipText(Localization.getString("panel.settings.stepping.label_progressspeed_tooltip"));
		sliderProgressSpeed.setMinimum(SettingsMSC.MIN_PROGRESS_SPEED);
		sliderProgressSpeed.setMaximum(SettingsMSC.MAX_PROGRESS_SPEED);
		
        initSliderValues();

        updateSliderFontSize();
        updateSliderMinStepping();
        updateSliderMaxStepping();
        updateSliderProgressSpeed();
		
        sliderMinStepping.addChangeListener(new ChangeListener() {
        	public void stateChanged (ChangeEvent e) {
        		updateSliderMinStepping();
				auxSettingsMSC.setMinStepping(sliderMinStepping.getValue());
        	}
        });
        
        sliderFontSize.addChangeListener(new ChangeListener() {
        	public void stateChanged (ChangeEvent e) {
        		updateSliderFontSize();
        		auxSettingsMSC.setFontSize(sliderFontSize.getValue());
        	}
        });
       
        sliderProgressSpeed.addChangeListener(new ChangeListener() {
        	public void stateChanged (ChangeEvent e) {
        		updateSliderProgressSpeed();
        		auxSettingsMSC.setProgressSpeed(sliderProgressSpeed.getValue());
        	}
        });

        sliderMaxStepping.addChangeListener(new ChangeListener() {
        	public void stateChanged (ChangeEvent e) {
        		updateSliderMaxStepping();
        		auxSettingsMSC.setMaxStepping(sliderMaxStepping.getValue());
        	}
        });

        this.setLayout(new java.awt.GridBagLayout());
		this.add(slidersPanel);
	}

	// Initializes the sliders values from the Settings object
	private void initSliderValues() {
        sliderFontSize.setValue(auxSettingsMSC.getFontSize());
		sliderMinStepping.setValue(auxSettingsMSC.getMinStepping());
		sliderMaxStepping.setValue(auxSettingsMSC.getMaxStepping());
		sliderProgressSpeed.setValue(auxSettingsMSC.getProgressSpeed());
	}

	
	/**
	 * Makes sure the stepping settings sliders are in sync with each other.
	 */
	public void updateSliders () {
		updateSliderFontSize();
	}
	
	// Updates the value displayed in the font size slider.
	// Also updates the minimal minimum stepping, since it
	// depends on font size and chosen variables columns
	private void updateSliderFontSize () {

		tfFontSize.setText(Localization.getString(
				"panel.settings.stepping.display_font",
				new Integer(sliderFontSize.getValue())));

		int fontHeight = (int) (
			new Font("verdana", Font.PLAIN, sliderFontSize.getValue())).getMaxCharBounds(
					new FontRenderContext(null, false, false)).getHeight();

		int variableAmount = getMaxVariableAmount();
		sliderMinStepping.setMinimum(fontHeight * variableAmount);
		if (fontHeight * variableAmount > sliderMinStepping.getMaximum())
			sliderMinStepping.setMaximum(fontHeight * variableAmount);
	}
	
	// Update the min stepping sliders displayed value.
	// Also updates the minimal maximum stepping according
	// to minstepping.
	private void updateSliderMinStepping () {
		tfMinStepping.setText(Localization.getString(
				"panel.settings.stepping.display_stepping",
				new Integer(sliderMinStepping.getValue())));

		sliderMaxStepping.setMinimum(sliderMinStepping.getValue());
	}
	
	// Updates max stepping slider displayed value.
	private void updateSliderMaxStepping () {
		sliderMaxStepping.setMinimum(sliderMinStepping.getValue());
		tfMaxStepping.setText(Localization.getString(
				"panel.settings.stepping.display_stepping",
				new Integer(sliderMaxStepping.getValue())));
	}
	
	// Updates progress speed slider's displayed value.
	private void updateSliderProgressSpeed () {
		tfProgressSpeed.setText(Localization.getString(
				"panel.settings.stepping.display_speed",
				new Integer(sliderProgressSpeed.getValue())));
	}

	private JLabel labelMinStepping;
	private JSlider sliderMinStepping;
	private JTextField tfMinStepping;
	
	private JLabel labelMaxStepping;
	private JSlider sliderMaxStepping;
	private JTextField tfMaxStepping;
	
	private JLabel labelProgressSpeed;
	private JSlider sliderProgressSpeed;
	private JTextField tfProgressSpeed;
	
	private JLabel labelFontSize;
	private JSlider sliderFontSize;
	private JTextField tfFontSize;
	
	private GeneralSettings generalSettings;
	private SettingsMSC auxSettingsMSC;
	private boolean scenarioMode;
}
