/*
 * Created on 25.2.2005
 */
package fi.helsinki.dacopan.ui.tsc;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.Flow;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;


/**
 * Draws most of the TSC stuff: segments, grid etc.
 * 
 * @author mjvaanan
 */
public class DrawingArea extends JPanel implements MouseMotionListener, MouseListener {
	private static final int UNIT_MIN_HEIGHT = 5;
	private static final int GAP_TO_BORDER = 20;
	private static final int MAX_SIZE_PX = 10000000;
	private static final String FLAG_SYN = "flag_syn";
	private static final String FLAG_ACK = "flag_ack";
	private static final String FLAG_FIN = "flag_fin";
	private static final String FLAG_TRUE = "1";
	private static final String VAR_SEQ = "seq";
	private static final String VAR_ACK_SEQ = "ack_seq";
	private static final String VAR_SACK = "sack";
	private static final String VAR_WINDOW = "window";
	
	private PacketInfoPanel infoPanel;
	private JViewport view;
	private MainPanel main;
	private DataView data;
	private Map var_defs = new HashMap();	// String -> VariableDefinition
	private List events = new ArrayList();
	private int event_count;
	private Map unit_rects = new HashMap();
	private TransferUnit active_unit;
	private Host host;
	private Xaxis xAxis;
	private Yaxis yAxis;

	private final float anim_length_sec;
	private final long anim_length_seqno;
	private final int total_bytes;
	private double visible_time_range;
	private long visible_seqno_range;
	private long min_seqno;
	private long max_seqno;
	private Xaxis.Label[] xLabels;
	private Yaxis.Label[] yLabels;
	private Point cross_point;
	private DecimalFormat formatter = new DecimalFormat();
	
	private int unit_width_px = 10;
	private int hit_count = 0;

	
	public DrawingArea(MainPanel main, DataView data, int host_id, long flow_id, 
			Xaxis xAxis, Yaxis yAxis, PacketInfoPanel infoPanel) {
		this.infoPanel = infoPanel;
		setDoubleBuffered(true);
		addMouseMotionListener(this);
		addMouseListener(this);
		addComponentListener(new ComponentListener() {
			public void componentResized(ComponentEvent arg0) {
				updateUnitRects();
			}

			public void componentMoved(ComponentEvent arg0) {}

			public void componentShown(ComponentEvent arg0) {}

			public void componentHidden(ComponentEvent arg0) {}
		});
		formatter.setMaximumFractionDigits(4);

		setBackground(Color.WHITE);
		setOpaque(true);
		this.main = main;
		this.data = data;
		this.xAxis = xAxis;
		this.yAxis = yAxis;
		
		List layers = data.getLayers();
		Iterator i = layers.iterator();
		Layer trans_layer = null;
		while (i.hasNext()) {
			Layer layer = (Layer) i.next();
			if (layer.getId() == 2) {	// TODO: onko kyseenalainen tapa olettaa id??
				trans_layer = layer;
				break;
			}
		}
		
		List variables = data.getAvailableVariables(trans_layer);
		i = variables.iterator();
		while (i.hasNext()) {
			VariableDefinition var = (VariableDefinition) i.next();
			var_defs.put(var.getRawName(), var);
		}
		
		min_seqno = Long.MAX_VALUE;
		max_seqno = 0;
		int total_bytes = 0;
		if (trans_layer != null) {
			host = data.getHostById(host_id);
			List units = data.getUnitsForLayer(trans_layer);
			Iterator iterator = units.iterator();
			VariableDefinition var_seq = (VariableDefinition) var_defs.get(VAR_SEQ);
			VariableDefinition var_ack_seq = (VariableDefinition) var_defs.get(VAR_ACK_SEQ);
			if (var_seq != null && var_ack_seq != null) {
				while (iterator.hasNext()) {
					TransferUnit unit = (TransferUnit) iterator.next();
					Flow flow = unit.getFlow();
					if (flow.getId() != flow_id) {
						// skip other connections
						continue;
					}
					String protocol_name = unit.getProtocol().getRawName();
					if (protocol_name.equalsIgnoreCase("tcp")) {
						events.add(unit);
						if (unit.getSource().equals(host)) {
							String value = unit.getValue(var_seq);
							if (value == null) {
								errorDialog(Localization.getString("panel.tsc.error.no_seqno"));
							} else {
								long seq = Long.parseLong(value);
								if (seq != 0) {
									min_seqno = Math.min(seq, min_seqno);
									max_seqno = Math.max(seq, max_seqno);
								}
								if (!unit.isDropped()) {
									int payload = unit.getPayloadSize();
									total_bytes += payload;
								}
							}
						} else {
							String value = unit.getValue(var_ack_seq);
							if (value == null) {
								errorDialog(Localization.getString("panel.tsc.error.no_seqno"));
							} else {
								long seq = Long.parseLong(value);
								if (seq != 0) {
									min_seqno = Math.min(seq, min_seqno);
									max_seqno = Math.max(seq, max_seqno);
								}
							}
						}
					}
				}
			}
		}
		this.total_bytes = total_bytes;
		if (events.size() == 0) {
			errorDialog(Localization.getString("panel.tsc.error.no_tcp_data"));
		}
		
		anim_length_sec = data.getEndTime() * 1.05f;
		anim_length_seqno = max_seqno - min_seqno;
		visible_time_range = anim_length_sec;
		visible_seqno_range = anim_length_seqno;
		//System.out.println("max " + max_seqno + " min " + min_seqno);
        infoPanel.clear();
	}
	
	private SettingsTSC getSettings() {
		return main.getMainFrame().getSettings().getSettingsTSC();
	}
	
	int getEventCount() {
		return events.size();
	}
	
	private void errorDialog(String msg) {
		JOptionPane.showMessageDialog(this, msg, Localization.getString("panel.tsc.errordialog.title"), JOptionPane.ERROR_MESSAGE);
	}
	
	private void searchActive() {
		if (active_unit != null) {
			if (event_count > 0 && events.size() > 0 && events.get(event_count - 1) != active_unit) {
				for (int i = 0; i < events.size(); i++) {
					if (events.get(i) == active_unit) {
						event_count = i + 1;
						break;
					}
				}
			}
		}
	}
	
	public boolean stepForward() {
		searchActive();
		event_count++;
		if (event_count > events.size()) {
			event_count = events.size();
			return false;
		} else {
			selectActiveUnit();
			return true;
		}
	}
	
	public boolean stepBackward() {
		searchActive();
		event_count--;
		if (event_count < 0) {
			event_count = 0;
			return false;
		} else {
			selectActiveUnit();
			return true;
		}
	}
	
	public void toStart() {
		event_count = 0;
		selectActiveUnit();
	}
	
	public void toEnd() {
		event_count = events.size();
		selectActiveUnit();
	}
	
	private void selectActiveUnit() {
		if (event_count > 0) {
			active_unit = (TransferUnit) events.get(event_count - 1);
		} else {
			active_unit = null;
		}
		
		showActiveUnit();
	}
	
	private void showActiveUnit() {
		if (active_unit != null) {
			Rectangle rect = (Rectangle) unit_rects.get(active_unit);
			if (rect != null) {
				Rectangle visible = view.getViewRect();
				int gap = 50;	// minimum distance to drawing area borders
				Rectangle bounds = new Rectangle(rect.x - gap, rect.y - gap, rect.width + gap*2, rect.height + gap*2);
				if (visible.contains(bounds) == false) {
					scrollRectToVisible(bounds);
				}
				cross_point = new Point(rect.x, rect.y);
			}
            infoPanel.repaint();            
		}
	}
	
	/**
	 * Scales the time axis
	 * @param s The scale between 0 and 1
	 */
	void setXScale(double s) {
		getSettings().setXScale(s);
		visible_time_range = s * anim_length_sec;
		scale();
	}

	/**
	 * Scales the seqno axis
	 * @param s The scale between 0 and 1
	 */
	void setYScale(double s) {
		getSettings().setYScale(s);
		visible_seqno_range = Math.round(s * anim_length_seqno);
		scale();
	}
	
	private void scale() {
		Rectangle viewRect = view.getViewRect();
		int w = (int) (viewRect.width / getSettings().getXScale());
		int h = (int) (viewRect.height / getSettings().getYScale());
		int x_max_size = Math.round(anim_length_sec * 1000 * 100); // 1 ms takes max 100 px
		int y_max_size = (int) anim_length_seqno * 20;	// one segment takes max 20 px
		if (w > x_max_size) w = x_max_size;
		if (h > y_max_size) h = y_max_size;
		if (w > MAX_SIZE_PX) w = MAX_SIZE_PX;
		if (h > MAX_SIZE_PX) h = MAX_SIZE_PX;
		setPreferredSize(new Dimension(w, h));
		setSize(new Dimension(w, h));
		updateUnitRects();
		showActiveUnit();
	}
	
	void setViewport(JViewport view) {
		this.view = view;
		view.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent arg) {
				main.repaint();
			}
		});
	}
	
	private void updateLabels() {
		xAxis.setLabels(xLabels);
		yAxis.setLabels(yLabels);
	}
	
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
        
		drawGrid(g);
		updateLabels();
		Graphics2D g2 = (Graphics2D) g;
		Stroke stroke = new BasicStroke(2);
		g2.setStroke(stroke);
		drawTransferUnits(g);
		if (getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_CONNECTING_LINE)) {
			drawConnectingLines(g);
		}
		if (getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_ADVERTIZED_WINDOW_LINE)) {
			drawAdvertizedWindow(g);
		}
		if (getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_SACK)) {
			drawSACKs(g);
		}
		if (getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_CROSSHAIR)) {
			drawCross(g);
		}
	}

	private void drawConnectingLines(Graphics g) {
		int prev_x = -1;
		int prev_y = -1;
		g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_CONNECTING_LINE));
		for (int i = 0; i < event_count; i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			if (!unit.getSource().equals(host)) {
				continue;
			}
			long seqno = getSEQNO(unit);
			int unit_height_px = getUnitHeight(unit);
			int y = getYCoordForSEQNO(seqno) - unit_height_px;
			int x = getXCoordForTime(unit.getSendStart());
			if (prev_x != -1 && prev_y != -1) {
				g.drawLine(prev_x, prev_y, x, y);
			}
			prev_x = x;
			prev_y = y;
		}
	}
	
	private void drawAdvertizedWindow(Graphics g) {
		boolean first = true;
		int prev_x = 0;
		int prev_y = 0;
		g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_ADVERTIZED_WINDOW_LINE));
		Rectangle visible = view.getViewRect();
		for (int i = 0; i < event_count; i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			// only acks accepted
			if (unit.getSource().equals(host)) {
				continue;
			}
			if (unit.isDropped()) {
				continue;
			}
			int x = getXCoordForTime(unit.getReceiveStart());
			long seqno = getSEQNO(unit);
			int window = getAdvWindow(unit);
			int y = getYCoordForSEQNO(seqno) - getWindowHeight(window);
			if (!first) {
				// horizontal
				g.drawLine(prev_x, prev_y, x, prev_y);
				// vertical
				g.drawLine(x, prev_y, x, y);
			} else {
				first = false;
				g.drawLine(x, getSize().height, x, y);
			}
			prev_x = x;
			prev_y = y;
		}
		if (!first) {
			// horizontal long line
			g.drawLine(prev_x, prev_y, visible.x + visible.width, prev_y);
		}
	}
	
	private void drawSACKs(Graphics g) {
		g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_SACK));
		for (int i = 0; i < event_count; i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			// only acks accepted
			if (unit.getSource().equals(host)) {
				continue;
			}
			if (unit.isDropped()) {
				continue;
			}
			int x = getXCoordForTime(unit.getReceiveStart());
			long[] sacks = getSACKs(unit);
			for (int s = 0; s < sacks.length - 1; s+=2) {
				long seq1 = sacks[s];
				long seq2 = sacks[s+1];
				int y1 = getYCoordForSEQNO(seq1);
				int y2 = getYCoordForSEQNO(seq2);
				g.drawLine(x, y1, x, y2);
			}
		}
	}
	
	// 100-200,300-350
	private long[] getSACKs(TransferUnit unit) {
		VariableDefinition def = (VariableDefinition) var_defs.get(VAR_SACK);
		if (def == null) return new long[0];
		String str = unit.getValue(def);
		if (str == null) return new long[0];
		String[] parts = str.split(",|-");
		if (parts.length % 2 != 0) {
			throw new IllegalArgumentException("illegal sack format: " + str);
		}
		long[] sacks = new long[parts.length]; 
		for (int i = 0; i < parts.length; i++) {
			sacks[i] = Long.parseLong(parts[i].trim());
		}
		return sacks;
	}
	
	private long getSEQNO(TransferUnit unit) {
		String variable;
		if (unit.getSource().equals(host)) {
			variable = VAR_SEQ;
		} else {
			variable = VAR_ACK_SEQ;
		}
		VariableDefinition def = (VariableDefinition) var_defs.get(variable);
		if (def == null) {
			throw new IllegalStateException("no variable definition for " + variable);
		}
		String str = unit.getValue(def);
		if (str != null) {
			return Long.parseLong(str);
		} else {
			return 0;
		}
	}
	
	private int getAdvWindow(TransferUnit unit) {
		VariableDefinition def = (VariableDefinition) var_defs.get(VAR_WINDOW);
		String str = unit.getValue(def);
		if (str != null) {
			return Integer.parseInt(str);
		} else {
			return 0;
		}
	}

	private void drawTransferUnits(Graphics g) {
		if (unit_rects.size() == 0) {
			updateUnitRects();
		}
		for (int i = 0; i < event_count; i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			Rectangle rect = (Rectangle) unit_rects.get(unit);
			drawTransferUnit(g, unit, rect);
		}
	}
	
	private void drawTransferUnit(Graphics g, TransferUnit unit, Rectangle rect) {
		int x = rect.x;
		int y = rect.y;
		int unit_width_px = rect.width;
		int unit_height_px = rect.height;
		if (!isShowing(unit)) {
			return;
		}
		if (getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_DELAY_LINE)) {
			drawDelayLine(g, unit);
		}
		Color unit_color = getUnitColor(unit);
		g.setColor(unit_color);
		g.fillRect(x, y, unit_width_px-1, unit_height_px-1);
		g.setColor(Color.BLACK);
		g.drawRect(x, y, unit_width_px-1, unit_height_px-1);
		if (unit == active_unit) {
			g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_ACTIVE));
			g.drawRect(x-2, y-2, unit_width_px-1+4, unit_height_px-1+4);
		}
		if (unit.isDropped()) {
			g.setColor(Color.RED);
			g.drawLine(x-5, y-5, x+unit_width_px+5, y+unit_height_px+5);
			g.drawLine(x+unit_width_px+5, y-5, x-5, y+unit_height_px+5);
		}
	}
	
	private void drawDelayLine(Graphics g, TransferUnit unit) {
		g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_DELAY_LINE));
		if (unit.isDropped()) {
			return;
		}
		long seqno = getSEQNO(unit);
		int unit_height_px = getUnitHeight(unit);
		int y = getYCoordForSEQNO(seqno) - unit_height_px/2 - 1;
		g.drawLine(getXCoordForTime(unit.getSendStart()), y, getXCoordForTime(unit.getReceiveStart()), y);
	}
	
	private void updateUnitRects() {
		for (int i = 0; i < events.size(); i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			int unit_height_px = getUnitHeight(unit);
			long seq = getSEQNO(unit);
			int x;
			if (unit.getSource().equals(host)) {
				x = getXCoordForTime(unit.getSendStart());
			} else {
				x = getXCoordForTime(unit.getReceiveStart());
				if (hasFlag(unit, FLAG_SYN) || hasFlag(unit, FLAG_FIN) || hasFlag(unit, FLAG_ACK)) {
				} else {
					continue;
				}
			}
			int y = getYCoordForSEQNO(seq) - unit_height_px;
			Rectangle rect = new Rectangle(x, y, unit_width_px, unit_height_px);
			unit_rects.put(unit, rect);
		}
	}
	
	private boolean hasFlag(TransferUnit unit, String flag) {
		VariableDefinition def = (VariableDefinition) var_defs.get(flag);
		if (def != null) {
			String value = unit.getValue(def);
			if (value != null && value.trim().equals(FLAG_TRUE)) {
				return true;
			}
		}
		return false;
	}
	
	private int getUnitHeight(TransferUnit unit) {
		if (unit.getSource().equals(host) == false) {
			return UNIT_MIN_HEIGHT;
		}
		long dist = max_seqno - min_seqno;
		int payload = unit.getPayloadSize();
		double rel = payload / (double) dist;
		int height = (int) Math.round((getSize().height-GAP_TO_BORDER) * rel);
		if (height < UNIT_MIN_HEIGHT) {
			height = UNIT_MIN_HEIGHT;
		}
		return height;
	}
	
	private int getXCoordForTime(float time) {
		double rel = time / anim_length_sec;
		return (int) (getSize().width * rel);
	}
	
	private float getTimeForXCoord(int x) {
		double rel = (double) x / (double) getSize().width;
		return (float) (anim_length_sec * rel);
	}
	
	private int getYCoordForSEQNO(long seqno) {
		long dist = max_seqno - min_seqno;
		double rel = (seqno-min_seqno) / (double) dist;
		return (int) Math.round(getSize().height - (getSize().height-GAP_TO_BORDER) * rel);
	}
	
	private long getSEQNOForYCoord(double y) {
		y = getSize().height - y;
		double rel = (double) y / (double) getSize().height;
		long dist = max_seqno - min_seqno;
		long result = Math.round(dist * rel);
		if (!getSettings().isRelativeSEQNO()) {
			result += min_seqno;
		}
		return result;
	}
	
	private int getBytesForYCoord(int y) {
		y = getSize().height - y;
		double rel = (double) y / (double) getSize().height;
		return (int) Math.round(total_bytes * rel);
	}
	
	private int getWindowHeight(int bytes) {
		double rel = bytes / (double) total_bytes;
		return (int) Math.round(getSize().height * rel);
	}

	private float getGapForTime(float seconds) {
		return seconds * getSize().width / anim_length_sec;
	}
	
	private float getGapForSEQNOrange(int range) {
		return range * (float) getSize().height / anim_length_seqno;
	}
	
	private void drawTimeAxisGrid(float time_scale, boolean isMajor, Map xLabels, Graphics g) {
		if (isMajor) {
			g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_GRID_MAJOR));
		} else {
			g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_GRID_MINOR));
		}
		Rectangle viewRect = view.getViewRect();
		float xLabelGap = getGapForTime(time_scale);
		if (xLabelGap < 10) {	// too tight, skip this level
			return;
		}
		boolean draw = getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_GRID);
		// TODO: optimoitavaa (iteroi isoa looppia vaikka vain osa piirretn)
		int xLabelCount = Math.round(getSize().width / xLabelGap);
		for (int i = 1; i <= xLabelCount; i++) {
			int x = Math.round(i * xLabelGap);
			if (x > viewRect.x && x < viewRect.x + viewRect.width) {
				int label_pos = x - viewRect.x;
				Integer key = new Integer(label_pos);
				if (xLabels.containsKey(key) == false) {
					if (isMajor) {
						xLabels.put(key, new Xaxis.Label(label_pos, getTimeForXCoord(x)));
					}
					if (draw) {
						g.drawLine(x, viewRect.y, x, viewRect.y+viewRect.height);
					}
				}
			}
		}
	}

	private void drawSEQNOAxisGrid(int seqno_range, boolean isMajor, Map yLabels, Graphics g) {
		if (isMajor) {
			g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_GRID_MAJOR));
		} else {
			g.setColor(getSettings().getColor(SettingsTSC.COLORKEY_GRID_MINOR));
		}
		Rectangle viewRect = view.getViewRect();
		float gap = getGapForSEQNOrange(seqno_range);
		if (gap < 10) {	// too tight, skip this level
			return;
		}
		boolean draw = getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_GRID);
		int labelCount = Math.round(getSize().height / gap);
		for (int i = 1; i <= labelCount; i++) {
			double y = getSize().height - i * gap;
			if (y > viewRect.y && y < viewRect.y + viewRect.height) {
				int label_pos = (int) Math.round(y) - viewRect.y;
				Integer key = new Integer(label_pos);
				if (yLabels.containsKey(key) == false) {
					if (isMajor) {
						yLabels.put(key, new Yaxis.Label(label_pos, getSEQNOForYCoord(y)));
					}
					if (draw) {
						g.drawLine(viewRect.x, (int) Math.round(y), viewRect.x+viewRect.width, (int) Math.round(y));
					}
				}
			}
		}
	}

	private void drawGrid(Graphics g) {
		Rectangle viewRect = view.getViewRect();
		
		// vertical lines
		Map xLabels = new TreeMap();
		xAxis.setExtent(visible_time_range);
		float x_major;
		if (visible_time_range > 15) {
			x_major = 10;
		} else if (visible_time_range > 1.5) {
			x_major = 1;
		} else if (visible_time_range > 0.15) {
			x_major = 0.1f;
		} else if (visible_time_range > 0.015) {
			x_major = 0.01f;
		} else {
			x_major = 0.001f;
		}
		float x_minor = x_major / 10;
		drawTimeAxisGrid(x_major, true, xLabels, g);
		drawTimeAxisGrid(x_minor, false, xLabels, g);
		this.xLabels = (Xaxis.Label[]) xLabels.values().toArray(new Xaxis.Label[0]);
		
		// horizontal lines
		Map yLabels = new TreeMap();
		int y_major;
		if (visible_seqno_range > 10000000) {
			y_major = 10000000;
		} else if (visible_seqno_range > 1000000) {
			y_major = 1000000;
		} else if (visible_seqno_range > 100000) {
			y_major = 100000;
		} else if (visible_seqno_range > 10000) {
			y_major = 10000;
		} else if (visible_seqno_range > 1000) {
			y_major = 1000;
		} else if (visible_seqno_range > 100) {
			y_major = 100;
		} else {
			y_major = 10;
		}
		int y_minor = y_major / 10;
		drawSEQNOAxisGrid(y_major, true, yLabels, g);
		drawSEQNOAxisGrid(y_minor, false, yLabels, g);
		this.yLabels = (Yaxis.Label[]) yLabels.values().toArray(new Yaxis.Label[0]);
	}
	
	private void drawCross(Graphics g) {
		if (cross_point != null) {
			g.setColor(Color.BLACK);
		
			Rectangle viewRect = view.getViewRect();
			String vert_text = formatter.format(getTimeForXCoord(cross_point.x)) + " " + 
				Localization.getString("panel.tsc.crosshair.label.xaxis");
			int swap_x = 0;
			if (cross_point.x - viewRect.x > viewRect.width - 70) {
				swap_x = -70;
			}
			g.drawString(vert_text, cross_point.x + 5 + swap_x, viewRect.y + viewRect.height - 5);
			
			String horiz_text = getBytesForYCoord(cross_point.y) + " " + 
				Localization.getString("panel.tsc.crosshair.label.yaxis");
			int swap_y = 0;
			if (cross_point.y < 20) {
				swap_y = 20;
			}
			g.drawString(horiz_text, viewRect.x + 5, cross_point.y - 5 + swap_y);
			
			if (hit_count > 1) {
				g.setColor(Color.WHITE);
				int size = 20;
				g.fillRect(cross_point.x+2, cross_point.y-size-2, size, size);
				g.setColor(Color.BLACK);
				g.drawRect(cross_point.x+2, cross_point.y-size-2, size, size);
				g.drawString("" + hit_count, cross_point.x + 5, cross_point.y - 7);
			}
			
			Graphics2D g2 = (Graphics2D) g;
			float dash[] = {5.0f};
	        g2.setStroke(new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 
	        		10.0f, dash, 0.0f));
			g.drawLine(cross_point.x, viewRect.y, cross_point.x, viewRect.y+viewRect.height);
			g.drawLine(viewRect.x, cross_point.y, viewRect.x+viewRect.width, cross_point.y);
		}
	}
	
	public void mouseDragged(MouseEvent e) {}
	
	public void mouseMoved(MouseEvent e) {
		TransferUnit[] units = pickUnits(e.getPoint());
		hit_count = units.length;
		cross_point = e.getPoint();
		main.repaint();
	}

	public void mouseClicked(MouseEvent e) {}
	
	public void mousePressed(MouseEvent e) {
		TransferUnit[] units = pickUnits(e.getPoint());
		if (units.length > 0) {
			active_unit = units[0];
			infoPanel.repaint();
			main.repaint();
		}
	}
	
	/**
	 * Finds units intersecting pick rectangle.
	 * 
	 * TODO: optimointi mahdollista
	 */
	private TransferUnit[] pickUnits(Point p) {
		int pick_x = (int) p.getX();
		int pick_y = (int) p.getY();
		List result = new ArrayList();
		for (int i = 0; i < event_count; i++) {
			TransferUnit unit = (TransferUnit) events.get(i);
			if (!isShowing(unit)) {
				continue;
			}
			long seq = getSEQNO(unit);
			float time;
			if (unit.getSource().equals(host)) {
				time = unit.getSendStart();
			} else {
				time = unit.getReceiveStart();
			}
			int x = getXCoordForTime(time) + unit_width_px;
			int y = getYCoordForSEQNO(seq);
			if (x >= pick_x && x <= pick_x + unit_width_px) {
				if (y >= pick_y && y <= pick_y + getUnitHeight(unit)) {
					result.add(unit);
				}
			}
		}
		return (TransferUnit[]) result.toArray(new TransferUnit[0]);
	}
	
	private boolean isShowing(TransferUnit unit) {
		if (unit.getSource().equals(host)) {
			if (hasFlag(unit, FLAG_SYN)) {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_SYN)) {
					return false;
				}
			} else if (hasFlag(unit, FLAG_FIN)) {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_FIN)) {
					return false;
				}
			} else {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_SEGMENT)) {
					return false;
				}
			}
		} else {
			if (hasFlag(unit, FLAG_SYN)) {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_SYN)) {
					return false;
				}
			} else if (hasFlag(unit, FLAG_FIN)) {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_FIN)) {
					return false;
				}
			} else if (hasFlag(unit, FLAG_ACK)) {
				if (!getSettings().isElementDisplayed(SettingsTSC.ELEMENTKEY_ACK)) {
					return false;
				}
			} else {
				return false;
			}
		}
		return true;
	}
	
	private Color getUnitColor(TransferUnit unit) {
		if (hasFlag(unit, FLAG_SYN)) {
			return getSettings().getColor(SettingsTSC.COLORKEY_SYN);
		} 
		if (hasFlag(unit, FLAG_FIN)) {
			return getSettings().getColor(SettingsTSC.COLORKEY_FIN);
		}
		if (unit.getSource().equals(host)) {
			return getSettings().getColor(SettingsTSC.COLORKEY_SEGMENT);
		} else {
			if (hasFlag(unit, FLAG_ACK)) {
				return getSettings().getColor(SettingsTSC.COLORKEY_ACK);
			}
		}
		// this never happens
		return null;
	}

	public void mouseReleased(MouseEvent e) {}

	public void mouseEntered(MouseEvent e) {}

	public void mouseExited(MouseEvent e) {
		cross_point = null;
		main.repaint();
	}
	
	/**
	 * Calculates the coordinates for the specified TransferUnit. The coordinates
	 * are screen coordinates, not native DrawingArea coordinates.
	 * @param u
	 * @return Point object containing the coordinates
	 */
	public Point getUnitCoordinates(TransferUnit u) {			
        Point p = new Point();
        if(u == null) return p;
		
        Rectangle rect = (Rectangle) unit_rects.get(u);
        if(rect == null) return p;

        Rectangle screenAdj = getParent().getBounds();        
        Rectangle visRect = view.getViewRect();
                
        Point myPoint = getLocationOnScreen();
        Point mainPoint = main.getLocationOnScreen();
                
        p.x = rect.x-visRect.x;
        p.y = rect.y-visRect.y;
        
        //TODO: Remove hardcoded offset for unit location        
        p.x+=100;
        p.y+=120;
                
        return p;
	}
	
    public boolean isUnitVisible(TransferUnit u) {
    	if (isShowing(u) == false) {
    		return false;
    	}
    	boolean drawn = false;
    	for (int i = 0; i < event_count; i++) {
    		if (events.get(i) == u) {
    			drawn = true;
    			break;
    		}
    	}
    	if (!drawn) {
    		return false;
    	}
    	Rectangle uRect = (Rectangle) unit_rects.get(u);
        Rectangle vRect = view.getViewRect();        
        return SwingUtilities.isRectangleContainingRectangle(vRect,uRect);
    }
    
    public TransferUnit getActiveUnit() {
       return active_unit;
    }
    
	public void setActiveUnit(TransferUnit u) {
		if(u != null) {
			active_unit = u;
            showActiveUnit();
            main.repaint();
        }
    }
    
    public List getUnits() {
    	return events;
    }
}
