/*
 * Created on 13.4.2005
 */
package fi.helsinki.dacopan.ui.tsc;

import fi.helsinki.dacopan.Localization;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * A class to store an occurrence of some VariableDefinition
 * and its Value in a Transfer Unit, which triggers a notice to be displayed in
 * the TSC animation.
 */
public class NoticeTrigger {
	
	/** Key for equality test */
	public static final int EQUALS = 0;
	/** Key for testing if a variable exists for the transfer unit*/
	public static final int EXISTS = 1;
	/** Key for greater than test*/
	public static final int GREATER = 2;
	/** Greater than or equals test*/
	public static final int GREATER_OR_EQUALS = 3;
	/** Smaller than*/
	public static final int SMALLER = 4;
	/** Smaller than or equals*/
	public static final int SMALLER_OR_EQUALS = 5;

	/** The variable to test for */
	private VariableDefinition variable;
	/** The value to test*/
	private String value;
	/** Type of test */
	private int type;
	/** Is this trigger enabled? */
	private boolean enabled;
	/** The text to display on the notice*/
	private String notice;
	
	/**
	 * Constructor  to be used when creating a new NoticeTrigger
	 * 
	 * @param variable The VariableDefinition to test.
	 * @param value The value to test.
	 * @param type The type of test
	 * @param notice The notice text for the notice.
	 */
	public NoticeTrigger(VariableDefinition variable, String value, int type, String notice) {
		this.variable = variable;
		this.value = value;
		this.type = type;
		this.enabled = true;
		this.notice = notice;
	}

	/**
	 * Copy constructor
	 * 
	 * @param sourceTrigger
	 */
	public NoticeTrigger(NoticeTrigger sourceTrigger) {
		this.variable = sourceTrigger.variable;
		this.value = sourceTrigger.value;
		this.type = sourceTrigger.type;
		this.enabled = sourceTrigger.enabled;
		this.notice = sourceTrigger.notice;
	}
	
	/**
	 * Tests if the notice has been triggered for a 
	 * given TransferUnit.
	 * 
	 * @param unit Testable unit.
	 * @return Returns true if the trigger is enabled
	 * and if the test result is true, otherwise false.
	 */
	public boolean triggered(TransferUnit unit) {
		if (this.enabled) {
			String variableValue = unit.getValue(variable);
			double unitValue, triggerValue;
			try {
				// the tests require that the values be
				// parsable to numbers
				unitValue = Double.parseDouble(variableValue);
				triggerValue = Double.parseDouble(value);
			}
			catch (Exception e) {
				if (e instanceof NumberFormatException) {
					// The exists test can be made even without a value 
					if (type == EXISTS)
						return (unit.getValue(variable) != null);
					return false;
				}
				else {
					return false;
				}
			}
			
			// Make the correct test
			switch (this.type) {
				case EQUALS:
					return (unitValue == triggerValue);
				case EXISTS:
					return (unit.getValue(variable) != null);
				case GREATER:
					return (unitValue > triggerValue);
				case GREATER_OR_EQUALS:
					return (unitValue >= triggerValue);
				case SMALLER:
					return (unitValue < triggerValue);
				case SMALLER_OR_EQUALS:
					return (unitValue <= triggerValue);
				default:
					return false;
			}
		}
		else
			return false;
	}
	
	/**
	 * Getter for the notice text
	 * @return The text.
	 */
	public String getNotice() {
		return notice;
	}
	
	/**
	 * @return Returns true if enabled.
	 */
	public boolean isEnabled() {
		return enabled;
	}
	/**
	 * @param enabled The enabled to set.
	 */
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}
	/**
	 * @return Returns the type.
	 */
	public int getType() {
		return type;
	}
	/**
	 * @return Returns the value.
	 */
	public String getValue() {
		return value;
	}
	/**
	 * @return Returns the variable.
	 */
	public VariableDefinition getVariable() {
		return variable;
	}

	/**
	 * @param notice The notice to set.
	 */
	public void setNotice(String notice) {
		this.notice = notice;
	}
	/**
	 * @param type The type to set.
	 */
	public void setType(int type) {
		this.type = type;
	}
	/**
	 * @param value The value to set.
	 */
	public void setValue(String value) {
		this.value = value;
	}
	/**
	 * @param variable The variable to set.
	 */
	public void setVariable(VariableDefinition variable) {
		this.variable = variable;
	}	

	/* (non-Javadoc)
	 * Returns the localized string for this notice.
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String ret = this.notice;
		if (!this.enabled)
			ret += " [" + Localization.getString("panel.settings.tsc.notice_disabled") + "]";
		return ret;
	}
}
