package fi.helsinki.dacopan.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import fi.helsinki.dacopan.AbstractTestCase;

/**
 * Test case for the DataView class.
 * 
 * @version $Id: DataViewTestCase.java,v 1.16 2004/05/11 09:21:05 aarniala-dacopan Exp $
 */
public class DataViewTestCase extends AbstractTestCase {
    
    private DefaultDataView view;
    private TransferUnit firstApplication;
    private TransferUnit secondApplication;
    private long idSequence = 0;

    protected void createAndPopulateView() {
        view = new DefaultDataView();
        firstApplication = createApplicationUnit(new float[] {
                .040F, .110F,
                .050F, .120F,
                .060F, .120F,
                .070F, .095F
        });
        view.addTransferUnit(firstApplication, true);
        secondApplication = createApplicationUnit(new float[] {
                .010F, .100F,
                .030F, .090F,
                .040F, .110F,
                .060F, .105F
        });
        view.addTransferUnit(secondApplication, true);
    }
    
    /**
     * Helper method for creating bulks of data. Creates an application
     * unit that contains transport units that in turn consist of network units.
     * For example:
     * 
     *           app_1
     *     trans_1    trans_2
     *  net_1 net_2  net_3 net_4  
     */
    protected TransferUnit createApplicationUnit(float[] timestamps) {
        int length = timestamps.length;
        if(length % 8 != 0) {
            throw new IllegalArgumentException("Invalid number of timestamps");
        }
        // first create the network units
        List units = new ArrayList();
        for(int i = 0; i < length / 2; i ++) {
            units.add(new TransferUnit(getNextId(), Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, timestamps[i * 2], timestamps[i * 2 + 1], false));
        }
        // then wrap them in transport units
        List transportUnits = new ArrayList();
        for(int i = 0; i < length / 4; i++) {
            TransferUnit transport = new TransferUnit(getNextId(), Collections.EMPTY_MAP, TCP, null, HOST_A, HOST_B, units.subList(i * 2, i * 2 + 2), false);
            transportUnits.add(transport);
        }
        return new TransferUnit(getNextId(), Collections.EMPTY_MAP, HTTP, null, HOST_A, HOST_B, transportUnits, false);
    }

    /**
     * For the given lists, checks that the units at corresponding
     * positions have the same timestamps.
     */
    protected void assertTimestampsAreCorrect(String msg, List a, List b) {
        assertEquals(msg + ": sizes of lists don't match", a.size(), b.size());
        for(int i = 0; i < a.size(); i++) {
            TransferUnit first = (TransferUnit)a.get(i);
            TransferUnit second = (TransferUnit)b.get(i);
            assertEquals(msg + ": send start incorrect", first.getSendStart(), second.getSendStart(), 0);
            assertEquals(msg + ": receive end incorrect", first.getReceiveEnd(), second.getReceiveEnd(), 0);
        }
    }
    
    /* actual tests follow */
    
    public void testGetUnits() {
        
        createAndPopulateView();

        List appUnits = view.getUnitsForLayer(APPLICATION);
        List expectedAppUnits = new ArrayList();
        expectedAppUnits.add(firstApplication);
        expectedAppUnits.add(secondApplication);
        Collections.sort(expectedAppUnits);
        assertTimestampsAreCorrect("App units", expectedAppUnits, appUnits);

        List transUnits = view.getUnitsForLayer(TRANSPORT);
        List expectedTransUnits = new ArrayList();
        expectedTransUnits.addAll(firstApplication.getChildren());
        expectedTransUnits.addAll(secondApplication.getChildren());
        Collections.sort(expectedTransUnits);
        assertTimestampsAreCorrect("Trans units", expectedTransUnits, transUnits);

        List networkUnits = view.getUnitsForLayer(NETWORK);
        List expectedNetworkUnits = new ArrayList();
        for(Iterator i = expectedTransUnits.iterator(); i.hasNext(); ) {
            TransferUnit unit = (TransferUnit)i.next();
            expectedNetworkUnits.addAll(unit.getChildren());
        }
        Collections.sort(expectedNetworkUnits);
        assertTimestampsAreCorrect("Network units", expectedNetworkUnits, networkUnits);

    }
    
    public void testGetVisibleUnitsWithInvalidInterval() {
        createAndPopulateView();
        try {
            view.getUnitsForLayer(NETWORK, .100F, .090F);
            fail("Should've failed");
        } catch(IllegalArgumentException expected) { }
    }
    
    public void testGetVisibleUnits() {
        
        createAndPopulateView();

        /*
         * Timestamps of network units:
         * 10, 100,
         * 30, 90,
         * 40, 110,
         * 40, 110,
         * 50, 120,
         * 60, 105,
         * 60, 120,
         * 70, 95
         */

        assertTrue("There should be no units in 0.0-0.09", view.getUnitsForLayer(NETWORK, 0, .009F).isEmpty());
        assertTrue("There should be no units in 0.125-0.200", view.getUnitsForLayer(NETWORK, .125F, .200F).isEmpty());
        
        List visibleNetworkUnits = view.getUnitsForLayer(NETWORK, 0, .040F);
        List expectedNetworkUnits = new ArrayList();
        expectedNetworkUnits.add(createUnit(.010F, .100F));
        expectedNetworkUnits.add(createUnit(.030F, .090F));
        expectedNetworkUnits.add(createUnit(.040F, .110F));
        expectedNetworkUnits.add(createUnit(.040F, .110F));
        Collections.sort(expectedNetworkUnits);
        Collections.sort(visibleNetworkUnits);
        assertTimestampsAreCorrect("Visible network units in 0.0-0.40", expectedNetworkUnits, visibleNetworkUnits);

        visibleNetworkUnits = view.getUnitsForLayer(NETWORK, .100F, .150F);
        expectedNetworkUnits = new ArrayList();
        expectedNetworkUnits.add(createUnit(.010F, .100F));
        expectedNetworkUnits.add(createUnit(.040F, .110F));
        expectedNetworkUnits.add(createUnit(.040F, .110F));
        expectedNetworkUnits.add(createUnit(.050F, .120F));
        expectedNetworkUnits.add(createUnit(.060F, .105F));
        expectedNetworkUnits.add(createUnit(.060F, .120F));
        Collections.sort(expectedNetworkUnits);
        Collections.sort(visibleNetworkUnits);
        assertTimestampsAreCorrect("Visible network units in .100-.150", expectedNetworkUnits, visibleNetworkUnits);

        // all network layer units should be visible between 80-85
        visibleNetworkUnits = view.getUnitsForLayer(NETWORK, .080F, .085F);
        expectedNetworkUnits = view.getUnitsForLayer(NETWORK);
        Collections.sort(expectedNetworkUnits);
        Collections.sort(visibleNetworkUnits);
        assertTimestampsAreCorrect("Visible network units in .080-.085", expectedNetworkUnits, visibleNetworkUnits);
        
    }
    
    public void testGetUnitById() {
        createAndPopulateView();
        assertEquals("First application unit by id", firstApplication, view.getUnitById(firstApplication.getId()));
        assertEquals("Second application unit by id", secondApplication, view.getUnitById(secondApplication.getId()));
    }
    
    public void testStepIterator() {
        createAndPopulateView();
        /*
        * 10, 100,
        * 30, 90,
        * 40, 110,
        * 40, 110,
        * 50, 120,
        * 60, 105,
        * 60, 120,
        * 70, 95
        * 
        * --> as steps:
        * 
        * 0  1   2   3   4   5   6   7   8   9    10   11   12
        * 0, 10, 30, 40, 50, 60, 70, 90, 95, 100, 105, 110, 120
        */
        StepIterator it = view.getStepIterator(NETWORK);
        assertFalse("hasPrevious() at 0", it.hasPrevious());
        assertTrue("hasNext() at 0", it.hasNext());
        assertEquals("current() at 0", 0.0f, it.current(), 0.0f);
        assertEquals("Step 0->1", 0.01f, it.next(), 0.0);
        assertTrue("hasPrevious() at 1", it.hasPrevious());
        assertEquals("Step 1->0", 0.0f, it.previous(), 0.0);
        it.next(); // 0 -> 1
        assertEquals("Step 1->2", 0.03f, it.next(), 0.0);
        assertEquals("Step 2->3", 0.04f, it.next(), 0.0);
        assertEquals("Step 3->2", 0.03f, it.previous(), 0.0);
        it.next(); // 2 -> 3
        assertEquals("Step 3->4", 0.05f, it.next(), 0.0);
        assertEquals("Step 4->5", 0.06f, it.next(), 0.0);
        assertEquals("current() at 5", 0.06f, it.current(), 0.0f);
        it.next();
        it.next();
        it.next();
        assertEquals("Step 8->9", 0.10f, it.next(), 0.0);
        it.next();
        it.next();
        assertEquals("Step 11->12", 0.12f, it.next(), 0.0);
        assertFalse("hasNext() at end", it.hasNext());
        assertEquals("current() at end", 0.12f, it.current(), 0.0f);
        assertEquals("Step 12->11", 0.11f, it.previous(), 0.0);
    }
    
    public void testStepBackAtBeginning() {
        createAndPopulateView();
        StepIterator it = view.getStepIterator(NETWORK);
        try {
            it.previous();
            fail("Should've failed with illegal state");
        } catch(IllegalStateException expected) { }
    }
    
    public void testStepForwardAtEnd() {
        createAndPopulateView();
        StepIterator it = view.getStepIterator(NETWORK);
        for(int i = 0; i < 12; i++) {
            it.next();
        }
        assertFalse("hasNext() at end", it.hasNext());
        try {
            it.next();
            fail("Should've failed with illegal state");
        } catch(IllegalStateException expected) { }
    }
    
    public void testStepToFirst() {
        createAndPopulateView();
        StepIterator it = view.getStepIterator(NETWORK);
        it.next();
        it.next();
        assertEquals("first()", 0.0f, it.first(), 0.0f);
        assertEquals("current() at start", 0.0f, it.current(), 0.0f);
        assertTrue("hasNext() at start", it.hasNext());
        assertFalse("hasPrevious() at start", it.hasPrevious());
    }
    
    public void testStepToLast() {
        createAndPopulateView();
        StepIterator it = view.getStepIterator(NETWORK);
        assertEquals("last()", 0.12f, it.last(), 0.0f);
        assertEquals("current() at end", 0.12f, it.current(), 0.0f);
        assertFalse("hasNext() at end", it.hasNext());
        assertTrue("hasPrevious() at end", it.hasPrevious());
        view.addTransferUnit(createUnit(0.500f, 0.600f), false);
        it = view.getStepIterator(APPLICATION);
        assertEquals("Iterator should step to end also on higher layers", 0.600f, it.last(), 0.0f);
    }
    
    public void testLayersAreSorted() {
        view = new DefaultDataView();
        Layer second = new Layer(2, "Second", 10);
        view.addLayer(new Layer(1, "First", 20));
        view.addLayer(second);
        assertEquals("Layers are sorted", second, ((Layer)view.getLayers().get(0)));
    }

    public void testNextForTime() {
        createAndPopulateView();
        /*
        * 0, 10, 30, 40, 50, 60, 70, 90, 95, 100, 105, 110, 120
        */
        StepIterator it = view.getStepIterator(NETWORK);
        assertEquals("Next from 0.049", 0.050f, it.getNextForTime(0.049f), 0.0f);
        assertEquals("Next from 0.050", 0.060f, it.getNextForTime(0.050f), 0.0f);
        assertEquals("Next from 0.051", 0.060f, it.getNextForTime(0.051f), 0.0f);
        assertEquals("Next from 0.013", 0.030f, it.getNextForTime(0.013f), 0.0f);
        assertEquals("Next from 0.001", 0.010f, it.getNextForTime(0.001f), 0.0f);
        assertEquals("Next from 0.093", 0.095f, it.getNextForTime(0.093f), 0.0f);
        assertEquals("Next from 0.12", 0.12f, it.getNextForTime(0.12f), 0.0f);
        assertEquals("Next from 0.13", 0.12f, it.getNextForTime(0.13f), 0.0f);
    }

    public void testPreviousForTime() {
        createAndPopulateView();
        /*
        * 0, 10, 30, 40, 50, 60, 70, 90, 95, 100, 105, 110, 120
        */
        StepIterator it = view.getStepIterator(NETWORK);
        assertEquals("Previous from 0.0", 0.0f, it.getPreviousForTime(0.0f), 0.0f);
        assertEquals("Previous from 0.033", 0.030f, it.getPreviousForTime(0.033f), 0.0f);
        assertEquals("Previous from 0.050", 0.040f, it.getPreviousForTime(0.050f), 0.0f);
        assertEquals("Previous from 0.049", 0.040f, it.getPreviousForTime(0.049f), 0.0f);
        assertEquals("Previous from 0.125", 0.120f, it.getPreviousForTime(0.125f), 0.0f);
    }
    
    public void testStepWithHigherLevelUnits() {
        createAndPopulateView();
        StepIterator it = view.getStepIterator(APPLICATION);
        assertEquals(0.010f, it.next(), 0.0f);
        assertEquals(0.040f, it.next(), 0.0f);
        assertEquals(0.060f, it.next(), 0.0f);
        assertEquals(0.070f, it.next(), 0.0f);
        assertEquals(0.090f, it.next(), 0.0f);
        assertEquals(0.095f, it.next(), 0.0f);
        assertEquals(0.110f, it.next(), 0.0f);
        assertEquals(0.120f, it.next(), 0.0f);
    }
    
    public void testEndTime() {
        createAndPopulateView();
        assertEquals("End time of whole scenario", 0.120f, view.getEndTime(), 0.0f);
    }

    public void testEndTimeForLayer() {
        createAndPopulateView();
        assertEquals("End time of network layer", 0.120f, view.getEndTimeForLayer(NETWORK), 0.0f);
    }
    
    public void testValueMaxLength() {
        
        view = new DefaultDataView();
        view.addHost(HOST_A);
        view.addHost(HOST_B);
        view.addLayer(APPLICATION);
        view.addLayer(TRANSPORT);
        view.addLayer(NETWORK);
        view.addProtocol(HTTP);
        view.addProtocol(TCP);
        view.addProtocol(IPv4);
        VariableDefinition a = new VariableDefinition("a", VariableDefinition.UNIT_FIELD, TCP);
        VariableDefinition b = new VariableDefinition("b", VariableDefinition.UNIT_FIELD, TCP);
        VariableDefinition c = new VariableDefinition("c", VariableDefinition.UNIT_FIELD, TCP);
        VariableDefinition d = new VariableDefinition("d", VariableDefinition.UNIT_FIELD, TCP);
        TCP.getVariables().add(a);
        TCP.getVariables().add(b);
        TCP.getVariables().add(c);
        TCP.getVariables().add(d);
        
        Map values = new HashMap();
        values.put("a", "abc");
        values.put("b", "123123123");
        values.put("c", "abcdef");
        view.addTransferUnit(new TransferUnit(1, values, TCP, null, HOST_A, HOST_B, 0.0f, 0.1f, false), false);
        
        values = new HashMap();
        values.put("a", "abcdefg");
        values.put("c", "123");
        view.addTransferUnit(new TransferUnit(1, values, TCP, null, HOST_A, HOST_B, 0.0f, 0.1f, false), false);

        values = new HashMap();
        values.put("c", "1234");
        values.put("b", "abcabcabcabc");
        view.addTransferUnit(new TransferUnit(1, values, TCP, null, HOST_A, HOST_B, 0.0f, 0.1f, false), false);
        
        assertEquals("Max length of a", 7, view.getValueMaxLength(a));
        assertEquals("Max length of b", 12, view.getValueMaxLength(b));
        assertEquals("Max length of c", 6, view.getValueMaxLength(c));
        assertEquals("Max length of d", 0, view.getValueMaxLength(d));
        
        List expectedAvailableVariables = new ArrayList();
        expectedAvailableVariables.add(a);
        expectedAvailableVariables.add(b);
        expectedAvailableVariables.add(c);
        assertListsHaveSameContent("Available variables", expectedAvailableVariables, view.getAvailableVariables(TRANSPORT));
        
    }

}
