package fi.helsinki.dacopan.model;

import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fi.helsinki.dacopan.AbstractTestCase;

/**
 * Test case for the ENC tree model.
 */
public class ENCTreeModelTestCase extends AbstractTestCase {
			
	private DataView dataView;
	private List[] layeredEncTree;
	private List[] layeredEncTreeToDraw;
	private TransferUnit randomAppUnit;
	private TransferUnit randomTransUnit;
	private TransferUnit randomNetUnit;
	
	private static final int NET = 0;
	private static final int TRA = 1;
	private static final int APP = 2;

	protected void setUp(){
					
		// Create test data 
		// 1 app, 6 trans, 6 net
		int i;
		
		List netUnits = new LinkedList();
		for(i = 0; i < 6; i++) {
			TransferUnit unit = new TransferUnit(i, Collections.EMPTY_MAP, IPv4, null, HOST_A, HOST_B, (float)i, (float)(i+1), false);
			netUnits.add(unit);			
		}
			
		List transUnits = new LinkedList();
		
		i = 10;
		Iterator netUnitIterator = netUnits.iterator();
		while(netUnitIterator.hasNext()) {
			TransferUnit child = (TransferUnit)netUnitIterator.next();
			transUnits.add(new TransferUnit(i, Collections.EMPTY_MAP, TCP, null, HOST_A, HOST_B, Arrays.asList(new TransferUnit[]{child}), false));
			i++;
		}
		
		List appUnits = new LinkedList();
		TransferUnit rootUnit = new TransferUnit(20, Collections.EMPTY_MAP, HTTP, null, HOST_A, HOST_B, transUnits, false);
		appUnits.add(rootUnit);
		
		
		// Create data view
		dataView = new DefaultDataView();
		dataView.addHost(HOST_A);
		dataView.addHost(HOST_B);
		
		dataView.addLayer(NETWORK);
		dataView.addLayer(TRANSPORT);
		dataView.addLayer(APPLICATION);
				
		dataView.addProtocol(IPv4);
		dataView.addProtocol(TCP);
		dataView.addProtocol(HTTP);
		
		// Add data
		dataView.addTransferUnit(rootUnit, true);
		
		// Create layeredEncTree
	    layeredEncTree = new List[3];
	    layeredEncTree[NET] = netUnits;
	    layeredEncTree[TRA] = transUnits;
	    layeredEncTree[APP] = appUnits;
	    
	    // Select some units for testing
	    randomAppUnit = rootUnit;
	    randomTransUnit = (TransferUnit)transUnits.get(1);
	    randomNetUnit = (TransferUnit)netUnits.get(0);	    	
	}  
	 
	public void testTreeBuildingApp() {
	    ENCTreeModel treeModel = new ENCTreeModel(randomAppUnit, dataView);
	    List[] encTree = treeModel.getLayeredEncTree();
	    
	    assertListsHaveSameContent("App lists", encTree[2], layeredEncTree[2]);
	}

	public void testTreeBuildingTrans() {
	    ENCTreeModel treeModel = new ENCTreeModel(randomTransUnit,dataView);
	    List[] encTree = treeModel.getLayeredEncTree();
	    
	    assertListsHaveSameContent("Trans lists", encTree[1], layeredEncTree[1]);
	}
	
	public void testTreeBuildingNet() {
	    ENCTreeModel treeModel = new ENCTreeModel(randomNetUnit,dataView);
	    List[] encTree = treeModel.getLayeredEncTree();
	    
	    assertListsHaveSameContent("Network lists", encTree[0], layeredEncTree[0]);
	}


	// Network
	public void testDrawTreeBuildingNetMiddle() {
		
        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(1)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(2)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(3)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(1)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(2)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(3)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[NET].get(2), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertTrue("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertTrue("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertTrue("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertTrue("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertTrue("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertTrue("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());
	}

	public void testDrawTreeBuildingNetFirst() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(0)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(1)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(2)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(0)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(1)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(2)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[NET].get(0), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertFalse("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertTrue("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertFalse("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertTrue("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertFalse("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertTrue("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());
	
	}

	public void testDrawTreeBuildingNetLast() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(3)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(4)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(5)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(3)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(4)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(5)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[NET].get(5), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertTrue("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertFalse("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertTrue("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertFalse("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertTrue("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertFalse("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());

	}

	
	// Transport
	public void testDrawTreeBuildingTransMiddle() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(1)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(2)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(3)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(1)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(2)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(3)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[TRA].get(2), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertTrue("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertTrue("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertTrue("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertTrue("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertTrue("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertTrue("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());
	    
	}

	public void testDrawTreeBuildingTransFirst() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(0)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(1)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(2)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(0)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(1)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(2)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[TRA].get(0), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertFalse("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertTrue("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertFalse("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertTrue("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertFalse("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertTrue("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());
	    
    }

	public void testDrawTreeBuildingTransLast() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(3)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(4)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(5)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(3)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(4)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(5)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[TRA].get(5), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertTrue("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertFalse("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertTrue("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertFalse("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertTrue("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertFalse("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());

	}
	
	
	// Application
	public void testDrawTreeBuildingApp() {

        ENCTreeModel.DrawableEncLayer[] handMadeEncTree = new ENCTreeModel.DrawableEncLayer[3];
		handMadeEncTree[APP] = new ENCTreeModel.DrawableEncLayer(layeredEncTree[APP]);
		
		handMadeEncTree[TRA] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(0)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(1)));
		handMadeEncTree[TRA].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[TRA].get(2)));

		handMadeEncTree[NET] = new ENCTreeModel.DrawableEncLayer();
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(0)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(1)));
		handMadeEncTree[NET].getDrawableUnits().add(new ENCTreeModel.DrawableTransferUnit((TransferUnit)layeredEncTree[NET].get(2)));
		
		
		ENCTreeModel treeModel = new ENCTreeModel((TransferUnit)layeredEncTree[APP].get(0), dataView);
	    ENCTreeModel.DrawableEncLayer[] builtEncTree = treeModel.getLayeredEncTreeToDraw();
	    
	    assertFalse("App unit continues to left", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheLeft());
	    assertTrue("App unit continues to right", ((ENCTreeModel.DrawableTransferUnit)builtEncTree[APP].getDrawableUnits().get(0)).hasMoreToTheRight());
	    
	    assertFalse("Tra has siblings to left", builtEncTree[TRA].hasSiblingsToTheLeft());
	    assertTrue("Tra has siblings to right", builtEncTree[TRA].hasSiblingsToTheRight());
	    assertFalse("Net has siblings to left", builtEncTree[NET].hasSiblingsToTheLeft());
	    assertTrue("Net has siblings to right", builtEncTree[NET].hasSiblingsToTheRight());
	    
	    assertListsHaveSameContent("Net middle - APP", handMadeEncTree[APP].getDrawableUnits(), builtEncTree[APP].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - TRA", handMadeEncTree[TRA].getDrawableUnits(), builtEncTree[TRA].getDrawableUnits());
	    assertListsHaveSameContent("Net middle - NET", handMadeEncTree[NET].getDrawableUnits(), builtEncTree[NET].getDrawableUnits());

	}

}
