package fi.helsinki.dacopan.pef;

import java.io.InputStreamReader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import fi.helsinki.dacopan.AbstractTestCase;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.DefaultDataView;
import fi.helsinki.dacopan.model.Flow;
import fi.helsinki.dacopan.model.Host;
import fi.helsinki.dacopan.model.Layer;
import fi.helsinki.dacopan.model.Link;
import fi.helsinki.dacopan.model.Protocol;
import fi.helsinki.dacopan.model.StaticVariable;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * Test case for the XML protocol events reader. Due to its extensive use
 * of DataView, this isn't a unit test exactly, rather an integration test
 * between ProtocolEventsReader and DataView.
 * 
 * @version $Id: XMLProtocolEventsReaderTestCase.java,v 1.13 2005/04/25 17:26:57 pamietti Exp $
 */
public class XMLProtocolEventsReaderTestCase extends AbstractTestCase {
    
    protected XMLProtocolEventsReader reader = new XMLProtocolEventsReader();
    protected DataView view = new DefaultDataView();
    
    protected Map hosts;
    protected Map layers;
    protected Map protocols;
    protected Map flows;
    protected Map links;

    /**
     * Sets up data for the test case. This is basically a reconstruction
     * of most of the data in the test PEF instance.
     */
    protected void setUp() {
        
        hosts = new LinkedHashMap();

        Host a = new Host(1, "123.123.123.123", "somehost.com");
        hosts.put(new Long(1), a);
        
        Host b = new Host(2, "111.111.111.111", null);
        b.setStaticVariables(Collections.singletonList(new StaticVariable("HOST_CONSTANT", "def", b, null, null)));
        hosts.put(new Long(2), b);
        
        layers = new LinkedHashMap();
        Layer network = new Layer(1, "network", 0);
        Layer transport = new Layer(2, "transport", 10);
        Layer application = new Layer(3, "application", 20);
        layers.put(new Long(1), network);
        layers.put(new Long(2), transport);
        layers.put(new Long(3), application);

        flows = new LinkedHashMap();
        flows.put(new Long(1), new Flow(1, a, 1024, b, 2048));
        flows.put(new Long(2), new Flow(2, a, 1111, b, 2334));

        protocols = new LinkedHashMap();
        
        Protocol ipv4 = new Protocol(1, "ipv4", (Layer)layers.get(new Long(1)));
        protocols.put(new Long(1), ipv4);
        List ipVariables = new ArrayList();
        ipVariables.add(new VariableDefinition("IP_1", VariableDefinition.UNIT, ipv4));
        ipVariables.add(new VariableDefinition("IP_2", VariableDefinition.UNIT_FIELD, ipv4));
        ipVariables.add(new VariableDefinition("SHARED", VariableDefinition.UNIT, ipv4));
        ipv4.setVariables(ipVariables);
        network.getProtocols().add(ipv4);

        Protocol tcp = new Protocol(2, "tcp", (Layer)layers.get(new Long(2)));
        protocols.put(new Long(2), tcp);
        StaticVariable h1andTcp = new StaticVariable("PROTO_AND_HOST", "ghi", a, null, tcp);
        a.setStaticVariables(Collections.singletonList(h1andTcp));
        List tcpStaticVariables = new ArrayList();
        tcpStaticVariables.add(new StaticVariable("TCP_CONSTANT", "abc", null, null, tcp));
        tcpStaticVariables.add(h1andTcp);
        tcp.setStaticVariables(tcpStaticVariables);
        transport.getProtocols().add(tcp);

        Protocol udp = new Protocol(3, "udp", (Layer)layers.get(new Long(2)));
        protocols.put(new Long(3), udp);
        List udpVariables = new ArrayList();
        udpVariables.add(new VariableDefinition("UDP_X", VariableDefinition.FLOW, udp));
        udpVariables.add(new VariableDefinition("UDP_Y", VariableDefinition.UNIT_FIELD, udp));
        udpVariables.add(new VariableDefinition("SHARED", VariableDefinition.UNIT, udp));
        // we need payload_size to get over stupid problem with "purkkapaikka" at PEF reader
        udpVariables.add(new VariableDefinition("payload_size", VariableDefinition.UNIT_FIELD, udp));
        udp.setVariables(udpVariables);
        transport.getProtocols().add(udp);

        Protocol http = new Protocol(4, "http", (Layer)layers.get(new Long(3)));
        protocols.put(new Long(4), http);
        List httpVariables = new ArrayList();
        httpVariables.add(new VariableDefinition("STATUS", VariableDefinition.UNIT_FIELD, http));
        httpVariables.add(new VariableDefinition("data", VariableDefinition.UNIT_FIELD, http));
        http.setVariables(httpVariables);
        application.getProtocols().add(http);
        
        links = new LinkedHashMap();
        Link k1 = new Link(1, (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)));
        links.put(new Long(1), k1);
        k1.setStaticVariables(Collections.singletonList(new StaticVariable("MTU", "1234", null, k1, null)));

    }
    
    /**
     * Reads in the given classpath resource.
     */
    protected void readResource(String resource) {
        reader.read(new InputStreamReader(getClass().getResourceAsStream(resource)), view);
    }
    
    // actual tests follow
    
    public void testDTDResolving() throws Exception {
        assertNotNull("System id should resolve to a DTD", reader.resolveEntity(null, XMLProtocolEventsReader.DTD_SYSTEM_ID));
    }
    
    public void testReadWithInvalidDocumentFails() {
        try {
            reader.read(new StringReader("<?xml version=\"1.0\"?><invalid/>"), view);
            fail("Reading an invalid document didn't fail");
        } catch(ProtocolEventsDataException expected) { }
    }
    
    public void testReadWithWrongDoctypeFails() {
        try {
            reader.read(new StringReader("<?xml version=\"1.0\"?><!DOCTYPE abc [ <!ELEMENT abc EMPTY> ]><abc/>"), view);
            fail("Reading a document valid according to a wrong dtd didn't fail");
        } catch(ProtocolEventsDataException expected) {
            assertEquals("Wrong error message", "Wrong doctype", expected.getMessage());
        }
    }
    
    public void testHosts() {
        readResource("a.xml");
        assertListsHaveSameContent("Host listing", new ArrayList(hosts.values()), view.getHosts());
    }
    
    public void testFlows() {
        readResource("a.xml");
        assertListsHaveSameContent("Flows not correct", new ArrayList(flows.values()), view.getFlows());
    }
    
    public void testLinks() {
        readResource("a.xml");
        assertEquals("Link not correct", new ArrayList(links.values()), view.getLinks());
    }
    
    public void testLayers() {
        readResource("a.xml");
        List expectedLayers = new ArrayList(layers.values());
        List viewLayers = view.getLayers();
        assertEquals("Incorrect number of layers", 3, viewLayers.size());
        Layer last = null;
        for(int i = 0; i < layers.size(); i++) {
            Layer layer = (Layer)viewLayers.get(i);
            // check names of layers
            assertEquals("Layer name incorrect", ((Layer)expectedLayers.get(i)).getName(), layer.getName());
            // check list of protocols
            assertListsHaveSameContent("Layer.getProtocols for " + layer, ((Layer)expectedLayers.get(i)).getProtocols(), layer.getProtocols());
            if(last != null) {
                // if there is a previous layer, this layer must be above it
                assertTrue("Ordering of layers incorrect", layer.compareTo(last) > 0);
            }
        }
    }
    
    public void testProtocols() {
        readResource("a.xml");
        List expectedProtocols = new ArrayList(protocols.values());
        assertListsHaveSameContent("Protocols incorrect", expectedProtocols, view.getProtocols());
    }
    
    public void testVariables() {
        readResource("a.xml");

        Protocol ipv4 = view.getProtocolById(1);
        List expectedIp = new ArrayList();
        expectedIp.add(new VariableDefinition("IP_1", VariableDefinition.UNIT, ipv4));
        expectedIp.add(new VariableDefinition("IP_2", VariableDefinition.UNIT_FIELD, ipv4));
        expectedIp.add(new VariableDefinition("SHARED", VariableDefinition.UNIT, ipv4));
        assertListsHaveSameContent("IPv4", expectedIp, ipv4.getVariables());
        assertEquals(ipv4, ((VariableDefinition)ipv4.getVariables().get(2)).getProtocol());

        Protocol udp = view.getProtocolById(3);
        List expectedUdp = new ArrayList();
        expectedUdp.add(new VariableDefinition("UDP_X", VariableDefinition.FLOW, udp));
        expectedUdp.add(new VariableDefinition("UDP_Y", VariableDefinition.UNIT_FIELD, udp));
        expectedUdp.add(new VariableDefinition("SHARED", VariableDefinition.UNIT, udp));
        // fix for "purkkapaikka"
        expectedUdp.add(new VariableDefinition("payload_size", VariableDefinition.UNIT_FIELD, udp));
        assertListsHaveSameContent("UDP vars", expectedUdp, udp.getVariables());
        assertEquals(udp, ((VariableDefinition)udp.getVariables().get(2)).getProtocol());

    }
    
    public void testUnits() {
        readResource("a.xml");
        Map values = new HashMap();
        values.put("IP_1", "123");
        values.put("IP_2", "456");
        values.put("payload_size", "-1");
        TransferUnit u1 = new TransferUnit(1, values, (Protocol)protocols.get(new Long(1)), null, (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)), 0.0f, 0.03f, false);
        assertEquals("u1 incorrect", u1, view.getUnitById(1));
        values = new HashMap();
        values.put("IP_1", "789");
        values.put("IP_2", "123");
        values.put("payload_size", "-1");
        TransferUnit u2 = new TransferUnit(2, values, (Protocol)protocols.get(new Long(1)), null, (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)), 0.05f, 0.09f, false);
        assertEquals("u2 incorrect", u2, view.getUnitById(2));
        values = new HashMap();
        values.put("IP_1", "456");
        values.put("IP_2", "234");
        values.put("payload_size", "-1");
        TransferUnit u3 = new TransferUnit(3, values, (Protocol)protocols.get(new Long(1)), null, (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)), 0.10f, 0.16f, true);
        assertEquals("u3 incorrect", u3, view.getUnitById(3));
        List u12 = Arrays.asList(new Object[] { u1, u2 });
        values = new HashMap();
        values.put("UDP_X", "foo");
        values.put("UDP_Y", "bar");
        values.put("payload_size", "-1");
        TransferUnit u4 = new TransferUnit(4, values, (Protocol)protocols.get(new Long(3)), (Flow)flows.get(new Long(2)), (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)), u12, false);
        assertEquals("u4 incorrect", u4, view.getUnitById(4));
        assertListsHaveSameContent("u4's children", u4.getChildren(), view.getUnitById(4).getChildren());
        values = new HashMap();
        values.put("STATUS", "200");
        values.put("response_data", "<html><head><title>Blaa</title></head><body>Blaa blaa blalblalba</body></html>");
        values.put("payload_size", "-1");
        List u5Kids = new ArrayList();
        u5Kids.add(u4);
        TransferUnit u5 = new TransferUnit(5, values, (Protocol)protocols.get(new Long(4)), null, (Host)hosts.get(new Long(1)), (Host)hosts.get(new Long(2)), u5Kids, false);
        assertEquals("u5 incorrect", u5, view.getUnitById(5));
        assertListsHaveSameContent("u5's children", u5.getChildren(), view.getUnitById(5).getChildren());
    }
    
    public void testSentUnitsWithNoCorrespondingEndTag() {
        try {
            readResource("c.xml");
            fail("Should've failed");
        } catch(ProtocolEventsDataException e) {
            assertEquals("Wrong exception thrown", "No corresponding unit_received or unit_dropped for unit_sent with id 1", e.getMessage());
        }
    }
    
    public void testSentUnitsWithMissingChildren() {
        try {
            readResource("e.xml");
            fail("Should've failed");
        } catch(ProtocolEventsDataException expected) { }
    }
    
    public void testVariablesReferenceOwners() {
        readResource("a.xml");
        Link k1 = view.getLinkById(1);
        StaticVariable mtu = (StaticVariable)k1.getStaticVariables().get(0);
        assertEquals("MTU's reference to owner incorrect", k1, mtu.getLink());
        Host h1 = view.getHostById(1);
        Protocol p2 = view.getProtocolById(2);
        StaticVariable h1p2 = (StaticVariable)h1.getStaticVariables().get(0);
        assertEquals("PROTO_AND_HOST's reference to host incorrect", h1, h1p2.getHost());
        assertEquals("PROTO_AND_HOST's reference to protocol incorrect", p2, h1p2.getProtocol());
    }
    
    public void testConstantsRequireAtLeastOneOwner() {
        try {
            readResource("b.xml");
            fail("Should've failed");
        } catch(ProtocolEventsDataException e) {
            assertEquals("Wrong exception thrown", "Constants require owner(s)", e.getMessage());
        }
    }

    public void testHandlesChildrenInAnyOrder() {
        readResource("d.xml");
    }

}
