package fi.helsinki.dacopan.scenario;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.HashSet;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;

import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.DefaultDataView;
import fi.helsinki.dacopan.pef.XMLProtocolEventsReader;

import junit.framework.TestCase;

/**
 * Test case for the ScenarioFile class.
 * 
 * @version $Id: ScenarioFileTestCase.java,v 1.7 2005/04/25 15:18:14 mjvaanan Exp $
 */
public class ScenarioFileTestCase extends TestCase {

    private File jar;
    private Collection saveables;
    private Collection values;

    private static int counter;

    /**
     * Creates a temporary file that's deleted at the end of the test run.
     */
    private void createFile() throws IOException {
        if(jar != null) {
            throw new IllegalStateException("A file already created for this test case");
        }
        File tempDir = new File(System.getProperty("java.io.tmpdir"));
        assertTrue("Temp dir doesn't exist", tempDir.isDirectory());
        jar = new File(tempDir, "dacopan_scenario_file_test" + (++counter));
        assertFalse("Temp file already exists", jar.exists());
    }
    
    /**
     * Returns a protocol data reader.
     */
    private Reader getProtocolDataReader() {
        return new InputStreamReader(getClass().getResourceAsStream("/fi/helsinki/dacopan/pef/a.xml"));
    }

    /**
     * Sets up sample values for the tests.
     */
    protected void setUp() {
        saveables = new HashSet();
        values = new HashSet();
        saveables.add(new A("abc"));
        values.add("abc");
        saveables.add(new B(123));
        values.add(new Integer(123));
    }
    
    /**
     * Deletes the archive created during the latest run, if any.
     */
    protected void tearDown() throws Exception {
        if(jar != null) {
            jar.delete();
        }
    }
    
    // actual test cases
    
    public void testCreateNew() throws Exception {
        createFile();
        ScenarioFile file = new ScenarioFile(jar, getProtocolDataReader());
        assertEquals("File incorrect", jar, file.getArchiveFile());
        DummySerializer dummy = new DummySerializer();
        file.setObjectSerializer(dummy);
        file.save(saveables);
        assertEquals("Saved values incorrect", values, dummy.values);
    }
    
    public void testCreateNewWithNullParameters() throws Exception {
        createFile();
        try {
            ScenarioFile file = new ScenarioFile(null, getProtocolDataReader());
            fail("Should've failed with null file");
        } catch (IllegalArgumentException e) { }
        try {
            ScenarioFile file = new ScenarioFile(jar, (Reader)null);
            fail("Should've failed with null PEF");
        } catch (IllegalArgumentException e) { }
    }

    public void testLoadNonExisting() throws Exception {
        try {
            ScenarioFile file = new ScenarioFile(new File("no_such_file_should_exist"));
            fail("Didn't fail with non-existing file");
        } catch(FileNotFoundException e) { }
    }

    public void testLoadArbitraryFile() throws Exception {
        createFile();
        FileWriter writer = new FileWriter(jar);
        writer.write("Sure, this ain't no scenario data");
        writer.close();
        try {
            new ScenarioFile(jar);
            fail("Reading an arbitrary file should've failed");
        } catch(InvalidScenarioFileException expected) {
            assertEquals("Invalid Jar archive", expected.getMessage());
        }
    }
    
   /* public void testLoadRegularJarFile() throws Exception {
        createFile();
        JarOutputStream out = new JarOutputStream(new FileOutputStream(jar));
        out.putNextEntry(new JarEntry("some_entry"));
        out.write("text".getBytes());
        out.closeEntry();
        out.close();
        try {
            new ScenarioFile(jar);
            fail("Reading a normal Jar file should've failed");
        } catch(InvalidScenarioFileException expected) {
            assertEquals("Protocol data missing", expected.getMessage());
        }
    }*/

    public void testLoadExisting() throws Exception {
        createFile();
        ScenarioFile file = new ScenarioFile(jar, getProtocolDataReader());
        file.save(saveables);
        ScenarioFile loaded = new ScenarioFile(jar);
        assertEquals("File incorrect", jar, loaded.getArchiveFile());
        Collection emptySaveables = new HashSet();
        A a = new A();
        B b = new B();
        emptySaveables.add(a);
        emptySaveables.add(b);
        loaded.load(emptySaveables);
        assertEquals("Loaded values incorrect", saveables, emptySaveables);
    }
    
    public void testOverwrite() throws Exception {
        createFile();
        // save first batch
        ScenarioFile file = new ScenarioFile(jar, getProtocolDataReader());
        file.save(saveables);
        // load file, overwrite existing
        ScenarioFile loaded = new ScenarioFile(jar);
        Collection overwritten = new HashSet();
        overwritten.add(new A("overwritten"));
        overwritten.add(new B(123456));
        loaded.save(overwritten);
        // load back values
        ScenarioFile loadedAgain = new ScenarioFile(jar);
        Collection empty = new HashSet();
        empty.add(new A());
        empty.add(new B());
        loadedAgain.load(empty);
        assertEquals("Loaded values incorrect", overwritten, empty);
    }
    
    public void testSaveManyObjectsOfSameType() throws IOException {
        createFile();
        ScenarioFile file = new ScenarioFile(jar, getProtocolDataReader());
        Collection duplicates = new HashSet();
        duplicates.add(new A("first"));
        duplicates.add(new A("second"));
        try {
            file.save(duplicates);
            fail("Should've failed when saving multiple objects of same type");
        } catch(IllegalArgumentException expected) {
            assertEquals("Can't save multiple objects of same type", expected.getMessage());
        }
    }
    
    public void testProtocolData() throws Exception {
        createFile();
        ScenarioFile file = new ScenarioFile(jar, getProtocolDataReader());
        file.save(saveables);
        ScenarioFile loaded = new ScenarioFile(jar);
        DataView expected = new DefaultDataView();
        new XMLProtocolEventsReader().read(getProtocolDataReader(), expected);
        DataView actual = new DefaultDataView();
        loaded.populateView(actual);
        assertEquals("Protocol data incorrect", expected, actual);
    }
    
    class A implements Saveable {
        private String s;
        private A() { }
        private A(String s) {
            this.s = s;
        }
        public Object getData() {
            return s;
        }
        public void setData(Object o) {
            this.s = (String)o;
        }
        public boolean equals(Object o) {
            return (o instanceof A && ((A)o).s.equals(s));
        }
        public int hashCode() {
            return s != null ? s.hashCode() : 0;
        }
        public String toString() {
            return s;
        }
    }
    
    class B implements Saveable {
        private int i;
        private B() { }
        private B(int i) {
            this.i = i;
        }
        public Object getData() {
            return new Integer(i);
        }
        public void setData(Object o) {
            this.i = ((Integer)o).intValue();
        }
        public boolean equals(Object o) {
            return (o instanceof B && ((B)o).i == i);
        }
        public int hashCode() {
            return i;
        }
        public String toString() {
            return Integer.toString(i);
        }
    }

    /**
     * Dummy serializer, each object passed to save is put to a collection.
     */
    class DummySerializer implements ObjectSerializer {
        
        private Collection values = new HashSet();

        public void saveObject(Object o, Writer w) {
            values.add(o);
        }

        public Object loadObject(Reader r) {
            throw new UnsupportedOperationException("Tried to call loadObject()");
        }

        public void setDataView(DataView view) {
        }
        
    }
    
}
