package fi.helsinki.dacopan.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import fi.helsinki.dacopan.AbstractTestCase;
import fi.helsinki.dacopan.model.DataView;
import fi.helsinki.dacopan.model.TransferUnit;
import fi.helsinki.dacopan.model.VariableDefinition;

/**
 * @author pamietti
 *
 * JUnit tests for CalcYCoord. Tests that y-coordinates are in order and that minimum stepping and
 * maximum gap are as they should be. It is not possible to test maximum stepping for random data
 * case because minimum stepping might have changed the length of packet line. Test are also done
 * for fat units.
 */
public class CalcYCoordTestCase extends AbstractTestCase {
	
	private static long NETWORK_ID = 1;
	private static long APPLICATION_ID = 3;
	
	public void testYCoordsAreInOrder() {
		
		DataView data = AbstractTestCase.getRandomData(1, 100, 10, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(NETWORK_ID), new VariableDefinition[0], new VariableDefinition[0]);
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(NETWORK_ID));
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			times.add(new Float(unit.getSendStart()));
			times.add(new Float(unit.getReceiveEnd()));
		}
		Collections.sort(times);
		for (int i = 1; i < times.size(); i++) {
			int prevYCoord = coord.getYCoordForTime(((Float)times.get(i-1)).floatValue());
			int thisYCoord = coord.getYCoordForTime(((Float)times.get(i)).floatValue());
			assertTrue("events y-coords are not in order", prevYCoord <= thisYCoord);
		}
		for (float i = 0.1f; i <= data.getEndTimeForLayer(data.getLayerById(3)); i += 0.1f) {
			int prevYCoord = coord.getYCoordForTime(i-0.1f);
			int thisYCoord = coord.getYCoordForTime(i);
			assertTrue("random y-coords are not in order", prevYCoord <= thisYCoord);
		}
		for (int i = coord.getFirstYCoord() + 1; i < coord.getLastYCoord(); i++) {
			float prevTime = coord.getTimeForYCoord(i-1);
			float thisTime = coord.getTimeForYCoord(i);
			assertTrue("random times are not in order", prevTime <= thisTime);
		}
	}
	
	public void testMinimumStepping() {
		
		DataView data = AbstractTestCase.getRandomData(1, 100, 10, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(NETWORK_ID), new VariableDefinition[0], new VariableDefinition[0]);
		int minStepping = 10;
		settings.setMinStepping(minStepping);
		settings.setMaxStepping(minStepping+1);
		assertEquals("minStepping is not correctly set", minStepping, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", minStepping+1, settings.getMaxStepping());
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(NETWORK_ID));
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			EventClass event = new EventClass(unit.getSendStart(), unit.getId());
			times.add(event);
			event = new EventClass(unit.getReceiveEnd(), unit.getId());
			times.add(event);
		}
		Collections.sort(times);
		for (int i = 1; i < times.size(); i++) {
			float prevTime = ((EventClass)times.get(i-1)).getTime();
			int prevYCoord = coord.getYCoordForTime(prevTime);
			long prevId = ((EventClass)times.get(i-1)).getId();
			float thisTime = ((EventClass)times.get(i)).getTime();
			int thisYCoord = coord.getYCoordForTime(thisTime);
			long thisId = ((EventClass)times.get(i)).getId();
			assertTrue("space is smaller than minStepping: "+thisYCoord+", "+prevYCoord+"; ids:"
					+thisId+", "+prevId, 
					(thisYCoord - prevYCoord >= minStepping && prevId != thisId && thisTime != prevTime)
					|| (thisYCoord >= prevYCoord && prevId == thisId)
					|| (thisYCoord == prevYCoord && thisTime == prevTime));
		}
		
		minStepping = 1000;
		settings.setMinStepping(minStepping);
		settings.setMaxStepping(minStepping);
		assertEquals("minStepping is not correctly set", minStepping, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", minStepping, settings.getMaxStepping());
		coord = new CalcYCoord(data, settings);
		for (int i = 1; i < times.size(); i++) {
			float prevTime = ((EventClass)times.get(i-1)).getTime();
			int prevYCoord = coord.getYCoordForTime(prevTime);
			long prevId = ((EventClass)times.get(i-1)).getId();
			float thisTime = ((EventClass)times.get(i)).getTime();
			int thisYCoord = coord.getYCoordForTime(thisTime);
			long thisId = ((EventClass)times.get(i)).getId();
			assertTrue("space is smaller than large minStepping: "+thisYCoord+", "+prevYCoord+"; ids:"
					+thisId+", "+prevId, 
					(thisYCoord - prevYCoord >= minStepping && prevId != thisId && thisTime != prevTime)
					|| (thisYCoord >= prevYCoord && prevId == thisId)
					|| (thisYCoord == prevYCoord && thisTime == prevTime));	
		}
	}
	
	public void testMaximumGap() {
		
		DataView data = AbstractTestCase.getRandomData(1, 100, 10, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(NETWORK_ID), new VariableDefinition[0], new VariableDefinition[0]);
		int maxStepping = 10;
		settings.setMinStepping(1);
		settings.setMaxStepping(maxStepping);
		assertEquals("minStepping is not correctly set", 1, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", maxStepping, settings.getMaxStepping());
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(NETWORK_ID));
		assertEquals("not have correct number of units", units.size(), 100 * 10);
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			EventClass event = new EventClass(unit.getSendStart(), unit.getId());
			event.setSendStart(true);
			times.add(event);
			event = new EventClass(unit.getReceiveEnd(), unit.getId());
			event.setReceiveEnd(true);
			times.add(event);
		}
		Collections.sort(times);
		
		for (int i = 1; i < times.size(); i++) {
			float prevTime = ((EventClass)times.get(i-1)).getTime();
			int prevYCoord = coord.getYCoordForTime(prevTime);
			long prevId = ((EventClass)times.get(i-1)).getId();
			float thisTime = ((EventClass)times.get(i)).getTime();
			int thisYCoord = coord.getYCoordForTime(thisTime);
			long thisId = ((EventClass)times.get(i)).getId();
			assertTrue("gap is bigger than maximum gap: "+thisYCoord+", "+prevYCoord+"; ids:"
					+thisId+", "+prevId+"; times:"+thisTime+", "+prevTime,
					thisYCoord - prevYCoord <= maxStepping+10);
		}
	}
	
	public void testDiscreteYAxis() {
		
		DataView data = AbstractTestCase.getRandomData(1, 100, 10, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(NETWORK_ID), new VariableDefinition[0], new VariableDefinition[0]);
		int stepping = 1;
		settings.setMinStepping(stepping);
		settings.setMaxStepping(stepping);
		assertEquals("minStepping is not correctly set", stepping, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", stepping, settings.getMaxStepping());
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(NETWORK_ID));
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			EventClass event = new EventClass(unit.getSendStart(), unit.getId());
			times.add(event);
			event = new EventClass(unit.getReceiveEnd(), unit.getId());
			times.add(event);
		}
		Collections.sort(times);
		
		for (int i = 1; i < times.size(); i++) {
			int prevYCoord = coord.getYCoordForTime(((EventClass)times.get(i-1)).getTime());
			int thisYCoord = coord.getYCoordForTime(((EventClass)times.get(i)).getTime());
			assertTrue("y-axis is not discrete: "+thisYCoord+", "+prevYCoord,
					thisYCoord - prevYCoord <= stepping);
		}
	}
	
	public void testMinimumSteppingForFatUnits() {
		
		DataView data = AbstractTestCase.getRandomData(1000, 2, 2, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(APPLICATION_ID), new VariableDefinition[0], new VariableDefinition[0]);
		int minStepping = 10;
		settings.setMinStepping(minStepping);
		settings.setMaxStepping(minStepping+1);
		assertEquals("minStepping is not correctly set", minStepping, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", minStepping+1, settings.getMaxStepping());
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(APPLICATION_ID));
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			EventClass event = new EventClass(unit.getSendStart(), unit.getId());
			event.setSendStart(true);
			times.add(event);
			event = new EventClass(unit.getReceiveEnd(), unit.getId());
			event.setReceiveEnd(true);
			times.add(event);
			event = new EventClass(unit.getReceiveStart(), unit.getId());
			event.setReceiveStart(true);
			times.add(event);
			event = new EventClass(unit.getSendEnd(), unit.getId());
			event.setSendEnd(true);
			times.add(event);
		}
		Collections.sort(times);
		for (int i = 1; i < times.size(); i++) {
			float prevTime = ((EventClass)times.get(i-1)).getTime();
			int prevYCoord = coord.getYCoordForTime(prevTime);
			long prevId = ((EventClass)times.get(i-1)).getId();
			float thisTime = ((EventClass)times.get(i)).getTime();
			int thisYCoord = coord.getYCoordForTime(thisTime);
			long thisId = ((EventClass)times.get(i)).getId();
			assertTrue("space is smaller than minStepping: "+thisYCoord+", "+prevYCoord+"; ids:"
					+thisId+", "+prevId, 
					(thisYCoord - prevYCoord >= minStepping && prevId != thisId && thisTime != prevTime)
					|| (thisYCoord >= prevYCoord && prevId == thisId)
					|| (thisYCoord == prevYCoord && thisTime == prevTime));
		}
		
		minStepping = 1000;
		settings.setMinStepping(minStepping);
		settings.setMaxStepping(minStepping);
		assertEquals("minStepping is not correctly set", minStepping, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", minStepping, settings.getMaxStepping());
		coord = new CalcYCoord(data, settings);
		for (int i = 1; i < times.size(); i++) {
			float prevTime = ((EventClass)times.get(i-1)).getTime();
			int prevYCoord = coord.getYCoordForTime(prevTime);
			long prevId = ((EventClass)times.get(i-1)).getId();
			float thisTime = ((EventClass)times.get(i)).getTime();
			int thisYCoord = coord.getYCoordForTime(thisTime);
			long thisId = ((EventClass)times.get(i)).getId();
			assertTrue("space is smaller than large minStepping: "+thisYCoord+", "+prevYCoord+"; ids:"
					+thisId+", "+prevId, 
					(thisYCoord - prevYCoord >= minStepping && prevId != thisId && thisTime != prevTime)
					|| (thisYCoord >= prevYCoord && prevId == thisId)
					|| (thisYCoord == prevYCoord && thisTime == prevTime));
		}
	}
	
	public void testMaximumGapForFatUnits() {
		
		DataView data = AbstractTestCase.getRandomData(1, 100, 10, 0.1f, false);
		SettingsMSC settings = new SettingsMSC(data.getHostById(1), data.getHostById(2), 
				data.getLayerById(APPLICATION_ID), new VariableDefinition[0], new VariableDefinition[0]);
		int maxStepping = 10;
		settings.setMinStepping(1);
		settings.setMaxStepping(maxStepping);
		assertEquals("minStepping is not correctly set", 1, settings.getMinStepping());
		assertEquals("maxStepping is not correctly set", maxStepping, settings.getMaxStepping());
		CalcYCoord coord = new CalcYCoord(data, settings);
		List units = data.getUnitsForLayer(data.getLayerById(APPLICATION_ID));
		List times = new ArrayList();
		Iterator iter = units.iterator();
		while (iter.hasNext()) {
			TransferUnit unit = (TransferUnit)iter.next();
			EventClass event = new EventClass(unit.getSendStart(), unit.getId());
			event.setSendStart(true);
			times.add(event);
			event = new EventClass(unit.getReceiveEnd(), unit.getId());
			event.setReceiveEnd(true);
			times.add(event);
			event = new EventClass(unit.getReceiveStart(), unit.getId());
			event.setReceiveStart(true);
			times.add(event);
			event = new EventClass(unit.getSendEnd(), unit.getId());
			event.setSendEnd(true);
			times.add(event);
		}
		Collections.sort(times);
		
		for (int i = 1; i < times.size(); i++) {
			int prevYCoord = coord.getYCoordForTime(((EventClass)times.get(i-1)).getTime());
			int thisYCoord = coord.getYCoordForTime(((EventClass)times.get(i)).getTime());
			assertTrue("gap is bigger than maximum gap: "+thisYCoord+", "+prevYCoord,
					thisYCoord - prevYCoord <= maxStepping + 10);
		}
	}

	
	private class EventClass implements Comparable {

		private float time;
		private long id;
		private boolean fatUnit = false;
		private boolean sendStart = false;
		private boolean sendEnd = false;
		private boolean receiveStart = false;
		private boolean receiveEnd = false;
		
		public EventClass() {
			this.time = 0.0f;
		}
		
		public EventClass(float time, long id) {
			this.time = time;
			this.id = id;
		}
		/* (non-Javadoc)
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(Object o) {
			EventClass other = (EventClass)o;
			return Float.compare(time, other.getTime());
		}
		
		public boolean equals(EventClass e) {
			return (this.time == e.getTime() && this.id == e.getId());
		}
		
		public float getTime() {
			return time;
		}
		
		
		public long getId() {
			return id;
		}
		
		public void setTime(float time) {
			this.time = time;
		}
		
		
		public void setFatUnit(boolean fatUnit) {
			this.fatUnit = fatUnit;
		}
		
		public void setSendStart(boolean sendStart) {
			this.sendStart = sendStart;
		}
		
		public void setSendEnd(boolean sendEnd) {
			this.sendEnd = sendEnd;
		}
		
		public void setReceiveStart(boolean receiveStart) {
			this.receiveStart = receiveStart;
		}
		
		public void setReceiveEnd(boolean receiveEnd) {
			this.receiveEnd = receiveEnd;
		}
		
		public boolean isFatUnit() {
			return fatUnit;
		}
		
		public boolean isSendStart() {
			return sendStart;
		}
		
		public boolean isSendEnd() {
			return sendEnd;
		}
		
		public boolean isReceiveStart() {
			return receiveStart;
		}
		
		public boolean isReceiveEnd() {
			return receiveEnd;
		}
	}


}
